#ifndef _ALIGNER_COMMON_DDEQUE_H_
#define _ALIGNER_COMMON_DDEQUE_H_

#include <assert.h>
#include <deque>
#include <iostream>
#include <fstream>
#include <iterator>
#include <algorithm>
#include <ext/algorithm>

using __gnu_cxx::copy_n;

/**
 * A ``dumpable'' deque: contains methods which allow *efficiently*
 * dumping the contents to a stream, and reconstructing a DDeque
 * from such a dump.
 *
 * Works under the assumption that the contents of the deque are
 * either DDeques, or primitives, or structs/classes containing
 * no pointers (i.e. the objects can be reconstructed by simply
 * copying the memory they occupy).
 *
 * The output produced by operator>>() is *NOT* portable.  Only
 * reload the contents of a DDeque on the same (or an identical)
 * machine as the one which produced the output!
 *
 * Working around brokenness in the STL deque implementation.
 * The problem occurs when the size of a deque is over 2^31,
 * because deque::iterator::operator+() takes signed, rather
 * than unsigned, 32-bit values.  When deque::operator[]()
 * and deque::resize() take values larger than 2^31, and this
 * value is added to _M_start, it becomes negative, and all
 * hell breaks loose.
 *
 * The workaround is kind of ugly, and only fixes these two
 * specific instances of the problem.  (There are probably deque
 * methods other than operator[]() and resize() which are broken
 * for the same reason; the real fix would be to rewrite
 * deque::iterator, which I neither have the time nor energy
 * to do.)  It should be noted that the workaround is both
 * platform-specific (assumes that sizeof(size_type) >= 4)
 * and heavily implementation-specific (assumes that nodes
 * are exactly size 512)
 *
 * -afn
 */

template< typename T, typename Alloc = std::allocator<T> >
class DDeque : public std::deque<T, Alloc>
{
protected:
	typedef std::deque<T, Alloc> _Base;

public:
	typedef typename _Base::reference       reference;
	typedef typename _Base::const_reference const_reference;
	typedef typename _Base::size_type       size_type;
	typedef typename _Base::value_type      value_type;
	typedef typename _Base::iterator        iterator;
	typedef typename _Base::const_iterator  const_iterator;

protected:
	/* The maximum amount it's safe to advance an iterator by */
	static const size_type MAX_SIZE = 0x7fffffff - (sizeof(T) < 512 ? size_type(512 / sizeof(T)) : size_type(1));

public:
	DDeque() : _Base() { }

	inline reference operator[](size_type __n)
	{
		iterator it = this->begin();

		while(__n > MAX_SIZE)
		{
			it += MAX_SIZE;
			__n -= MAX_SIZE;
		}

		return *(it + __n);
	}

	inline const_reference operator[](size_type __n) const
	{
		const_iterator it = this->begin();

		while(__n > MAX_SIZE)
		{
			it += MAX_SIZE;
			__n -= MAX_SIZE;
		}

		return *(it + __n);
	}

	inline void resize(size_type __n, const value_type & __v)
	{
		const size_type __len  = this->size();

		if(__n < __len)
		{
			iterator it = this->begin();
			while(__n > MAX_SIZE)
			{
				it += MAX_SIZE;
				__n -= MAX_SIZE;
			}

			it += __n;

			erase( it, this->end() );
		}
		else
		{
			__n -= __len;

			while(__n > MAX_SIZE)
			{
				insert( this->end(), MAX_SIZE, __v );
				__n -= MAX_SIZE;
			}

			insert( this->end(), __n, __v );
		}
	}

	inline void resize(size_type __n)
	{
		resize(__n, value_type());
	}

	//
	// Declaring << and >> operators as friend methods
	//

	template< typename _T, class _CharT, class _Traits >
	friend std::basic_ostream< _CharT, _Traits > &
	operator<<( std::basic_ostream< _CharT, _Traits > & stream,
	            const DDeque<_T> & deque );

	template< typename _T, class _CharT, class _Traits >
	friend std::basic_istream< _CharT, _Traits > &
	operator>>( std::basic_istream< _CharT, _Traits > & stream,
		    DDeque<_T> & deque );
};

/**
 * Stream << DDeque of primitives
 */
template< typename _T, class _CharT, class _Traits >
std::basic_ostream< _CharT, _Traits > &
operator<<( std::basic_ostream< _CharT, _Traits > & stream,
	    const DDeque<_T> & __d )
{
	size_t size = __d.size();
	stream.write( (char *) &size, sizeof(size) );

	for( _T *cur        = __d.begin()._M_cur,  /* next data item to process */
	        *last       = __d.end()._M_cur,    /* last data item */
	        **cur_node  = __d.begin()._M_node, /* pointer to beginning of node containing cur */
	        **last_node = __d.end()._M_node ;  /* pointer to beginning of node containing last */
	     cur != last ; )
	{
		// Get pointer to end of the current node
		_T *node_end = *cur_node + DDeque<_T>::iterator::_S_buffer_size();

		// If we're on the last node, then rewind node_end to the end of actual data
		if( cur_node == last_node )
			node_end = last;

		// Write (node_end - cur) bytes to stream
		size_t bytes_to_write = ((node_end - cur)*(sizeof (_T))) / (sizeof (const char));
		stream.write( (const char *) cur, bytes_to_write );

		// Advance to beginning of next node
		if( cur_node != last_node )
			cur = *(++cur_node);
		else
			cur = last;
	}

	return stream;
}

/**
 * Stream << DDeque of DDeques
 */
template< typename _T, class _CharT, class _Traits >
std::basic_ostream< _CharT, _Traits > &
operator<<( std::basic_ostream< _CharT, _Traits > & stream,
	    const DDeque< DDeque< _T > > & deque )
{
	size_t size = deque.size();
	stream.write( (char *) &size, sizeof(size) );

	copy_n( deque.begin(),
	        size,
	        std::ostream_iterator< DDeque<_T> > (stream) );

	return stream;
}

/**
 * Stream >> DDeque of primitives
 */
template< typename _T, class _CharT, class _Traits >
std::basic_istream< _CharT, _Traits > &
operator>>( std::basic_istream< _CharT, _Traits > & stream,
	    DDeque<_T> & __d )
{
	size_t size;
	stream.read( (char *) &size, sizeof(size) );
	assert( stream );
	__d.resize( size );

	for( _T *cur        = __d.begin()._M_cur,  /* next data item to process */
	        *last       = __d.end()._M_cur,    /* last data item */
	        **cur_node  = __d.begin()._M_node, /* pointer to beginning of node containing cur */
	        **last_node = __d.end()._M_node ;  /* pointer to beginning of node containing last */
	     cur != last ; )
	{
		// Get pointer to end of the current node
		_T *node_end = *cur_node + DDeque<_T>::iterator::_S_buffer_size();

		// If we're on the last node, then rewind node_end to the end of actual data
		if( cur_node == last_node )
			node_end = last;

		// Read (node_end - cur) bytes from
		stream.read( (char *) cur, ((node_end - cur)*(sizeof (_T))) / (sizeof (char)) );
		assert( stream );

		// Advance to beginning of next node
		if( cur_node != last_node )
			cur = *(++cur_node);
		else
			cur = last;
	}

	return stream;
}

/**
 * Stream >> DDeque of DDeques
 */
template< typename _T, class _CharT, class _Traits >
std::basic_istream< _CharT, _Traits > &
operator>>( std::basic_istream< _CharT, _Traits > & stream,
	    DDeque< DDeque< _T > > & deque )
{
	size_t size;
	stream.read( (char *) &size, sizeof(size) );
	deque.resize( size );

	for( size_t i = 0 ; i < size ; i++ )
	{
		stream >> deque[i];
	}

	return stream;
}

/**
 * Save: Filename << DDeque
 */
template< typename _T, class _CharT, class _Traits >
void
operator<<( const std::basic_string< _CharT, _Traits > & filename,
            const DDeque< _T > & deque )
{
	std::ofstream os( filename.c_str() );
	os << deque;
}

/**
 * Save: DDeque >> Filename
 */
template< typename _T, class _CharT, class _Traits >
void
operator>>(  const DDeque< _T > & deque ,
             const std::basic_string< _CharT, _Traits > & filename )
{
	operator<<( filename, deque );
}

/**
 * Load: Filename >> DDeque
 */
template< typename _T, class _CharT, class _Traits >
void
operator>>( const std::basic_string< _CharT, _Traits > & filename,
            DDeque< _T > & deque )
{
	std::ifstream is( filename.c_str() );
	is >> deque;

	// Sanity check: ensure that there's no trailing garbage in the file
 	assert( is.peek() == EOF );
}

/**
 * Load: DDeque << Filename
 */
template< typename _T, class _CharT, class _Traits >
void
operator<<( DDeque< _T > & deque ,
            const std::basic_string< _CharT, _Traits > & filename )
{
	operator>>( filename, deque );
}

#endif //_ALIGNER_COMMON_DDEQUE_H_
