#ifndef _ALIGNER_COMMON_DVECTOR_H_
#define _ALIGNER_COMMON_DVECTOR_H_

#include "assert.h"
#include "Vec.h"
#include <iostream>
#include <fstream>
#include <iterator>
#include <ext/algorithm>

using __gnu_cxx::copy_n;

/**
 * A ``dumpable'' vector: contains methods which allow *efficiently*
 * dumping the contents to a stream, and reconstructing a DVector
 * from such a dump.
 *
 * Works under the assumption that the contents of the vector are
 * either DVectors, or primitives, or structs/classes containing
 * no pointers (i.e. the objects can be reconstructed by simply
 * copying the memory they occupy).  In particular, I would not
 * expect classes with virtual methods to serialize properly.
 *
 * The output produced by operator>>() is *NOT* portable.  Only
 * reload the contents of a DVector on the same (or an identical)
 * machine as the one which produced the output!
 *
 * -afn
 */

template< typename T, typename Alloc = std::allocator<T> >
class DVector : public Vec<T, Alloc>
{
protected:
	typedef Vec<T, Alloc>                  _Parent;
public:
	typedef typename _Parent::size_type    size_type;
protected:

	/**
	 * Return the raw data in the vector
	 */
	char *getRawData()
	{
		return (char *) &*this->begin();
	}

	/**
	 * Return the raw data in the vector
	 */
	const char *getRawData() const
	{
		return (const char *) &*this->begin();
	}

	/**
	 * Return the length of the raw data (in chars)
	 */
	size_type getRawDataLength() const
	{
		return (char *) &*this->end() - (char *) &*this->begin();
	}

public:
	DVector() : _Parent() { }

	//
	// Declaring << and >> operators as friend methods
	//

	template< typename _T, class _CharT, class _Traits >
	friend std::basic_ostream< _CharT, _Traits > &
	operator<<( std::basic_ostream< _CharT, _Traits > & stream,
	            const DVector<_T> & vector );

	template< typename _T, class _CharT, class _Traits >
	friend std::basic_istream< _CharT, _Traits > &
	operator>>( std::basic_istream< _CharT, _Traits > & stream,
		    DVector<_T> & vector );
};

/**
 * Stream << DVector of primitives
 */
template< typename _T, class _CharT, class _Traits >
std::basic_ostream< _CharT, _Traits > &
operator<<( std::basic_ostream< _CharT, _Traits > & stream,
	    const DVector<_T> & vector )
{
	size_t size = vector.size();
	stream.write( (char *) &size, sizeof(size) );
	return stream.write( vector.getRawData(), vector.getRawDataLength() );
}

/**
 * Stream << DVector of DVectors
 */
template< typename _T, class _CharT, class _Traits >
std::basic_ostream< _CharT, _Traits > &
operator<<( std::basic_ostream< _CharT, _Traits > & stream,
	    const DVector< DVector< _T > > & vector )
{
	size_t size = vector.size();
	stream.write( (char *) &size, sizeof(size) );

	copy( vector.begin(),
	      vector.end(),
	      std::ostream_iterator< DVector<_T> > (stream) );

	return stream;
}

/**
 * Stream >> DVector of primitives
 */
template< typename _T, class _CharT, class _Traits >
std::basic_istream< _CharT, _Traits > &
operator>>( std::basic_istream< _CharT, _Traits > & stream,
	    DVector<_T> & vector )
{
	size_t size;
	stream.read( (char *) &size, sizeof(size) );
	assert( stream );

	vector.resize( size );

	// Haha, this is awesome: stream.read() takes a signed value for the length.
	// Man, I *LOVE* the STL!  It's also totally cool that there's no warning
	// about the casting between signed and unsigned types-- woohoo! C++ rulez
	//
	// The following loop replaces the line:
	//     stream.read( vector.getRawData(), vector.getRawDataLength() );

	for( typename DVector<_T>::size_type __n = vector.getRawDataLength(), __off = 0 ;
	     __n > 0 ; )
	{
		typename DVector<_T>::size_type bytes = __n > 0x7fffffff ? 0x7fffffff : __n;
		stream.read( vector.getRawData() + __off, bytes );
		assert( stream );

		__n   -= bytes;
		__off += bytes;
	}

	assert( stream );

	return stream;
}

/**
 * Stream >> DVector of DVectors
 */
template< typename _T, class _CharT, class _Traits >
std::basic_istream< _CharT, _Traits > &
operator>>( std::basic_istream< _CharT, _Traits > & stream,
	    DVector< DVector< _T > > & vector )
{
	size_t size;
	stream.read( (char *) &size, sizeof(size) );
	assert( stream );
	vector.resize( size );

	for( size_t i = 0 ;
	     i < size ;
	     ++i )
	{
		stream >> vector[i];
	}

	return stream;
}

/**
 * Save: Filename << DVector
 */
template< typename _T, class _CharT, class _Traits >
void
operator<<( const std::basic_string< _CharT, _Traits > & filename,
            const DVector< _T > & vector )
{
	std::ofstream os( filename.c_str() );
	os << vector;
}

/**
 * Save: DVector >> Filename
 */
template< typename _T, class _CharT, class _Traits >
void
operator>>(  const DVector< _T > & vector ,
             const std::basic_string< _CharT, _Traits > & filename )
{
	operator<<( filename, vector );
}

/**
 * Load: Filename >> DVector
 */
template< typename _T, class _CharT, class _Traits >
void
operator>>( const std::basic_string< _CharT, _Traits > & filename,
            DVector< _T > & vector )
{
	std::ifstream is( filename.c_str() );
	is >> vector;

	// Sanity check: ensure that there's no trailing garbage in the file
 	assert( is.peek() == EOF );
}

/**
 * Load: DVector << Filename
 */
template< typename _T, class _CharT, class _Traits >
void
operator<<( DVector< _T > & vector ,
            const std::basic_string< _CharT, _Traits > & filename )
{
	operator>>( filename, vector );
}

#endif //_ALIGNER_COMMON_DVECTOR_H_
