#include "Profile.h"
#include <math.h>

void ProfileColumn::set_to(float a_, float c_, float g_, float t_,
										float ge_, float go_, float gc_, float gs_,
										float gt_, float undef_) {
	_freqs[ NUCL_A_PROF_IDX ] = a_;
	_freqs[ NUCL_C_PROF_IDX ] = c_;
	_freqs[ NUCL_G_PROF_IDX ] = g_;
	_freqs[ NUCL_T_PROF_IDX ] = t_;

	_freqs[ GAP_EXTEND_PROF_IDX ] = ge_;
	_freqs[ GAP_OPEN_PROF_IDX   ] = go_;
	_freqs[ GAP_CLOSE_PROF_IDX  ] = gc_;
	_freqs[ GAP_SINGLE_PROF_IDX ] = gs_;
	_freqs[ GAPS_TOTAL_PROF_IDX ] = gt_;
	_freqs[ UNDEFINED_PROF_IDX  ] = undef_;
}	

ostream& ProfileColumn::print_consensus(ostream &o) const {
	
	const_iterator max_iter = max_element(_freqs.begin(), _freqs.end());
	assert( max_iter < _freqs.end() );
	
	int idx = (int) ( max_iter - _freqs.begin() );
	
	o << IDX_PROFILE_CHAR[ idx ];
	return o;
}

float match_freqs_of_column_pair( const ProfileColumn &c1, const ProfileColumn &c2 ) {

	float match_freqs = 0;
	match_freqs += c1[ NUCL_A_PROF_IDX ] * c2[ NUCL_A_PROF_IDX ];
	match_freqs += c1[ NUCL_C_PROF_IDX ] * c2[ NUCL_C_PROF_IDX ];
	match_freqs += c1[ NUCL_G_PROF_IDX ] * c2[ NUCL_G_PROF_IDX ];
	match_freqs += c1[ NUCL_T_PROF_IDX ] * c2[ NUCL_T_PROF_IDX ];
	return match_freqs;
}


void Profile::initialize_from_seq(const Seq &s, Seq::size_type begin_pos, Seq::size_type end_pos,
                                  bool normalize_bp_freqs) {

	assert(!_valid);
	assert(end_pos - begin_pos > 0 && begin_pos > 0 && end_pos <= s.getLength());
	assert( s.getLength() >= 2 );
	
	_freqs.resize(s.getLength()+1, ProfileColumn());
	
	Seq::const_iterator e_iter = s.getIterator(end_pos);
	iterator f_iter = _freqs.begin() + begin_pos;
	for (Seq::const_iterator i_iter = s.getIterator(begin_pos); i_iter <= e_iter; ++i_iter, ++f_iter)
		(*f_iter)[ CHAR_PROFILE_IDX[ (int) *i_iter ] ] = 1;	

	if (normalize_bp_freqs)
	{
		iterator e_iter = get_end();
		for (iterator i_iter = get_begin(); i_iter < e_iter; ++i_iter)
		{
			ProfileColumn &c = *i_iter;
			if (c[ UNDEFINED_PROF_IDX ] > SMALL_FLOAT)
				c[ NUCL_A_PROF_IDX ] = 
					c[ NUCL_C_PROF_IDX ] = 
					c[ NUCL_G_PROF_IDX ] = 
					c[ NUCL_T_PROF_IDX ] = .25;
		}
	}
	
	_num_seqs = 1;
	_valid    = true;
}


void Profile::initialize_from_mseq(const Mseq &m, unsigned int begin_pos, unsigned int end_pos,
                                   const Vec<float> &weights, bool normalize_bp_freqs) {
	
	assert(!_valid);
	assert(_num_seqs = m.getSize());
	assert(end_pos - begin_pos > 0);
	
	size_type this_length = (size_type) (end_pos - begin_pos + 1);
	
	_freqs.resize(this_length + 1, ProfileColumn() );
	
	Mseq::const_iterator       seq_end = m.getEnd();
	Vec<float>::const_iterator wt_iter = weights.begin();
	Vec<float>::const_iterator wt_end  = weights.end();

 	for (Mseq::const_iterator seq_iter = m.getBegin(); seq_iter < seq_end; ++seq_iter, ++wt_iter) {
 		Seq::const_iterator pos_iter = seq_iter->getBegin() + begin_pos - 1;
 		Seq::const_iterator pos_end  = pos_iter + end_pos;
 		assert(pos_end <= seq_iter->getEnd());
		
 		if (pos_end == seq_iter->getEnd())
 			pos_end--; // if end_pos is the last pos in mseq, we have to treat end specially

		assert(wt_iter < wt_end);
 		float seq_weight = *wt_iter;

		iterator f_iter = get_begin();
		
 		if (begin_pos == 1) { // if we are at the beginning of mseq, we cannot get precise gap info for begin_pos
 			(*f_iter)[ CHAR_PROFILE_IDX[ (int) (*pos_iter) ]] += seq_weight;

 			if (*pos_iter == '-') {
				//(*f_iter)[ GAPS_TOTAL_PROF_IDX ] += seq_weight; /* Already counted previously */

			  if (*(pos_iter+1) != '-') // it does not make sense to classify the first gap as GAP_SINGLE
 				  (*f_iter)[ GAP_CLOSE_PROF_IDX ] += seq_weight;   
				else
					(*f_iter)[ GAP_EXTEND_PROF_IDX ] += seq_weight;
			}
 			++pos_iter;
			++f_iter;
 		}
 		// else, we are allowed to look at the position preceding begin_pos
		
 		for (; pos_iter < pos_end; ++pos_iter, ++f_iter) {

 			(*f_iter)[ CHAR_PROFILE_IDX[ (int) (*pos_iter) ]] += seq_weight;
			
			if (*pos_iter == '-') {
				//(*f_iter)[ GAPS_TOTAL_PROF_IDX ] += seq_weight; /* Already counted previously */
				if (*(pos_iter - 1) == '-') {
					
					if (*(pos_iter + 1) == '-') // gap extension
						(*f_iter)[ GAP_EXTEND_PROF_IDX ] += seq_weight;
					else                        // gap close
						(*f_iter)[ GAP_CLOSE_PROF_IDX  ] += seq_weight;
				}
				else // previous position is not a gap
					if (*(pos_iter + 1) == '-') // gap open
						(*f_iter)[ GAP_OPEN_PROF_IDX   ] += seq_weight;
					else                        // single gap!
						(*f_iter)[ GAP_SINGLE_PROF_IDX ] += seq_weight;
			}				
		}
 		if (pos_end+1 == seq_iter->getEnd()) { // then, pos_iter is now at the last pos of mseq, == end_pos
 			(*f_iter)[CHAR_PROFILE_IDX[ (int) (*pos_iter) ]] += seq_weight;
			
 			if (*pos_iter == '-') {
				//(*f_iter)[ GAPS_TOTAL_PROF_IDX ] += seq_weight; /* Already counter previously */

				if (*(pos_iter - 1) != '-') // it does not make sense to classify the last gap as GAP_SINGLE
					(*f_iter)[ GAP_OPEN_PROF_IDX ] += seq_weight; 
 				else (*f_iter)[ GAP_EXTEND_PROF_IDX ] += seq_weight;
			}
 		}
 	}

	// the following procedure normalizes the sum of letter frequencies to 1, if the flag is set
	if (normalize_bp_freqs) {
		iterator f_iter = get_begin();
		iterator f_end  = get_end();
		
		for (; f_iter < f_end; ++f_iter) {
			float acgt = 
				(*f_iter)[ NUCL_A_PROF_IDX ] + (*f_iter)[ NUCL_C_PROF_IDX ] + 
				(*f_iter)[ NUCL_G_PROF_IDX ] + (*f_iter)[ NUCL_T_PROF_IDX ];
			if ( acgt >= SMALL_FLOAT ) {
				assert( acgt < 1+SMALL_FLOAT );
				float scaling_factor = 1/acgt;
				
				(*f_iter)[ NUCL_A_PROF_IDX    ] *= scaling_factor;
				(*f_iter)[ NUCL_C_PROF_IDX    ] *= scaling_factor;
				(*f_iter)[ NUCL_G_PROF_IDX    ] *= scaling_factor;
				(*f_iter)[ NUCL_T_PROF_IDX    ] *= scaling_factor;
				(*f_iter)[ UNDEFINED_PROF_IDX ] *= scaling_factor;
			}
		}
	}
	
	_valid = true;
} 


void Profile::initialize_from_two_profiles(const Profile &px, const Profile &py, 
                                           float wx, float wy,
                                           const Vec<char> &alignment,
                                           bool normalize_bp_freqs) {

	assert( px.get_num_seqs() >= 1 && py.get_num_seqs() >= 1);
	
	_num_seqs = px.get_num_seqs() + py.get_num_seqs();
	
	assert(wx > 0 && wy > 0);
	assert(alignment.size() >= px.get_size() &&
	       alignment.size() >= py.get_size() &&
	       px.get_size() >= 2 && py.get_size() >= 2 );

	Vec<char>::const_iterator   a_iter = alignment.begin();
	Vec<char>::const_iterator   a_end  = alignment.end();

	unsigned int x_cnt=0, y_cnt=0;
	for (; a_iter < a_end; ++a_iter)
		switch (*a_iter) {
		case 'X': ++x_cnt; break;
		case 'Y': ++y_cnt; break;
		case 'B': ++x_cnt; ++y_cnt; break;
		default: 
			cout << "invalid *a_iter:" << *a_iter << " at position " << (int) (a_iter - alignment.begin()) << endl;
			assert(false);
		}
	cout << "x_cnt, y_cnt, pxsz, pysz: " << x_cnt << " " << y_cnt
	     << " " << px.get_size() << " " << py.get_size() << endl;

	assert( x_cnt == px.get_size());
	assert( y_cnt == py.get_size() );

	_freqs.resize( alignment.size() + 1, ProfileColumn() );

	const_iterator x_iter = px.get_begin();
	const_iterator y_iter = py.get_begin();
	const_iterator x_end  = px.get_end();
	const_iterator y_end  = py.get_end();

	// The following four columns are supposed gap columns before the beginning and end of x and y:
	ProfileColumn before_x_column, after_x_column, before_y_column, after_y_column, allgapextend_column;

	before_x_column.set_to(0, 0, 0, 0, 
	                       (*x_iter)[ GAPS_TOTAL_PROF_IDX ],          // gap extend
	                       0, (1 - (*x_iter)[ GAPS_TOTAL_PROF_IDX ]), // gap close
	                       0, 1,                                      // gaps total
	                       0);

	before_y_column.set_to(0, 0, 0, 0,
	                       (*y_iter)[ GAPS_TOTAL_PROF_IDX ],          // gap extend
	                       0, (1 - (*y_iter)[ GAPS_TOTAL_PROF_IDX ]), // gap close
	                       0, 1,                                      // gaps total
	                       0);
												 
	after_x_column.set_to(0, 0, 0, 0,
	                      (*(x_end-1))[ GAPS_TOTAL_PROF_IDX ],        // gap extend
	                      (1 - (*(x_end-1))[ GAPS_TOTAL_PROF_IDX ]),  // gap open
	                      0, 0, 1,                                    // gaps total
	                      0);

	after_y_column.set_to(0, 0, 0, 0,
	                      (*(y_end-1))[ GAPS_TOTAL_PROF_IDX ],        // gap extend
	                      (1 - (*(y_end-1))[ GAPS_TOTAL_PROF_IDX ]),  // gap open
	                      0, 0, 1,                                    // gaps total
	                      0);

	// The allgapextend column is padded to the end of each subprofile,
	//   touching the gap column that touches the first (or last) column
	//   of the subprofile.
	allgapextend_column.set_to(0, 0, 0, 0, 1, 0, 0, 0, 0, 0);         


	a_iter = alignment.begin();

	char c_prev, c_curr = *a_iter, c_next = *(a_iter+1);
	
	float a_  = 0, c_  = 0, g_  = 0, t_  = 0, acgt_ = 0;
	float go_ = 0, ge_ = 0, gc_ = 0, gs_ = 0, gt_ = 0, undef_ = 0;
	
	float a_xc, c_xc, g_xc, t_xc;
	float a_yc, c_yc, g_yc, t_yc;

	float ge_xp, ge_xc, ge_xn, go_xp, go_xc, go_xn, gc_xp, gc_xc, gc_xn, gs_xp, gs_xc, gs_xn;
	float ge_yp, ge_yc, ge_yn, go_yp, go_yc, go_yn, gc_yp, gc_yc, gc_yn, gs_yp, gs_yc, gs_yn;

	float gt_xp, gt_xc, gt_xn;
	float gt_yp, gt_yc, gt_yn;

	iterator f_iter = get_begin();

	bool  touched_first_xcol = false, touched_first_ycol = false;

	for ( a_iter = alignment.begin();
	      a_iter < a_end;
	      ++a_iter, ++f_iter ) {

		c_prev = (a_iter > alignment.begin()) ? *(a_iter-1) : 0;
		c_curr = *a_iter;
		c_next = (a_iter < a_end - 1)         ? *(a_iter+1) : 0;

		const ProfileColumn &xcol_prev = 
			( !touched_first_xcol && *a_iter == 'Y' ) ? 
			( allgapextend_column ) :
			( ( x_iter > px.get_begin() ) ? *(x_iter-1) : before_x_column );
		
		const ProfileColumn &xcol_curr = 
			( !touched_first_xcol && *a_iter == 'Y' ) ? 
			( ( c_next == 'Y' )  ? allgapextend_column : before_x_column ) :
			( ( x_iter < x_end ) ? *x_iter             : after_x_column );
		
		const ProfileColumn &xcol_next = 
			( !touched_first_xcol && *a_iter == 'Y' ) ? 

			( ( c_next == 'Y' ) ? before_x_column : *x_iter ) :

			( ( c_curr == 'Y' ) ? 
				( ( x_iter < x_end    ) ? *x_iter     : after_x_column)  :  
				( ( x_iter < x_end - 1) ? *(x_iter+1) : after_x_column ) );
		

		const ProfileColumn &ycol_prev = 
			( !touched_first_ycol && *a_iter == 'X' ) ? 
			( allgapextend_column ) :
			( ( y_iter > py.get_begin() ) ? *(y_iter-1) : before_y_column );

		const ProfileColumn &ycol_curr = 
			( !touched_first_ycol && *a_iter == 'X' ) ? 
			( ( c_next == 'X' )  ? allgapextend_column : before_y_column ) :
			( ( y_iter < y_end ) ? *y_iter             : after_y_column );
		
		const ProfileColumn &ycol_next =
			( !touched_first_ycol && *a_iter == 'X' ) ?

			( ( c_next == 'X' ) ? before_y_column : *y_iter ) :
			
			( ( c_curr == 'X' ) ? 
				( ( y_iter < y_end    ) ? *y_iter     : after_y_column ) :
				( ( y_iter < y_end - 1) ? *(y_iter+1) : after_y_column ) );
				

		// xp: x previous column; xc: x current column; xn: x next column

		a_xc = xcol_curr[ NUCL_A_PROF_IDX ];
		c_xc = xcol_curr[ NUCL_C_PROF_IDX ];
		g_xc = xcol_curr[ NUCL_G_PROF_IDX ];
		t_xc = xcol_curr[ NUCL_T_PROF_IDX ];
		
		a_yc = ycol_curr[ NUCL_A_PROF_IDX ];
		c_yc = ycol_curr[ NUCL_C_PROF_IDX ];
		g_yc = ycol_curr[ NUCL_G_PROF_IDX ];
		t_yc = ycol_curr[ NUCL_T_PROF_IDX ];
		
		// go: gap open; ge: gap extend;
		// gc: gap close; gs: gap single; gt: total gap = go + ge + gc + gs.
		
		ge_xp = xcol_prev[ GAP_EXTEND_PROF_IDX ];
		ge_xc = xcol_curr[ GAP_EXTEND_PROF_IDX ];
		ge_xn = xcol_next[ GAP_EXTEND_PROF_IDX ];

		go_xp = xcol_prev[ GAP_OPEN_PROF_IDX   ];
		go_xc = xcol_curr[ GAP_OPEN_PROF_IDX   ];
		go_xn = xcol_next[ GAP_OPEN_PROF_IDX   ];

		gc_xp = xcol_prev[ GAP_CLOSE_PROF_IDX  ];
		gc_xc = xcol_curr[ GAP_CLOSE_PROF_IDX  ];
		gc_xn = xcol_next[ GAP_CLOSE_PROF_IDX  ];

 		gs_xp = xcol_prev[ GAP_SINGLE_PROF_IDX ];
		gs_xc = xcol_curr[ GAP_SINGLE_PROF_IDX ];
		gs_xn = xcol_next[ GAP_SINGLE_PROF_IDX ];
		
		gt_xp = xcol_prev[ GAPS_TOTAL_PROF_IDX ];
		gt_xc = xcol_curr[ GAPS_TOTAL_PROF_IDX ];
		gt_xn = xcol_next[ GAPS_TOTAL_PROF_IDX ];
		

		ge_yp = ycol_prev[ GAP_EXTEND_PROF_IDX ];
		ge_yc = ycol_curr[ GAP_EXTEND_PROF_IDX ];
		ge_yn = ycol_next[ GAP_EXTEND_PROF_IDX ];

		go_yp = ycol_prev[ GAP_OPEN_PROF_IDX   ];
		go_yc = ycol_curr[ GAP_OPEN_PROF_IDX   ];
		go_yn = ycol_next[ GAP_OPEN_PROF_IDX   ];

		gc_yp = ycol_prev[ GAP_CLOSE_PROF_IDX  ];
		gc_yc = ycol_curr[ GAP_CLOSE_PROF_IDX  ];
		gc_yn = ycol_next[ GAP_CLOSE_PROF_IDX  ];

		gs_yp = ycol_prev[ GAP_SINGLE_PROF_IDX ];
		gs_yc = ycol_curr[ GAP_SINGLE_PROF_IDX ];
		gs_yn = ycol_next[ GAP_SINGLE_PROF_IDX ];
		
		gt_yp = ycol_prev[ GAPS_TOTAL_PROF_IDX ];
		gt_yc = ycol_curr[ GAPS_TOTAL_PROF_IDX ];
		gt_yn = ycol_next[ GAPS_TOTAL_PROF_IDX ];
		

		a_  = 0, c_  = 0, g_  = 0, t_  = 0, acgt_ = 0, undef_ = 0;
		go_ = 0, ge_ = 0, gc_ = 0, gs_ = 0, gt_ = 0;

		
		// First we set nucleotide freqs (this is easy)

		if ( c_curr != 'Y' ) {

			touched_first_xcol = true;
			
			a_ += a_xc * wx;
			c_ += c_xc * wx;
			g_ += g_xc * wx;
			t_ += t_xc * wx;

			undef_ += xcol_curr[ UNDEFINED_PROF_IDX ] * wx;
		}
		if ( c_curr != 'X' ) {

			touched_first_ycol = true;

			a_ += a_yc * wy;
			c_ += c_yc * wy;
			g_ += g_yc * wy;
			t_ += t_yc * wy;

			undef_ += ycol_curr[ UNDEFINED_PROF_IDX ] * wy;
		}

		acgt_ = a_ + c_ + g_ + t_;

		assert( ( !c_prev  || c_prev == 'X' || c_prev == 'Y' || c_prev == 'B') &&
		        (             c_curr == 'X' || c_curr == 'Y' || c_curr == 'B') &&
		        ( (!c_next || c_next == 'X' || c_next == 'Y' || c_next == 'B') ) );

		// Then, we set the gap contributions of X
		//
		// Denote by 0: a gapped column of X, 1: a present column of X. 
		// We distinguish between 0 cases, depending on whether the prev, current, and next
		//   columns are gapped.
		// 
		// 1. Case 0 0 0
		//
		//    Then, all the weight of X is allocated to gap extension in this column

		if      ( (!c_prev || c_prev == 'Y') && c_curr == 'Y' && (!c_next || c_next == 'Y') ) {
			ge_ += wx;
			gt_ += wx;
		}

		//
		// 2. Case 0 0 1
		//    
		//    Then, every letter in next column is a gap close in this column, and
		//      every gap in next column is a gap extension in this column.

		else if ( (!c_prev || c_prev == 'Y') && c_curr == 'Y' && c_next != 'Y' ) {
			ge_ += gt_xn * wx;
			gc_ += ( 1 - gt_xn ) * wx;
			gt_ += wx;
		}

		//
		// 3. Case 0 1 0
		//    
		//    Then, every gap in X's current column, is a gap extension in the combined profile.
		
		else if ( (!c_prev || c_prev == 'Y') && c_curr != 'Y' && (!c_next || c_next == 'Y') ) {
			ge_ += gt_xc * wx;
			gt_ += gt_xc * wx;
		}
		
		//
		// 4. Case 0 1 1
		//    
		//    Then, gap closes and gap singles in current X's column become all gap closes;
		//      gap opens and gap extends become gap extends.

		else if ( (!c_prev || c_prev == 'Y') && c_curr != 'Y' && c_next != 'Y' ) {
			ge_ += (go_xc + ge_xc) * wx;
			gc_ += (gc_xc + gs_xc) * wx;
			gt_ += (go_xc + ge_xc + gc_xc + gs_xc) * wx;
		}
						
		//
		// 5. Case 1 0 0
		//    
		//    Then, every letter in X's prev column becomes a gap open; every
		//      gap in X's prev column becomes a gap extension

		else if ( c_prev != 'Y' && c_curr == 'Y' && (!c_next || c_next == 'Y') ) {
			go_ += ( 1 - gt_xp ) * wx;
			ge_ += gt_xp * wx;
			gt_ += wx;
		}
						
		//
		// 6. Case 1 0 1
		//    
		//    This is a rather complicated case. Gap singles are estimated to be
		//      the minimum possible (no less than 0). (1 - gt_xp) + (1 - gt_xn) is
		//      the total number of letters in the prev and next column. Those
		//      letters have to become gap opens, and gap closes, respectively.
		//      However, there cannot be more gap opens + closes than number of
		//      total positions; hence, some of them will be gap singles.
		//    
		//    After gap singles are estimated, gap opens are all letters in prev
		//      column minus gap singles, and similarly for gap closes. Finally,
		//      the rest of the column becomes gap extensions.

		else if ( c_prev != 'Y' && c_curr == 'Y' && c_next != 'Y' ) {

			float go_temp = (go_xn + gs_xn) * wx;
			float gc_temp = (gc_xp + gs_xp) * wx;
			float ge_temp = (ge_xp + go_xp) * wx;
			float gs_temp = wx - (go_temp + gc_temp + ge_temp);
			
			assert( fabsf( (ge_xp + go_xp) - (ge_xn + gc_xn) ) < SMALL_FLOAT );

			ge_ += ge_temp;
			go_ += go_temp;
			gc_ += gc_temp;
			gs_ += gs_temp;
			
			gt_ += wx;
		}

		//
		// 7. Case 1 1 0
		//    
		//    This case is symmetrical to 0 1 1, with opens and closes reversed.

		else if ( c_prev != 'Y' && c_curr != 'Y' && (!c_next || c_next == 'Y') ) {
			ge_ += (gc_xc + ge_xc) * wx;
			go_ += (go_xc + gs_xc) * wx;
			gt_ += (go_xc + ge_xc + gc_xc + gs_xc) * wx;
		}

		//
		// 8. Case 1 1 1
		//
		//    In this case, X's gaps are exactly carried to the combined profile.

		else if ( c_prev != 'Y' && c_curr != 'Y' && c_next != 'Y' ) {
			ge_ += ge_xc * wx;
			go_ += go_xc * wx;
			gc_ += gc_xc * wx;
			gs_ += gs_xc * wx;
			gt_ += gt_xc * wx;
		}		

		//
		// 
		// Then, we set the gap contributions of Y, in a similar manner
		//
		//
		if      ( (!c_prev || c_prev == 'X') && c_curr == 'X' && (!c_next || c_next == 'X') ) {
			ge_ += wy;
			gt_ += wy;
		}
		else if ( (!c_prev || c_prev == 'X') && c_curr == 'X' && c_next != 'X' ) {
			ge_ += gt_yn * wy;
			gc_ += ( 1 - gt_yn ) * wy;
			gt_ += wy;
		}
		else if ( (!c_prev || c_prev == 'X') && c_curr != 'X' && (!c_next || c_next == 'X') ) {
			ge_ += gt_yc * wy;
			gt_ += gt_yc * wy;
		}
		else if ( (!c_prev || c_prev == 'X') && c_curr != 'X' && c_next != 'X' ) {
			ge_ += (go_yc + ge_yc) * wy;
			gc_ += (gc_yc + gs_yc) * wy;
			gt_ += (go_yc + ge_yc + gc_yc + gs_yc) * wy;
		}
		else if ( c_prev != 'X' && c_curr == 'X' && (!c_next || c_next == 'X') ) {
			go_ += ( 1 - gt_yp ) * wy;
			ge_ += gt_yp * wy;
			gt_ += wy;
		}
		else if ( c_prev != 'X' && c_curr == 'X' && c_next != 'X' ) {

			float go_temp = (go_yn + gs_yn) * wy;
			float gc_temp = (gc_yp + gs_yp) * wy;
			float ge_temp = (ge_yp + go_yp) * wy;
			float gs_temp = wy - (go_temp + gc_temp + ge_temp);

			assert( fabsf( (ge_yp + go_yp) - (ge_yn + gc_yn) ) < SMALL_FLOAT );

			ge_ += ge_temp;
			go_ += go_temp;
			gc_ += gc_temp;
			gs_ += gs_temp;
			
			gt_ += wy;
		}
		else if ( c_prev != 'X' && c_curr != 'X' && (!c_next || c_next == 'X') ) {
			ge_ += (gc_yc + ge_yc) * wy;
			go_ += (go_yc + gs_yc) * wy;
			gt_ += (go_yc + ge_yc + gc_yc + gs_yc) * wy;
		}
		else if ( c_prev != 'X' && c_curr != 'X' && c_next != 'X' ) {
			ge_ += ge_yc * wy;
			go_ += go_yc * wy;
			gc_ += gc_yc * wy;
			gs_ += gs_yc * wy;
			gt_ += gt_yc * wy;
		}
			
		if (normalize_bp_freqs) {
			// Frequencies of nucleotides are normalized to sum up to 1:
			if ( acgt_ < SMALL_FLOAT )
				f_iter->set_to(0,0,0,0, ge_, go_, gc_, gs_, gt_, undef_);
			
			else 
				f_iter->set_to(a_/acgt_, c_/acgt_, g_/acgt_, t_/acgt_, 
				               ge_, go_, gc_, gs_, gt_, undef_);
		}
		else
			f_iter->set_to(a_, c_, g_, t_, 
			               ge_, go_, gc_, gs_, gt_, undef_);

		if (c_curr != 'Y') ++x_iter;
		if (c_curr != 'X') ++y_iter;
	}

	_valid = true;
}


ostream& operator<<(ostream &o, const ProfileColumn &c) {
	assert ( c._freqs.size() == PROFILE_COLUMN_SIZE );
	cout << setiosflags(ios::fixed | ios::showpoint) << setprecision(2);
	
	for (ProfileColumn::size_type i = 0; i < c._freqs.size(); ++i)
		if (c._freqs[i] < SMALL_FLOAT) o << " .   ";
		else o << c._freqs[i] << " ";
	
	return o;
}

ostream& operator<<(ostream &o, const Profile &p) {
	assert( p.is_valid() );
	
	o << "    A    C    G    T    gt   N    go   gc   ge   gs" << endl;
	Profile::const_iterator p_iter = p.get_begin();
	Profile::const_iterator p_end  = p.get_end();
	
	for (unsigned int currpos = 1; p_iter < p_end; ++p_iter, ++currpos) {
		o << currpos << ": " << *p_iter << "  : ";
		p_iter->print_consensus(o);
		o << endl;
	}	
	
	return o;
}


unsigned int CHAR_PROFILE_IDX[ 256 ];
char         IDX_PROFILE_CHAR[ PROFILE_COLUMN_SIZE ];

void initialize_prof_idx() {
	for (unsigned int i = 0; i < 256; ++i)
		CHAR_PROFILE_IDX[i] = UNDEFINED_PROF_IDX;
	
	CHAR_PROFILE_IDX[ (int)'a' ] = CHAR_PROFILE_IDX[ (int)'A' ] = NUCL_A_PROF_IDX;
	CHAR_PROFILE_IDX[ (int)'c' ] = CHAR_PROFILE_IDX[ (int)'C' ] = NUCL_C_PROF_IDX;
	CHAR_PROFILE_IDX[ (int)'g' ] = CHAR_PROFILE_IDX[ (int)'G' ] = NUCL_G_PROF_IDX;
	CHAR_PROFILE_IDX[ (int)'t' ] = CHAR_PROFILE_IDX[ (int)'T' ] = NUCL_T_PROF_IDX;
	CHAR_PROFILE_IDX[ (int)'-' ] = GAPS_TOTAL_PROF_IDX;
	
	for (unsigned int i = 0; i < PROFILE_COLUMN_SIZE; ++i)
		IDX_PROFILE_CHAR[ i ] = '?';
	
	IDX_PROFILE_CHAR[ NUCL_A_PROF_IDX     ] = 'A';
	IDX_PROFILE_CHAR[ NUCL_C_PROF_IDX     ] = 'C';
	IDX_PROFILE_CHAR[ NUCL_G_PROF_IDX     ] = 'G';
	IDX_PROFILE_CHAR[ NUCL_T_PROF_IDX     ] = 'T';
	IDX_PROFILE_CHAR[ GAP_EXTEND_PROF_IDX ] = '-';
	IDX_PROFILE_CHAR[ GAP_OPEN_PROF_IDX   ] = '-';
	IDX_PROFILE_CHAR[ GAP_CLOSE_PROF_IDX  ] = '-';
	IDX_PROFILE_CHAR[ GAP_SINGLE_PROF_IDX ] = '-';
	IDX_PROFILE_CHAR[ GAPS_TOTAL_PROF_IDX ] = '-';
	IDX_PROFILE_CHAR[ UNDEFINED_PROF_IDX  ] = 'N';
}
 
