#ifndef PROFILE_H
#define PROFILE_H

//
// The Profile class holds a profile (currently of columns floats, that hold
//   eight frequencies: A, C, G, T, gap extend, gap open, gap close, unknown
//

#include <iomanip>
#include <cassert>
#include <iostream>
#include <algorithm>
#include "Vec.h"
#include "Seq.h"
#include "Mseq.h"
#include "Constants.h"

using namespace std;
using namespace BIO;


const  unsigned int PROFILE_COLUMN_SIZE   = 10;

enum ColumnIndex {
	NUCL_A_PROF_IDX       = 0,
	NUCL_C_PROF_IDX       = 1,
	NUCL_G_PROF_IDX       = 2,
	NUCL_T_PROF_IDX       = 3,
	GAPS_TOTAL_PROF_IDX   = 4, 
	UNDEFINED_PROF_IDX    = 5, 
	GAP_OPEN_PROF_IDX     = 6,
	GAP_CLOSE_PROF_IDX    = 7, // Used in SOPScore with a position displacement
	GAP_EXTEND_PROF_IDX   = 8, // Not currently used in SOPScore
	GAP_SINGLE_PROF_IDX   = 9  // Used in SOPScore as both open and close
};


extern unsigned int CHAR_PROFILE_IDX[];
extern char         IDX_PROFILE_CHAR[];

void  initialize_prof_idx();

class ProfileColumn {

 private:

	Vec< float > _freqs;

	typedef Vec< float >::iterator       iterator;
	typedef Vec< float >::const_iterator const_iterator;

 public:

	typedef Vec< float >::size_type size_type;
	

	ProfileColumn() : _freqs( PROFILE_COLUMN_SIZE, 0) {}

	void   set_to(float a_, float c_, float g_, float t_,
	              float ge_, float go_, float gc_, float gs_,
	              float gt_, float undef_);

	float  get_freq_of(char c) const { return _freqs[ CHAR_PROFILE_IDX[ (int) c ] ]; }

	float& operator[] (int  i)       { return _freqs[i]; }
	float  operator[] (int  i) const { return _freqs[i]; }

	ostream& print_consensus(ostream &o) const;


	friend ostream &operator<<(ostream &o, const ProfileColumn &c);
};


// This very simple function, returns the number of pairs of letters that match.
//   For example, if column c1 = { A = 0.5, C = 2, G = 0, T = 0, UNDEFINED = 3, gaps: anything }
//                          c2 = { A = 1.2, C = 1, G = 3, T = 0, UNDEFINED = 1, gaps: anything }
//   Then, the function returns 0.5*1.2 + 2*1 = 2.6.
float match_freqs_of_column_pair( const ProfileColumn &c1, const ProfileColumn &c2 );


// A Profile holds the frequencies of letters and gaps in a multiple alignment.
//
//   Restriction: the length of a profile has to be >= 2.
//
//
//   There are three ways to initialize a profile:
// 
//   1. Initialize from a Seq. This creates a simple profile of all the letters
//      in the seq, with no gaps.
//
//   2. Initialize from a Mseq.
//
//   3. Initialize from two profiles. Given two profiles, and a weight associated
//      with each (e.g., the two weights could sum up to 1), a combined profile
//      is constructed.
//

class Profile {

 private:

	Vec< ProfileColumn > _freqs;    // frequencies of letters and gaps in each column
	unsigned int  _num_seqs; // number of sequences represented in the profile
	bool          _valid;    // is the profile initialized?

 public:

	typedef Vec< ProfileColumn >::size_type       size_type;
	typedef Vec< ProfileColumn >::iterator        iterator;
	typedef Vec< ProfileColumn >::const_iterator  const_iterator;
	



	Profile() : _num_seqs(0), _valid(false) { _freqs.resize(0); }

	//
	// Initializes a profile of 8 weighted frequencies, from an mseq object
	// Note that gap open, gap extend, gap single, and gap close are distinguished.
	// Their sum is equal to GAPS_TOTAL.
	// 
	// If begin_pos > 1 then the position begin_pos - 1 is examined to determine gap open/single
	//   at begin_pos. 
	//
 	void initialize_from_mseq(const Mseq &m, unsigned int begin_pos, unsigned int end_pos,
	                          const Vec<float> &weights, bool normalize_bp_freqs = false); 

	//
	// Initializes a profile from a single sequence, which has to be
	//   at least 2-long.
	void initialize_from_seq(const Seq &s, Seq::size_type begin_pos, Seq::size_type end_pos,
	                         bool normalize_bp_freqs = false);

	//
	// Initializes a profile from two smaller profiles, and an alignment.
	// 
	// The alignment has to be the exact length required in order to globally align
	//   the two profiles.
	// 
	void initialize_from_two_profiles(const Profile &px, const Profile &py, 
	                                  float wx, float wy, 
	                                  const Vec<char> &alignment,
	                                  bool normalize_bp_freqs = false);


	bool           is_valid()     const { return _valid;             }
	size_type      get_size()     const { return _freqs.size() - 1;  }
	unsigned int   get_num_seqs() const { return _num_seqs;          }
	const_iterator get_begin()    const { return _freqs.begin() + 1; } // like Seq, Profile is 1-indexed
	const_iterator get_end()      const { return _freqs.end();       }

	iterator       get_begin()          { return _freqs.begin() + 1; } // like Seq, Profile is 1-indexed
	iterator       get_end()            { return _freqs.end();       }	

	ProfileColumn &      operator[](Profile::size_type i)       { return _freqs[i]; }
	const ProfileColumn &operator[](Profile::size_type i) const { return _freqs[i]; }

	const_iterator get_iterator(size_type index) const { return _freqs.begin() + index; }
	iterator       get_iterator(size_type index)       { return _freqs.begin() + index; }

	friend ostream& operator<<(ostream &o, const Profile &p);
};


ostream& operator<<(ostream &o, const ProfileColumn  &c);
ostream& operator<<(ostream &o, const Profile &p);


#endif
