#ifndef SCORINGPARAMETERS_H
#define SCORINGPARAMETERS_H

#include "Vec.h"
#include "Constants.h"
#include <math.h>
#include <fstream>
#include <iostream>
#include <sstream>

using namespace std;

/* Some default values */
#define SP_DEF_PBPWindowSize  5
#define SP_DEF_PBPFunction    PBP_MIDDLE_OF_BS
#define SP_DEF_ScoringModel   SOP_OPTIMISTIC

template <class ScoreType>
class ScoringParameters
{
	/* Public part first, for type declaration */

	public:

	typedef Vec< Vec<ScoreType> > ScoringMatrix;
	
	typedef enum
	{
		PBP_MIDDLE_OF_BS,
		PBP_WELL_MATCHED_BCOL
	} PBPFuncType;

	typedef enum
	{
		SOP_OPTIMISTIC,
		SOP_HISTORY
	} SModel;

	/* This is what parameters file should look like:
	 *
	 * #          A    C    G    T   -
	 * SubstitutionMatrix
	 *           91 -114  -31 -123 -25
	 *         -114  100 -125  -31 -25
	 *          -31 -125  100 -114 -25
	 *         -123  -31 -114   91 -25
	 *          -25  -25  -25  -25   0
	 * GapOpen -750
	 * PBPFunction MiddleOfBs
	 * # or the following:
	 * # PBPFunction WellMatchedBCol
	 * PBPWindowSize 7
	 * ScoringModel OptimisticSOP
	 * # or the following:
	 * # ScoringModel HistorySOP
	 *
	 */
	ScoringParameters()
	{
		_init();
		_g_valid = _S_valid = _ftype_valid = _wsize_valid = _smodel_valid = false;
	}
	
	ScoringParameters(string fname)
	{
		_init();
		read_from_file(fname);
	}

	void read_from_file(string fname)
	{
		/* Reset everything to false */
		_g_valid = _S_valid = _ftype_valid = _wsize_valid = _smodel_valid = false;

		ifstream is(fname.c_str());

		if (!is)
			throw BIO::E_OpenFailedEx();

		read_from_stream(is);

		if (!(is.rdbuf()->close()))
			throw BIO::E_CloseFailedEx();

	}

	void read_from_string(string s)
	{
		istringstream is(s);
		read_from_stream(is);
	}
		
	void read_from_stream(istream& is)
	{
		string str;
		
		/* Read a word (consecutive non-whitespace characters) at a time */
		while (is >> str)
		{
			/* TODO: Make things case-insensitive */
			if (str == "GapOpen")
			{
				if (_g_valid)
					throw BIO::E_SPFileParseErrorEx("Duplicate definition for GapOpen");

				if (is >> _g)
					_g_valid = true;
				else
					throw BIO::E_SPFileParseErrorEx("Invalid GapOpen parameter");
			}
			else if (str == "PBPFunction")
			{
				string fnc;
				
				if (_ftype_valid)
					throw BIO::E_SPFileParseErrorEx("Duplicate definition for PBPFunction");

				if (is >> fnc)
				{
					if (fnc == "MiddleOfBs")
					{
						_ftype = PBP_MIDDLE_OF_BS;
						_ftype_valid = true;
					}
					else if (fnc == "WellMatchedBCol")
					{
						_ftype = PBP_WELL_MATCHED_BCOL;
						_ftype_valid = true;
					}
					else
						throw BIO::E_SPFileParseErrorEx("Invalid PBPFunction '" + fnc + "'");
				}
				else
					throw BIO::E_SPFileParseErrorEx("Invalid PBPFunction parameter");
			}
			else if (str == "ScoringModel")
			{
				string sm;
				
				if (_smodel_valid)
					throw BIO::E_SPFileParseErrorEx("Duplicate definition for ScoringModel");

				if (is >> sm)
				{
					if (sm == "OptimisticSOP")
					{
						_smodel = SOP_OPTIMISTIC;
						_smodel_valid = true;
					}
					else if (sm == "HistorySOP")
					{
						_smodel = SOP_HISTORY;
						_smodel_valid = true;
					}
					else
						throw BIO::E_SPFileParseErrorEx("Invalid ScoringModel '" + sm + "'");
				}
				else
					throw BIO::E_SPFileParseErrorEx("Invalid ScoringModel parameter");
			}
			else if (str == "PBPWindowSize")
			{
				if (_wsize_valid)
					throw BIO::E_SPFileParseErrorEx("Duplicate definition for PBPWindowSize");

				if (is >> _wsize)
					_wsize_valid = true;
				else
					throw BIO::E_SPFileParseErrorEx("Invalid PBPWindowSize parameter");
			}
			else if (str == "SubstitutionMatrix")
			{
				if (_S_valid)
					throw BIO::E_SPFileParseErrorEx("Duplicate definition for SubstitutionMatrix");

				for (unsigned int i = 0; i < BIO::NUM_ALPHABET; ++i)
					for (unsigned int j = 0; j < BIO::NUM_ALPHABET; ++j)
						if (!(is >> _S[i][j]))
							throw BIO::E_SPFileParseErrorEx("Invalid SubstitutionMatrix parameter");

				_S_valid = true;
			}
			else if (str[0] == '#')
			{
				stringbuf tmp;
				is.get(tmp);
			}
			else
			{
				throw BIO::E_SPFileParseErrorEx("Invalid parameter '" + str + "'");
			}
		}

	}

	ScoreType get_gap_open()
	{
		return _g;
	}

	bool has_gap_open()
	{
		return _g_valid;
	}

	ScoringMatrix &get_substitution_matrix()
	{
		return _S;
	}

	bool has_substitution_matrix()
	{
		return _S_valid;
	}

	PBPFuncType get_pick_best_point_function()
	{
		return _ftype;
	}

	bool has_pick_best_point_function()
	{
		return _ftype_valid;
	}

	SModel get_scoring_model()
	{
		return _smodel;
	}

	bool has_scoring_model()
	{
		return _smodel_valid;
	}

	unsigned int get_pbpf_window_size()
	{
		return _wsize;
	}

	bool has_pbpf_window_size()
	{
		return _wsize_valid;
	}

	double getLambda(Mseq::size_type size1 = 1, Mseq::size_type size2 = 1)
	{
		if (!_karlinValid)
		{
			_getKarlin(size1, size2);
		}
		return _lambda;
	}
	
	double getK(Mseq::size_type size1 = 1, Mseq::size_type size2 = 1)
	{
		if (!_karlinValid)
		{
			_getKarlin(size1, size2);
		}
		return _K;
	}

	double getH(Mseq::size_type size1 = 1, Mseq::size_type size2 = 1)
	{
		if (!_karlinValid)
		{
			_getKarlin(size1, size2);
		}
		return _H;
	}

	/* Private Part */

	private:
	
	ScoreType _g;
	bool _g_valid;
	
	ScoringMatrix _S;
	bool _S_valid;

	PBPFuncType _ftype;
	bool _ftype_valid;

	unsigned int _wsize;
	bool _wsize_valid;

	SModel _smodel;
	bool _smodel_valid;

	double _lambda;
	double _K;
	double _H;
	double _karlinValid;

	void _getKarlin(Mseq::size_type size1, Mseq::size_type size2)
	{
		Vec<ScoreType> scores;
		Vec<unsigned int> nucCounts;
		calcComb(nucCounts, size1 > 1 ? BIO::NUM_ALPHABET - 1 : BIO::NUM_ALPHABET - 2, size1, size2, scores);

		ScoreType low = 0;
		ScoreType high = 0;

		for (typename Vec<ScoreType>::iterator it = scores.begin(); it != scores.end(); ++it)
		{
			if (*it < low)
			{
				low = *it;
			}
			else if (*it > high)
			{
				high = *it;
			}
		}

		unsigned int range = high - low;
		double pr[range + 1];
		for (unsigned int i = 0; i < range + 1; i++)
		{
			pr[i] = 0;
		}

		for (typename Vec<ScoreType>::iterator it = scores.begin(); it != scores.end(); ++it)
		{
			pr[*it - low]++;
		}			

		for (unsigned int i = 0; i < range + 1; i++)
		{
			pr[i] /= scores.size();
		}

		_karlin(low, high, pr, &_lambda, &_K, &_H);
		_karlinValid = true;
	}

	//the below two functions are for filling up our pr vector

	void calcComb(Vec<unsigned int>& nucCounts, Vec<unsigned int>::size_type maxInd, Mseq::size_type size, Mseq::size_type otherSize, Vec<ScoreType>& scores)
	{
		if (size == 0 || maxInd == 0)
		{
			if (maxInd == 0)
			{
				nucCounts[maxInd] = size;
			}
			//we've now filled the first counts, now get the second counts
			Vec<unsigned int> otherNucCounts;
			calcComb(nucCounts, otherNucCounts, otherSize > 1 ? BIO::NUM_ALPHABET - 1 : BIO::NUM_ALPHABET - 2, otherSize, scores);
		}
		else
		{
			Mseq::size_type maxValue = (maxInd == BIO::NUM_ALPHABET - 1 ? size - 1 : size);
			for (Mseq::size_type i = 0; i <= maxValue; i++)
			{
				nucCounts.clear();
				nucCounts.resize(BIO::NUM_ALPHABET, 0);
				nucCounts[maxInd] = i;
				calcComb(nucCounts, maxInd - 1, size - i, otherSize, scores);
			}
		}
	}

	void calcComb(const Vec<unsigned int>& otherNucCounts, Vec<unsigned int>& nucCounts, Vec<unsigned int>::size_type maxInd, Mseq::size_type size, Vec<ScoreType>& scores)
	{
		if (size == 0 || maxInd == 0)
		{
			if (maxInd == 0)
			{
				nucCounts[maxInd] = size;
			}
			//we've now filled both counts, now get the score
			ScoreType score = 0;
			for (unsigned int i = 0; i < BIO::NUM_ALPHABET; ++i)
			{
				for (unsigned int j = 0; j < BIO::NUM_ALPHABET; ++j)
				{
					score += nucCounts[i]*otherNucCounts[j]*_S[i][j];
				}
			}
			scores.push_back(score);
			return;
		}
		else
		{
			//we don't allow all gaps
			Mseq::size_type maxValue = (maxInd == BIO::NUM_ALPHABET - 1 ? size - 1 : size);
			for (Mseq::size_type i = 0; i <= maxValue; i++)
			{
				nucCounts.clear();
				nucCounts.resize(BIO::NUM_ALPHABET, 0);
				nucCounts[maxInd] = i;
				calcComb(otherNucCounts, nucCounts, maxInd - 1, size - i, scores);
			}
		}
	}

	void _karlin(int low, int high, double* pr, double* lambda, double* K, double* H)
		 // int low;			/* Lowest score (must be negative)    */
		 //	int high;			/* Highest score (must be positive)   */
		 //	double *pr;			/* Probabilities for various scores   */
		 //	double *lambda;			/* Pointer to parameter lambda        */
		 //	double *K;			/* Pointer to parameter K             */
		 //	double *H;			/* Pointer to parameter H             */
	{
		const int MAXIT = 200;
		int i,j,range,lo,hi,first,last;
		double up,nw,sum,Sum,av,beta,oldsum,ratio,ftemp;
		double *p,*P,*ptrP,*ptr1,*ptr2;
		
		/* Check that scores and their associated probabilities are valid     */
		
		if (low>=0) {
			fprintf(stderr,"Lowest score must be negative.\n");
			return;
		}
		for (i=range=high-low;i> -low && !pr[i];--i);
		if (i<= -low) {
			fprintf(stderr,"A positive score must be possible.\n");
			return;
		}
		for (sum=i=0;i<=range;sum+=pr[i++]) if (pr[i]<0) {
			fprintf(stderr,"Negative probabilities not allowed.\n");
			return;
		}
		if (sum<0.99995 || sum>1.00005)
			fprintf(stderr,"Probabilities sum to %.4f.  Normalizing.\n",sum);
		p= (double *) calloc((unsigned)range+1,sizeof(double));
		for (Sum=low,i=0;i<=range;++i) Sum+=i*(p[i]=pr[i]/sum);
		if (Sum>=0)
	  {fprintf(stderr,"Invalid (non-negative) expected score:  %.3f\n",Sum);
		free((char *)p); return;
	  }
		
		/* Calculate the parameter lambda */
		
		up=0.5;
		do {
			up*=2;
			ptr1=p;
			beta=exp(up);
			ftemp=exp(up*(low-1));
			for (sum=i=0;i<=range;++i) sum+= *ptr1++ * (ftemp*=beta);
		}
		while (sum<1.0);
		for (*lambda=j=0;j<25;++j) {
			nw=(*lambda+up)/2.0;
			beta=exp(nw);
			ftemp=exp(nw*(low-1));
			ptr1=p;
			for (sum=i=0;i<=range;++i) sum+= *ptr1++ * (ftemp*=beta);
			if (sum>1.0) up=nw;
			else *lambda=nw;
		}
		
		/* Calculate the pamameter K */
		
		ptr1=p;
		ftemp=exp(*lambda*(low-1));
		for (av=0,i=low;i<=high;++i) av+= *ptr1++ *i*(ftemp*=beta);
		*H= *lambda*av/log(2.0);
		if (low == -1 || high == 1) {
			*K = high==1 ? av : Sum*Sum/av;
			*K *= 1.0 - exp(- *lambda);
			free((char *)p);
			return;	/* Parameters calculated successfully */
		}
		Sum=lo=hi=0;
		P= (double *) calloc((unsigned)MAXIT*range+1,sizeof(double));
		for (*P=sum=oldsum=j=1;j<=MAXIT && sum>0.001;Sum+=sum/=j++) {
			first=last=range;
			for (ptrP=P+(hi+=high)-(lo+=low);ptrP>=P;*ptrP-- =sum) {
				ptr1=ptrP-first;
				ptr2=p+first;
				for (sum=0,i=first;i<=last;++i) sum+= *ptr1-- * *ptr2++;
				if (first) --first;
				if (ptrP-P<=range) --last;
			}
			ftemp=exp(*lambda*(lo-1));
			for (sum=0,i=lo;i;++i) sum+= *++ptrP * (ftemp*=beta);
			for (;i<=hi;++i) sum+= *++ptrP;
			ratio=sum/oldsum;
			oldsum=sum;
		}
		if (j>MAXIT)
	  {fprintf(stderr,"\nhsseg.c: ");
		fprintf(stderr,
						"Value for K may be too large due to insufficient iterations.\n");
	  }
		for (;j<=MAXIT;Sum+=oldsum/j++) oldsum*=ratio;
		for (i=low;!p[i-low];++i);
		for (j= -i;i<high && j>1;) if (p[++i-low]) j=_gcd(j,i);
		*K = (j*exp(-2*Sum))/(av*(1.0-exp(- *lambda*j)));
		free((char *)p); free((char *)P);
		return;	/* Parameters calculated successfully */
	}
	
	int _gcd(int a,int b)
	{
		int c;
		
		if (b<0) b= -b;
		if (b>a) { c=a; a=b; b=c; }
		for (;b;b=c) { c=a%b; a=b; }
		return a;
	}

	void _init()
	{
		/* Some default values */
		_wsize = SP_DEF_PBPWindowSize;
		_ftype = SP_DEF_PBPFunction;
		_smodel = SP_DEF_ScoringModel;

		/* Resize S */
		_S.resize(BIO::NUM_ALPHABET);
		for (unsigned int i = 0; i < BIO::NUM_ALPHABET; ++i)
			_S[i].resize(BIO::NUM_ALPHABET);

		_karlinValid = false;
	}

};

#endif
