//
// SparseMatrix.h: Sparse matrix template.
// 
// Initialization:
// Two main options are (1) initialize by passing a matrix in the form Vec of Vec<T>,
//   which  contains all the data, including the elts below threshold, and (2) first
//   construct an empty matrix with the default constructor, and then initialize it
//   to the transpose of another sparse matrix.
//
//
// Iteration through the elements is done by use of two iterators, one for rows, and
//   one for cols of a row.
//
// Element (i, j) can be accessed by reference using GetValue, which performs binary
//   search.
//
// ostream<< has been defined, and makes a pretty-print that can be used for debugging
//   mostly, as it is not too efficient.
// 

#ifndef SPARSE_MATRIX_H
#define SPARSE_MATRIX_H

#include <iostream>
#include <string>
#include <vector>
#include <cassert>

#include "Vec.h"

using namespace std;

template<class T> struct less_than_first_int {
	bool operator() (const pair<unsigned int, T> &a, const pair<unsigned int, T> &b) {
		return a.first < b.first;
	}
};	


template<class T> 
class SparseMatrix {
 
 private:

	unsigned int                          _m, _n;     // a matrix of size mn
	Vec< Vec< pair<unsigned int, T> > >   _matrix;    // data in the matrix
	T                                     _threshold; // threshold used to filter out small elements
	
 public:
	
	SparseMatrix() : _m(0), _n(0), _threshold(0) { _matrix.resize(0); }
	
	SparseMatrix(Vec< Vec< T > > &data, T threshold);
		
	void initialize_to_transpose(SparseMatrix<T> &s); 	
	
	T get_value(unsigned int m_row, unsigned int n_col) const {
		
 		typename Vec< pair< unsigned int, T > >::const_iterator col_iter = lower_bound( _matrix[m_row].begin(),  
		                                                                                _matrix[m_row].end(),  
		                                                                                make_pair(n_col, _threshold), 
		                                                                                less_than_first_int<T>() );
		if (col_iter != _matrix[m_row].end() && 
					((*col_iter).first == n_col)) 
				return (*col_iter).second; 
			else  
				return _threshold; 
	}
	
	
	unsigned int get_nrows()     const { return _m;         }
	unsigned int get_ncols()     const { return _n;         }
	T            get_threshold() const { return _threshold; }

	typename Vec< Vec< pair< unsigned int, T > > >::const_iterator row_iter_begin()                   const { return _matrix.begin();        }
	typename Vec< Vec< pair< unsigned int, T > > >::const_iterator row_iter_end()                     const { return _matrix.end();          }
	typename Vec<      pair< unsigned int, T >   >::const_iterator col_iter_begin(unsigned int m_row) const { return _matrix[m_row].begin(); }
	typename Vec<      pair< unsigned int, T >   >::const_iterator col_iter_end  (unsigned int m_row) const { return _matrix[m_row].end();   }


	friend ostream& operator<< <T>(ostream &o, const SparseMatrix &s);	
};

template< class T > ostream& operator<<(ostream &o, const SparseMatrix<T> &s);

template< class T> SparseMatrix<T>::SparseMatrix(Vec< Vec< T > > &data, T threshold) : _threshold(threshold) {
	
	_m = data.size();
	assert( _m > 0 );
	_matrix.resize(_m);
	_n = data[0].size();
	
	for (unsigned int i = 0; i < _m; ++i) {
		
		assert(_n == data[i].size());
		_matrix[i].resize(0);
		
		for (unsigned int j = 0; j < _n; ++j) {
			if (data[i][j] > _threshold)
				_matrix[i].push_back(make_pair(j, data[i][j]));
		}
	}
}	

template< class T > void SparseMatrix<T>::initialize_to_transpose(SparseMatrix<T> &s) {
	
	assert(s._m > 0 && s._n > 0);
	
	_m         = s._n;
	_n         = s._m;
	_threshold = s._threshold;
	
	_matrix.resize(_m);
	
	for (unsigned int i = 0; i < _n; ++i) {
		
		for (typename Vec< pair< unsigned int, T > >::const_iterator row_iter = s._matrix[i].begin();
		     row_iter < s._matrix[i].end();
		     ++row_iter)
			_matrix[row_iter->first].push_back(make_pair(i, row_iter->second));
	}
}	


template< class T > ostream& operator<<(ostream &o, const SparseMatrix<T> &s) {
	o << s.get_nrows() << " " << s.get_ncols() << " " << s.get_threshold() << endl;
	for (unsigned int i = 0; i <s.get_nrows(); ++i) {
		o << i << " [ ";
		for (typename Vec< pair< unsigned int, T > >::const_iterator col_iter = s.col_iter_begin(i);
		     col_iter < s.col_iter_end(i);
		     ++col_iter)
			o <<  col_iter->first << " : " << col_iter->second << "   ";
		
		o << "]" << endl;
	}
	return o;
}


#endif
