#include "HeuristicIndex.h"

void HeuristicIndex::_initializeScoreCache(Mseq& sequence, Vec<score_type>& scoreCache, ScoringParameters<score_type>& scoringParameters) const
{
	Scorer<score_type> scorer(scoringParameters);

	scorer.setN1(sequence.getSize());

	scoreCache.resize(sequence.getLength() + FIRST_SEQ_POSITION, 0);
	//this is the beginning
	scoreCache[FIRST_SEQ_POSITION - 1] = 0;

	MIter mIter;
	sequence.getMIBegin(mIter);
	MIter mEnd;
	sequence.getMIEnd(mEnd);

	for(Seq::size_type position = FIRST_SEQ_POSITION; mIter != mEnd; ++mIter, ++position)
	{
		score_type newScore = 0;
		newScore = scorer.scoreMIter(mIter);				

		scoreCache[position] = scoreCache[position - 1] + newScore;

		assert(scoreCache[position] - scoreCache[position - 1] == newScore);
	}
}

void HeuristicIndex::_initializePatternSet(Mseq& sequence, const Vec<score_type>& sequenceScoreCache, const Seq::size_type windowRight, const Seq::size_type windowLeft, const bool allowZerosAtEnds, const TyphonNS::Pattern::size_type minAllowableShift, const Vec<bool>& shouldUse)
{
	//this stores the scores of each pattern for the whole sequence
	typedef unsigned int pattern_key_type;
	typedef hash_map_type<pattern_key_type, pattern_score_type> pattern_score_table_type;
	pattern_score_table_type patternScoreTable;

	//this stores the current caches of scores for the positions
	typedef list<score_position_type> highest_scoring_positions_type;
	highest_scoring_positions_type highestScoringPositions;
	typedef queue<score_type> score_cache_type;
	score_cache_type scoreCache;

	//since we always pop before pushing
	score_position_type dummy(0,0);
	scoreCache.push(dummy.first);
	highestScoringPositions.insert(highestScoringPositions.end(), dummy);

	score_type totalScore = 0;

	//get statistics on the score
	Seq::size_type endPos = sequence.getLength() + FIRST_SEQ_POSITION;
	Seq::size_type base = FIRST_SEQ_POSITION;

	double meanScore = 0;
	double meanScoreSquared = 0;
	double devScore = 0;
	Seq::size_type effLength = 0;
	for (Seq::size_type seqPos = FIRST_SEQ_POSITION; seqPos != endPos; ++seqPos)
	{
		if (!shouldUse[seqPos])
		{
			continue;
		}
		effLength++;
		score_type currentScore = _getScore(sequenceScoreCache, seqPos, windowRight, windowLeft);

		meanScore += currentScore;
		meanScoreSquared += (static_cast<double>(currentScore) * currentScore);

	}
	meanScore /= effLength;
	meanScoreSquared /= effLength;
	devScore = meanScoreSquared - meanScore * meanScore;
	devScore = sqrt(devScore);

	//rampup; load the first values in there
	Seq::size_type rampupEnd = FIRST_SEQ_POSITION + _patternSpan - 1;
	for (Seq::size_type seqPos = FIRST_SEQ_POSITION; seqPos != rampupEnd; ++seqPos)
	{
		score_type currentScore = _getScore(sequenceScoreCache, seqPos);
		scoreCache.push(currentScore);
		totalScore += currentScore;
		highestScoringPositions.insert(upper_bound(highestScoringPositions.begin(), highestScoringPositions.end(), currentScore, &HeuristicIndexNS::compare_for_upper_score), score_position_type(currentScore, seqPos));
	}

	for (Seq::size_type seqPos = rampupEnd; seqPos != endPos; ++seqPos, ++base)
	{

		//extract the minimum position in the sequence
		score_type scorePopped = scoreCache.front();
		scoreCache.pop();
		totalScore -= scorePopped;
		//get the score of the new position
		score_type newScore = _getScore(sequenceScoreCache, seqPos);
		//insert the new position
		scoreCache.push(newScore);
		totalScore += newScore;

		//now remove it from the highestScoringPositions
		highest_scoring_positions_type::iterator toRemove = lower_bound(highestScoringPositions.begin(), highestScoringPositions.end(), scorePopped, &HeuristicIndexNS::compare_for_lower_score);
		highestScoringPositions.erase(toRemove);

		//now insert into the correct place in the highestScoringPositions
		highest_scoring_positions_type::iterator q = highestScoringPositions.end();

		highestScoringPositions.insert(upper_bound(highestScoringPositions.begin(), highestScoringPositions.end(), newScore, &HeuristicIndexNS::compare_for_upper_score), score_position_type(newScore, seqPos));

		if (!shouldUse[seqPos])
		{
			continue;
		}

 		score_type currentScore = _getScore(sequenceScoreCache, seqPos, windowRight, windowLeft);

		if (currentScore < meanScore)
		{
			//continue;
		}

		score_type lowScore = highestScoringPositions.rbegin()->first;
		score_type highScore = highestScoringPositions.begin()->first;
		if (highScore == lowScore)
		{
			//continue;
		}

		//now pick the best pattern
 		highest_scoring_positions_type::reverse_iterator hIt = highestScoringPositions.rbegin();

		pattern_key_type currentKey = 0;
		for (TyphonNS::Pattern::size_type i = 0; i < _numZeros; ++i, ++hIt)
		{
			TyphonNS::Pattern::size_type positionInPattern = hIt->second - base;
			if (!allowZerosAtEnds && (positionInPattern == 0 || positionInPattern == _patternSpan - 1))
			{
				--i;
				continue;
			}
			currentKey += (1u << positionInPattern);
		}

		patternScoreTable[currentKey] += currentScore;
	}

	assert(patternScoreTable.size() > 0);

	//now let's check each pattern and extract the highest
	list<pattern_score_key_type> sortedPatternList;

	pattern_score_table_type::const_iterator tEnd = patternScoreTable.end();
	for (pattern_score_table_type::const_iterator it = patternScoreTable.begin(); it != tEnd; ++it)
	{
		sortedPatternList.insert(upper_bound(sortedPatternList.begin(), sortedPatternList.end(), it->second, &HeuristicIndexNS::compare_for_upper_pattern), pattern_score_key_type(it->second, it->first));
	}

	//now get the number of patterns
	list<pattern_score_key_type>::iterator pEnd = sortedPatternList.end();
	pattern_num_type currentPatternNum = 0;
	Vec<pattern_key_type> patternKeysChosen;
 	for (list<pattern_score_key_type>::iterator it = sortedPatternList.begin(); currentPatternNum != _numPatterns && it != pEnd; it++)
	{
		//let's check if we should use this pattern
 		bool usePattern = true;
		Vec<pattern_key_type>::iterator kEnd = patternKeysChosen.end();
		for (Vec<pattern_key_type>::iterator kIt = patternKeysChosen.begin(); kIt != kEnd; kIt++)
		{
			unsigned int copyOfKeyChosen = *kIt;
			unsigned int copyOfCurrent = it->second;

			//see if this is a shift of it
			unsigned int numShifted = 0;
			while (copyOfKeyChosen > 0 && numShifted < minAllowableShift)
			{
				if (copyOfKeyChosen == it->second)
				{
					usePattern = false;
					break;
				}
				else
				{
					copyOfKeyChosen >>= 1;
					numShifted++;
				}
			}

			//see if this is a shift of it
			numShifted = 0;
			while (copyOfCurrent > 0 && numShifted < minAllowableShift)
			{
				if (copyOfCurrent == *kIt)
				{
					usePattern = false;
					break;
				}
				else
				{
					copyOfCurrent >>= 1;
					numShifted++;
				}
			}
		}

		//if not different enough skip
		if (!usePattern)
		{
			continue;
		}

		currentPatternNum++;

		patternKeysChosen.push_back(it->second);

		_patterns.push_back(TyphonNS::Pattern(~it->second, _patternSpan));

		//DEBUG MESSAGE
		ostringstream message;
		message << "Pattern " << currentPatternNum << " is " << _patterns.back() << " with count " << it->first;
		//cerr << "Debug: " << message.str() << endl;
	}

	//in case have more specified than exist
	_numPatterns = _patterns.size();


	//DEBUG MESSAGE
	ostringstream message;
	message << "We considered " << sortedPatternList.size() << " different patterns";
	//cerr << "Debug: " << message.str() << endl;

}

 void HeuristicIndex::_getSortedPositionToPosition(Vec<Seq::size_type>& sortedPositionToPosition, const Vec<score_type>& scoreCache, const Seq::size_type windowRight, const Seq::size_type windowLeft) const
{
	Vec<size_to_conservation_type> positionToConservationScore(scoreCache.size() - FIRST_SEQ_POSITION);
		
	//insert all of the correct values
	Seq::size_type pos = FIRST_SEQ_POSITION;
	Vec<size_to_conservation_type>::iterator cEnd = positionToConservationScore.end();	
	for (Vec<size_to_conservation_type>::iterator cIt = positionToConservationScore.begin(); cIt != cEnd; cIt++, pos++)
	{
		cIt->first = pos;
		cIt->second = _getScore(scoreCache, pos, windowRight, windowLeft);
	}

	//now we need to sort the list
	sort(positionToConservationScore.begin(), positionToConservationScore.end(), &HeuristicIndexNS::compare_for_sort);
				
	//store the mapping
	sortedPositionToPosition.resize(scoreCache.size());
				
	//in case it changed during sorting
	cEnd = positionToConservationScore.end();	
				
	Vec<Seq::size_type>::iterator pIt = sortedPositionToPosition.begin() + FIRST_SEQ_POSITION;
	for (Vec<size_to_conservation_type>::iterator cIt = positionToConservationScore.begin(); cIt != cEnd; ++cIt, ++pIt)
	{
		*pIt = cIt->first;
	}
}

void HeuristicIndex::_getRegionsToUse(Vec<bool>& shouldUse, const Vec<score_type>& scoreCache, const Seq::size_type windowRight, const Seq::size_type windowLeft, const bool highest, const float percentage) const
{
	Vec<Seq::size_type> sortedPositionToPosition;
	_getSortedPositionToPosition(sortedPositionToPosition, scoreCache, windowRight, windowLeft);
	_getRegionsToUse(shouldUse, sortedPositionToPosition, highest, percentage);
}

void HeuristicIndex::_getRegionsToUse(Vec<bool>& shouldUse, const Vec<Seq::size_type>& sortedPositionToPosition, const bool highest, const float percentage) const
{
	shouldUse.clear();
	shouldUse.resize(sortedPositionToPosition.size(), false);

	Seq::size_type start;
	Seq::size_type end;
	if (highest)
	{
		start = static_cast<Seq::size_type>((1 - percentage) * shouldUse.size());
		if (start < FIRST_SEQ_POSITION)
		{
			start = FIRST_SEQ_POSITION;
		}
		end = shouldUse.size();
	}
	else
	{
		start = FIRST_SEQ_POSITION;
		end = static_cast<Seq::size_type>(percentage * shouldUse.size());
		if (end > shouldUse.size())
		{
			end = shouldUse.size();
		}

	}

	for (Seq::size_type i = start; i != end; i++)
	{
		shouldUse[sortedPositionToPosition[i]] = true;
	}
}


HeuristicIndex::score_type HeuristicIndex::_getScore(const Vec<score_type>& scoreCache, Seq::size_type position, Seq::size_type windowRight, Seq::size_type windowLeft) const
{
	Seq::size_type length = scoreCache.size();

	if (position + windowRight >= length)
	{
		windowRight = length - position - 1;
	}
	if (windowLeft >= position)
	{
		windowLeft = position - 1;
	}

	return static_cast<score_type>(scoreCache[position + windowRight] - scoreCache[position - windowLeft - 1]);
}


bool HeuristicIndexNS::compare_for_sort(HeuristicIndex::size_to_conservation_type s, HeuristicIndex::size_to_conservation_type t)
{
	return s.second < t.second;
}

bool HeuristicIndexNS::compare_for_lower_score(HeuristicIndex::score_position_type sp, HeuristicIndex::score_type s)
{
	return sp.first > s;
}

bool HeuristicIndexNS::compare_for_upper_score(HeuristicIndex::score_type s, HeuristicIndex::score_position_type sp)
{
	return s > sp.first;
}

bool HeuristicIndexNS::compare_for_lower_pattern(HeuristicIndex::pattern_score_key_type sp, HeuristicIndex::pattern_score_type s)
{
	return sp.first > s;
}

bool HeuristicIndexNS::compare_for_upper_pattern(HeuristicIndex::pattern_score_type s, HeuristicIndex::pattern_score_key_type sp)
{
	return s > sp.first;
}

bool HeuristicIndexNS::compare_for_score_pattern(HeuristicIndex::score_type s, HeuristicIndex::score_pattern_num_type sp)
{
	return s < sp.first;
}
