#ifndef _heuristicindex_h_included_
#define _heuristicindex_h_included_

#include "System.h"
#include "MatchFinder.h"
#include "Mseq.h"
#include "ScoringParameters.h"
#include "LocalHit.h"
#include "Pattern.h"
#include <map>
#include <deque>
#include <ext/hash_map>
#include "Constants.h"
#include "Profile.h"
#include <cassert>
#include <list>
#include <queue>

/**
 * This is the main index class
 */

class HeuristicIndex
{

 public:

	/* Constants */
	/* ========= */
	static const Seq::size_type FIRST_SEQ_POSITION = 1;
	static const Seq::size_type WINDOW_SIZE = 50;

	/* Types */
	/* ===== */

	typedef LocalHit::score_type score_type;

	typedef float num_devs_type;
	typedef float avg_pattern_num_type;
	typedef Vec<TyphonNS::Pattern>::size_type pattern_num_type;
	typedef float avg_occurence_num_type;
	typedef Seq::size_type occurrence_num_type;
	typedef pair<score_type, pattern_num_type> score_pattern_num_type;

	typedef pair<score_type, Seq::size_type> score_position_type;
	typedef float pattern_score_type;
	typedef pair<pattern_score_type, unsigned int> pattern_score_key_type;

	typedef score_type conservation_score;
	typedef pair<Seq::size_type, conservation_score> size_to_conservation_type; 
	typedef unsigned char seq_letter_type;

	/* Constructor */
	/* =========== */
	
	/**
	 * Build an HeuristicIndex
	 * @param sequence: The sequence from which to build the index
	 * @param scoringParameters The parameters to use for scoring
	 * @param numDevs: kmers occuring more frequently that numDevs
	 *                 deviations above the mean will be discarded
	 *                 Specify 0 to discard no kmers
	 * @param patternSet: The patterns that we will use for indexing/scanning
	 *                    If it contains an empty pattern, we will fill it with dynamically
	                      chosen patterns
	 * @param kLength: The kLength for dynamic patterns; ignored if patternSet full
	 * @param numZeros: The number of zeros for dynamic patterns; ignored if patternSet full
	 * @avgNumPatternsPerPosition: the avg num of patterns to index per position
	 */

  template<class LetterPolicy> HeuristicIndex(Mseq& sequence, LetterPolicy letterPolicy, ScoringParameters<score_type>& scoringParameters, const float numDevs, const Vec<TyphonNS::Pattern>& patternSet, TyphonNS::Pattern::size_type kLength, const TyphonNS::Pattern::size_type numZeros, avg_pattern_num_type avgNumPatternsPerPosition, const float ratioToUse);

	/**
	 * Scan an HeuristicIndex
	 * @param sequence1 The sequence to scan
	 * @param sequence2 The sequence to scan
	 * @param scoringParameters The parameters to use for scoring
	 * @param start2 The offset that the 
	 * @offset: the offset in a larger sequence of this one; this is just for bookeeping	 */

  template<class LetterPolicy> void scanIndex(LetterPolicy letterPolicy, Mseq& sequence1, Mseq& sequence2, MatchFinder& m, ScoringParameters<score_type>& scoringParameters, const Seq::size_type indexOffset, const Seq::size_type queryOffset);

 private:

	/* Types */
	/* ===== */

	typedef Mseq::size_type num_keys_type;
	typedef Vec<Seq::size_type>::size_type key_type;
	typedef Vec<Vec<key_type>::size_type> last_type;
	typedef Vec<key_type> key_list_type;
	typedef deque<key_type> current_key_letters_type;
	typedef Vec<current_key_letters_type> word_cache_type;
	typedef deque<score_type> score_cache_type;

	typedef Vec<Seq::size_type> seq_pos_map;
	typedef Vec<Seq::size_type> fast_direct_kmer_map;
	typedef Vec<seq_pos_map::iterator> fast_seq_pos_iterator_map;
	typedef Vec<unsigned int> fast_direct_kmer_count_map;
	typedef hash_map_type<key_type, Seq::size_type> slow_direct_kmer_map;
	typedef hash_map_type<key_type, seq_pos_map::iterator> slow_seq_pos_iterator_map;
	typedef hash_map_type<key_type, unsigned int> slow_direct_kmer_count_map;



	/* Helper Methods */
	/* ============== */

 private:

	/**
	 * Copies into scoreCache the sum of pairs score for each position in the alignment
	 */
	void _initializeScoreCache(Mseq& sequence, Vec<score_type>& scoreCache, ScoringParameters<score_type>& scoringParameters) const;

	/**
	 * Initializes all of the patterns we'll use
	 */
	void _initializePatternSet(Mseq& sequence, const Vec<score_type>& sequenceScoreCache, const Seq::size_type windowRight, const Seq::size_type windowLeft, const bool allowZerosAtEnds, const TyphonNS::Pattern::size_type minAllowableShift, const Vec<bool>& shouldUse);

	/**
	 * Copies into sortedPositionToPosition a mapping from the sorted position to the actual position in the multiple alignment
	 */
	void _getSortedPositionToPosition(Vec<Seq::size_type>& sortedPositionToPosition, const Vec<score_type>& scoreCache, const Seq::size_type windowRight, const Seq::size_type windowLeft) const;

	/**
	 * These methods copy into shouldUse a vector of bools, which is true if the corresponding position should be indexed
	 */
	void _getRegionsToUse(Vec<bool>& shouldUse, const Vec<score_type>& scoreCache, const Seq::size_type windowRight, const Seq::size_type windowLeft, const bool highest, const float percentage) const;
	void _getRegionsToUse(Vec<bool>& shouldUse, const Vec<Seq::size_type>& sortedPositionToPosition, const bool highest, const float percentage) const;

	/**
	 * Given a score cache, a helper to get the score at a given position plus 
	 * a window
	 */
	score_type _getScore(const Vec<score_type>& scoreCache, Seq::size_type position, Seq::size_type windowRight = 0, Seq::size_type windowLeft = 0) const;

	/**
	 * Helper method for scanning an entire sequence
	 */

	typedef struct BI
	{
		Vec<unsigned int>* numKmers;
		Vec<fast_direct_kmer_count_map>* fastCount;
		Vec<slow_direct_kmer_count_map>* slowCount;
		Vec<fast_direct_kmer_map>* fastFirstOccurrence;
		Vec<Vec<Seq::size_type> >* nextOccurrence;
		Vec<Vec<Seq::size_type> >* kmerPosToSeqPos;
		Vec<score_type>* scoreCache;
		bool useTempVectors;
		bool lookupNumPatterns;
		Vec<pattern_num_type>* maxNumPatternsPerPosition;
		pattern_num_type patternsPerPosition;
		Vec<bool>* shouldUse;
	} BuildInformation;

	typedef struct SI
	{
		Mseq* sequenceIndexed;
		MatchFinder* m;
		bool useTempVectors;
	} ScanInformation;


	template<class LetterPolicy> void _scanSequence(Mseq& sequence
																									,LetterPolicy& letterPolicy
																									,BuildInformation* buildInfo
																									,ScanInformation* scanInfo
																									);

	/* Members */
	/* ======= */

 private:

	/**
	 * This stores how many bits we'll use for each letter in the key
	 */
	unsigned int _bitsPerLetter;

	/**
	 * Bookeeping information on memory
	 */
	unsigned int _totalMemory;
	unsigned int _maxMemory;
	unsigned int _permanentVectorsSize;
	unsigned int _numUniqueKmers;

	/**
	 * This stores information on the class of patterns that we'll be building
	 */
	pattern_num_type _numPatterns;
	Vec<TyphonNS::Pattern> _patterns;
	TyphonNS::Pattern::size_type _kLength;	
	TyphonNS::Pattern::size_type _numZeros;	
	TyphonNS::Pattern::size_type _patternSpan;

	/**
	 * We use this to implement non-consecutive patterns
	 * These pointers tell us what positions to subtract and
	 * what positions to add after each shift of a new character
	 */

	Vec<last_type> _firstZeroPositions;
	Vec<Vec<unsigned int> > _firstZeroShamt;
	Vec<last_type> _firstOnePositions;
	Vec<Vec<unsigned int> > _firstOneShamt;

	/**
	 * This is the storage structure for the index
	 */
	Vec<fast_seq_pos_iterator_map> _fastFirstIterator;
	Vec<seq_pos_map> _seqPosList;

};

namespace HeuristicIndexNS
{
	bool compare_for_sort(HeuristicIndex::size_to_conservation_type s, HeuristicIndex::size_to_conservation_type t);

	bool compare_for_lower_score(HeuristicIndex::score_position_type sp, HeuristicIndex::score_type s);
	bool compare_for_upper_score(HeuristicIndex::score_type s, HeuristicIndex::score_position_type sp);

	bool compare_for_lower_pattern(HeuristicIndex::pattern_score_key_type sp, HeuristicIndex::pattern_score_type s);
	bool compare_for_upper_pattern(HeuristicIndex::pattern_score_type s, HeuristicIndex::pattern_score_key_type sp);

	bool compare_for_score_pattern(HeuristicIndex::score_type s, HeuristicIndex::score_pattern_num_type sp);
}

template<class LetterPolicy> HeuristicIndex::HeuristicIndex(Mseq& sequence, LetterPolicy letterPolicy, ScoringParameters<score_type>& scoringParameters, const float numDevs, const Vec<TyphonNS::Pattern>& patternSet, TyphonNS::Pattern::size_type kLength, const TyphonNS::Pattern::size_type numZeros, avg_pattern_num_type avgNumPatternsPerPosition, const float ratioToUse) : _kLength(kLength), _numZeros(numZeros), _patternSpan(kLength + numZeros)
{
	time_t totalSeconds = time(NULL);

	//determine whether we need to pick a new pattern set or not
	bool needToConstructPatterns = false;
	_numPatterns = patternSet.size();
	if (_numPatterns == 0)
	{
		needToConstructPatterns = true;
		_numPatterns = 1;
	}
	else
	{
		_numPatterns = patternSet.size();
		_patterns.clear();
		_patterns.resize(_numPatterns);
		copy(patternSet.begin(), patternSet.end(), _patterns.begin());

		TyphonNS::Pattern::size_type maxKLength = 0;
		for (Vec<TyphonNS::Pattern>::iterator it = _patterns.begin(); it != _patterns.end(); ++it)
		{
			if (it->getKLength() > maxKLength)
			{
				maxKLength = it->getKLength();
			}
		}

		if (maxKLength == 0)
		{
			_patterns.clear();
			needToConstructPatterns = true;
		}
		else
		{
			for (Vec<TyphonNS::Pattern>::iterator it = _patterns.begin(); it != _patterns.end(); ++it)
			{
				if (it->getKLength() < maxKLength)
				{
					it->pad(maxKLength - it->getKLength(), true);
				}
			}
			//now all have the same weight.  We need to ensure all have the same span
			TyphonNS::Pattern::size_type maxPatternSpan = 0;
			for (Vec<TyphonNS::Pattern>::iterator it = _patterns.begin(); it != _patterns.end(); ++it)
			{
				if (it->getPatternLength() > maxPatternSpan)
				{
					maxPatternSpan = it->getPatternLength();
				}
			}

			for (Vec<TyphonNS::Pattern>::iterator it = _patterns.begin(); it != _patterns.end(); ++it)
			{
				if (it->getPatternLength() < maxPatternSpan)
				{
					it->pad(maxKLength - it->getKLength(), false);
				}
			}
			_kLength = maxKLength;
			_patternSpan = maxPatternSpan;
			_numZeros = _patternSpan - _kLength;
		}
	}

	//set up our data structures
	_numUniqueKmers = static_cast<unsigned int>(pow(static_cast<float>(BIO::NUM_NUCLEOTIDES), static_cast<float>(_kLength)));

	//determine whether we'll use vectors or hash maps for our internal and temporary data structures
	bool useTempVectors = true;
	{
		//determine the memory we'll assume
		_totalMemory = 500000000;
		_maxMemory = _totalMemory / 2;

		_permanentVectorsSize = _numPatterns * sizeof(Seq::size_type) * _numUniqueKmers;
		//temp vectors are a count for each kmer in each index, a total count for each kmer for each pattern, a last sequence position, and a last kmer position
		unsigned int numCounts = 1;

		unsigned int totalTempVectorsSize = numCounts * _numPatterns * sizeof(unsigned int) * _numUniqueKmers + sizeof(Seq::size_type) * _numUniqueKmers * 2 * _numPatterns;

		//determine whether to use vectors or hash maps
		if (_permanentVectorsSize + totalTempVectorsSize < _maxMemory)
		{
			useTempVectors = true;
		}
		else
		{
			useTempVectors = false;
		}		

		//DEBUG MESSAGE
		ostringstream message;
		message << "Assuming total memory of " << _totalMemory / 1000000<< "MB";
		message << endl << "Debug: Using vectors for internal storage";

		if (useTempVectors)
		{
			message << endl << "Debug: Using vectors for temporary storage";
		}
		else
		{
			message << endl << "Debug: Using hash tables for temporary storage";
		}
		//cerr << "Debug: " << message.str() << endl;
	}

	_bitsPerLetter = static_cast<unsigned int>(ceil(log(static_cast<double>(BIO::NUM_NUCLEOTIDES)) / M_LN2));
	//make sure that we won't overflow our keys
	if (_bitsPerLetter * _kLength >= sizeof(key_type) * CHAR_BIT)
	{
		throw BIO::E_BioEx("W is too big");
	}
	
	//go about actually building the index

	//BEGIN INITIALIZATION

	//required for converting letters to numbers
	initialize_prof_idx();

	const Seq::size_type windowLeft = WINDOW_SIZE;
	const Seq::size_type windowRight = WINDOW_SIZE;

	//this is a temp cache for storing sp scores
	//initialize the score cache
	Vec<score_type> scoreCache;
	_initializeScoreCache(sequence, scoreCache, scoringParameters);

	//this tells us whether to index certain portions of the sequence
	Vec<bool> shouldUse;
	_getRegionsToUse(shouldUse, scoreCache, windowRight, windowLeft, true, ratioToUse);

	time_t seconds = time(NULL);

	{
		ostringstream message("");
		message << "Initializing score cache took " << time(NULL) - seconds << " seconds";
		//cout << message.str() << endl;
	}

	//Pick the set of patterns
	seconds = time(NULL);

	if (needToConstructPatterns)
	{
		_initializePatternSet(sequence, scoreCache, windowRight, windowLeft, true, 6, shouldUse);
	}

	if (_numPatterns == 1)
	{
		avgNumPatternsPerPosition = 1;
	}

	{
		ostringstream message("");
		message << "Initializing patterns took " << time(NULL) - seconds << " seconds";
		//cout << message.str() << endl;
	}

	//store the patterns in a better format
	//now set up the positions that we'll need to track
	_firstZeroPositions.resize(_numPatterns);
	_firstZeroShamt.resize(_numPatterns);
	_firstOnePositions.resize(_numPatterns);
	_firstOneShamt.resize(_numPatterns);
	
	Vec<TyphonNS::Pattern>::const_iterator pEnd = _patterns.end();
	Vec<last_type>::iterator fzpIt = _firstZeroPositions.begin();
	Vec<Vec<unsigned int> >::iterator fzsIt = _firstZeroShamt.begin();
	Vec<last_type>::iterator fopIt = _firstOnePositions.begin();
	Vec<Vec<unsigned int> >::iterator fosIt = _firstOneShamt.begin();
	for (Vec<TyphonNS::Pattern>::const_iterator pIt = _patterns.begin(); pIt != pEnd; ++pIt, ++fzpIt, ++fzsIt, ++fopIt, ++fosIt)
	{
		bool seenOne = false;
		//the position off the end is a zero
		bool seenZero = true;
		unsigned int shamt = _kLength;
		for (TyphonNS::Pattern::size_type i = 0; i < _patternSpan; i++)
		{
			if ((*pIt)[i])
			{
				--shamt;
				//if we were just in a region of zeros, the last position is the last zero
				if (seenZero)
				{
					fopIt->push_back(i);
					fosIt->push_back(shamt * _bitsPerLetter);
				}
				seenOne = true;
				seenZero = false;
			}
			else
			{
				//if we were just in a region of ones, the last position is a last one
				if (seenOne)
				{
					fzpIt->push_back(i);
					fzsIt->push_back((shamt) * _bitsPerLetter);
				}
				seenZero = true;
				seenOne = false;
			}
		}
		if (seenOne)
		{
			fzpIt->push_back(_patternSpan);
			fzsIt->push_back(0);
		}
	}


	//now we are going to create the mapping from position to number of patterns
	Vec<pattern_num_type> maxNumPatternsPerPosition;

	bool lookupNumPatterns = false;
	pattern_num_type patternsPerPosition = static_cast<pattern_num_type>(avgNumPatternsPerPosition);

	seconds = time(NULL);

	//now we need to initialize the number of patterns used by each position

	if (avgNumPatternsPerPosition <= 1)
	{
		patternsPerPosition = 1;
	}
	else if (avgNumPatternsPerPosition == _numPatterns)
	{
		patternsPerPosition = _numPatterns;
	}
	else
	{
		//here we have to do some work
		lookupNumPatterns = true;
		maxNumPatternsPerPosition.resize(sequence.getLength() + FIRST_SEQ_POSITION, 0);
		
		//determine how many to choose for each numPatterns
		pattern_num_type maxNum = _numPatterns;
		//determine which way we'll weight the numbers
		Vec<float> ratios(maxNum);
		
		//we'll use the floor as the median
		pattern_num_type middle = static_cast<pattern_num_type>(avgNumPatternsPerPosition);
		float meanBelow = 0;
		float meanAbove = 0;
		float ratioBelow = 0;
		float ratioAverage = 0;
		float ratioAbove = 0;
		
		//find the mean of those above
		for (pattern_num_type i = middle + 1; i <= maxNum; i++)
		{
			meanAbove += i;
		}
		meanAbove /= maxNum - middle;
		
		if (middle == 1)
		{
			ratioAverage = static_cast<float>(meanAbove - avgNumPatternsPerPosition) / static_cast<float>(meanAbove - 1);
			ratioAbove = 1 - ratioAverage;
			ratioBelow = 0;
		}
		else
		{
			ratioAverage = 1.0f / 3.0f;
			
			for (pattern_num_type i = 1; i != middle; i++)
			{
				meanBelow += i;
			}
			meanBelow /= middle - 1;
			
			ratioBelow = (1 - ratioAverage) * static_cast<float>(meanAbove - avgNumPatternsPerPosition) / static_cast<float>(meanAbove - meanBelow);
			ratioAbove = 1 - ratioAverage - ratioBelow;
		}
		
		for (pattern_num_type i = 0; i < middle - 1; i++)
		{
			ratios[i] = ratioBelow / (middle - 1);
		}
		ratios[middle - 1] = ratioAverage;
		for (pattern_num_type i = middle; i < maxNum; i++)
		{
			ratios[i] = ratioAbove / (maxNum - middle);
		}
		
		//now determine the actual numbers
		
		Vec<float>::iterator rEnd = ratios.end();
		
		for(Vec<float>::iterator rIt = ratios.begin(); rIt != rEnd; rIt++)
		{
			*rIt *= (sequence.getLength() - _patternSpan + 1) * ratioToUse;
		}

		//now we need a list of the sequence positions in sorted order
		//this is what will do the actual sorting
		Vec <Seq::size_type> sortedPositionToPosition;
		_getSortedPositionToPosition(sortedPositionToPosition, scoreCache, windowRight, windowLeft);
		
		//fill up the data structure
		unsigned int numInThisBracket = 0;		
		pattern_num_type curNum = 1;
		
		//we need to start/end at different places
		Seq::size_type start = FIRST_SEQ_POSITION;
		Seq::size_type end = sequence.getLength();
		if (ratioToUse < 1)
		{
			//determine start/end based on number of kmers
			start = static_cast<Seq::size_type>((1 - ratioToUse) * sequence.getLength());
			if (start < FIRST_SEQ_POSITION)
			{
				start = FIRST_SEQ_POSITION;
			}
			end = sequence.getLength();
		}
		
		for (Seq::size_type i = start; i <= end - _patternSpan + 1; i++)
		{
			maxNumPatternsPerPosition[sortedPositionToPosition[i]] = curNum;
			numInThisBracket++;
		
			//if we aren't at the final bracket, and if we have exceeded the number per, then increment the index number
			if (curNum < maxNum && numInThisBracket > ratios[curNum - 1])
			{
				ostringstream message;
				message << numInThisBracket << " will use " << curNum << " patterns";
				//cerr << "Debug: " << message.str() << endl;
				
				curNum++;
				numInThisBracket = 0;
			}
		}
		ostringstream message;
		message << numInThisBracket << " will use " << curNum << " patterns";
		//cerr << "Debug: " << message.str() << endl;
	}
	
	{
		ostringstream message("");
		message << "Initializing patterns per position took " << time(NULL) - seconds << " seconds";
		//cout << message.str() << endl;
	}

	Vec<unsigned int> numKmers(_numPatterns, 0);
	//this is temporary; store the counts of each kmer
	Vec<fast_direct_kmer_count_map> fastCount;
	Vec<slow_direct_kmer_count_map> slowCount;
	Vec<fast_direct_kmer_map> fastFirstOccurrence(_numPatterns);
	//we make a temp because we don't know the size a priori
	//as it grows, it will have too much memory
	//so we want to initialize the actual one to trunc the un-needed memory
	Vec<Vec<Seq::size_type> > nextOccurrence;
	
	//another temp
	Vec<Vec<Seq::size_type> > kmerPosToSeqPos;

	//this is a utility to pass the parameters to the helper function
	BuildInformation buildInformation;
	buildInformation.numKmers = &numKmers;
	buildInformation.fastCount = &fastCount;
	buildInformation.slowCount = &slowCount;
  buildInformation.fastFirstOccurrence = &fastFirstOccurrence;
  buildInformation.nextOccurrence = &nextOccurrence;
  buildInformation.kmerPosToSeqPos = &kmerPosToSeqPos;
  buildInformation.scoreCache = &scoreCache;
  buildInformation.useTempVectors = useTempVectors;
  buildInformation.lookupNumPatterns = lookupNumPatterns;
  buildInformation.maxNumPatternsPerPosition = &maxNumPatternsPerPosition;
  buildInformation.patternsPerPosition = patternsPerPosition;
  buildInformation.shouldUse = &shouldUse;
 		
	_scanSequence(sequence, letterPolicy, &buildInformation, NULL);

	//BEGIN POST PROCESSING

	if (numDevs)
	{
		time_t seconds = time(NULL);

		Vec<unsigned int>::iterator nkEnd = numKmers.end();
		Vec<fast_direct_kmer_count_map>::iterator fcIt = fastCount.begin();
		Vec<slow_direct_kmer_count_map>::iterator scIt = slowCount.begin();
		Vec<fast_direct_kmer_map>::iterator ffoIt = fastFirstOccurrence.begin();

		for (Vec<unsigned int>::iterator nkIt = numKmers.begin(); nkIt != nkEnd; ++nkIt, ++fcIt, ++scIt, ++ffoIt)
		{

			const num_devs_type mean = static_cast<float>(*nkIt) / _numUniqueKmers;

			num_devs_type stddev = 0;
			num_devs_type totalNum = 0;

			if (useTempVectors)
			{
				//iterate through the first positions and clear out the frequent ones
				const fast_direct_kmer_count_map::iterator cEnd = fcIt->end();
				for (fast_direct_kmer_count_map::iterator cIt = fcIt->begin(); cIt != cEnd; ++cIt)
				{
					if (*cIt > 0)
					{
						num_devs_type diff = *cIt - mean;
						stddev += diff * diff;
						totalNum++;
					}
				}
			}
			else
			{
				const slow_direct_kmer_count_map::iterator cEnd = scIt->end();
				for (slow_direct_kmer_count_map::iterator cIt = scIt->begin(); cIt != cEnd; ++cIt)
				{
					if (cIt->second > 0)
					{
						num_devs_type diff = cIt->second - mean;
						stddev += diff * diff;
						totalNum++;
					}
				}
			}

			stddev += mean * mean * (_numUniqueKmers - totalNum);
			stddev /= _numUniqueKmers;

			//sqrt doesn't give correct performance when optimized for some reason
			//stddev = sqrt(stddev);
			stddev = (pow(stddev, static_cast<num_devs_type>(.5)));

			//DEBUG COMMENT
			ostringstream message("");
			message << "Average of " << mean << " positions per kmer; " << _numUniqueKmers << " unique but " << totalNum << " appear";

			//DEBUG COMMENT
			message << endl << "Debug: Standard deviation is " << stddev;

			unsigned int numAbove = 0;
			const float threshold = numDevs * stddev;

			if (useTempVectors)
			{
				//iterate through the first positions and clear out the frequent ones
				fast_direct_kmer_map::iterator fIt = ffoIt->begin();

				const fast_direct_kmer_count_map::iterator cEnd = fcIt->end();
				for (fast_direct_kmer_count_map::iterator cIt = fcIt->begin(); cIt != cEnd; cIt++, fIt++)
				{
					if (*fIt)
					{
						const float diff = *cIt - mean;
						if (diff > threshold)
						{
							numAbove++;
							*fIt = 0;
						}
					}
				}
			}
			else
			{
				const slow_direct_kmer_count_map::iterator cEnd = scIt->end();
				for (slow_direct_kmer_count_map::iterator cIt = scIt->begin(); cIt != cEnd; ++cIt)
				{
					if (cIt->second > 0)
					{
						const float diff = cIt->second - mean;
						if (diff > threshold)
						{
							numAbove++;
							(*ffoIt)[cIt->first] = 0;
						}
					}
				}
			}

			//DEBUG MESSAGE
			message << endl << "Debug: " << numAbove << " are above " << numDevs << " standard deviations";
			
			//cerr << "Debug: " << message.str() << endl;
		}

		{
			ostringstream message("");
			message << "Post processing took " << time(NULL) - seconds << " seconds";
			//cout << message.str() << endl;
		}
	}
	
	//END POST PROCESSING

	//BEGIN TRANSFER TO FINAL STORAGE
	_seqPosList.resize(_numPatterns);
	_fastFirstIterator.resize(_numPatterns);

	Vec<fast_direct_kmer_map>::iterator ffoEnd = fastFirstOccurrence.end();
	Vec<fast_direct_kmer_map>::iterator ffoIt = fastFirstOccurrence.begin();
	Vec<Vec<Seq::size_type> >::iterator tkIt = kmerPosToSeqPos.begin();
	Vec<Vec<Seq::size_type> >::iterator tnoIt = nextOccurrence.begin();
	Vec<seq_pos_map>::iterator spIt = _seqPosList.begin();
	Vec<fast_seq_pos_iterator_map>::iterator ffiIt = _fastFirstIterator.begin();
	for (; ffoIt != ffoEnd; ++ffoIt, ++tnoIt, ++spIt, ++ffiIt, ++tkIt)
	{
		Seq::size_type seqPosListSize = 0;

		fast_direct_kmer_map::reverse_iterator dIt = ffoIt->rbegin();
		fast_direct_kmer_map::reverse_iterator dEnd = ffoIt->rend();
		for(; dIt != dEnd; ++dIt)
		{
			for (Seq::size_type currentKmerPosition = *dIt; currentKmerPosition != 0; currentKmerPosition = (*tnoIt)[currentKmerPosition])		
			{
				seqPosListSize++;
			}
		}
		spIt->resize(seqPosListSize);
		seq_pos_map::iterator currentIterator = spIt->end();

		ffiIt->resize(ffoIt->size() + 1);
		//go in reverse order
		fast_seq_pos_iterator_map::reverse_iterator fIt = ffiIt->rbegin();
		dIt = ffoIt->rbegin();

		for(; dIt != dEnd; ++dIt, ++fIt)
		{
			*fIt = currentIterator;
			for (Seq::size_type currentKmerPosition = *dIt; currentKmerPosition != 0; currentKmerPosition = (*tnoIt)[currentKmerPosition])		
			{
				currentIterator--;
				*currentIterator = (*tkIt)[currentKmerPosition];
			}
		}
		*fIt = currentIterator;

		assert(currentIterator == spIt->begin());
	}

	//END TRANSFER TO FINAL STORAGE

	{
		ostringstream message("");
		unsigned int totalSize = 0;
		for (Vec<Vec<Seq::size_type> >::iterator it = nextOccurrence.begin(); it != nextOccurrence.end(); ++it)
		{
			totalSize += it->size();
		}
		message << "Debug: Total size is " << totalSize;
		//cerr << message.str() << endl;
	}	
	{
		ostringstream message("");
		message << "Built index on " << sequence.getLength() - _patternSpan + 1 << " kmers; took " << time(NULL) - totalSeconds << " seconds";
		//cout << message.str() << endl;
	}
}

template<class LetterPolicy> void HeuristicIndex::scanIndex(LetterPolicy letterPolicy, Mseq& sequence1, Mseq& sequence2, MatchFinder& m, ScoringParameters<score_type>& scoringParameters, const Seq::size_type indexOffset, const Seq::size_type queryOffset)
{
	m.initialize(scoringParameters);
	m._clearStats();
	//determine whether we'll use vectors or hash maps for our internal and temporary data structures
	bool useTempVectors = true;
	{
		unsigned int maxTempMemory = _maxMemory > _permanentVectorsSize ? _maxMemory - _permanentVectorsSize : 0;

		unsigned int totalTempVectorsSize = _numPatterns * sizeof(Seq::size_type) * _numUniqueKmers;

		//determine whether to use vectors or hash maps
		if (totalTempVectorsSize < maxTempMemory)
		{
			useTempVectors = true;
		}
		else
		{
			useTempVectors = false;
		}

		//DEBUG MESSAGE
		ostringstream message;
		message << "Assuming total memory of " << _totalMemory << "B";
		if (useTempVectors)
		{
			message << endl << "Debug: Using vectors for temporary storage";
		}
		else
		{
			message << endl << "Debug: Using hash tables for temporary storage";
		}
		//cerr << "Debug: " << message.str() << endl;

	}

	//set up the start offsets
	m._start1Offset = indexOffset;
	m._start2Offset = queryOffset;

	time_t seconds = time(NULL);
	
	//get all of the matches

	ScanInformation scanInformation;
	scanInformation.sequenceIndexed = &sequence1;
	scanInformation.m = &m;
	scanInformation.useTempVectors = useTempVectors;

	_scanSequence(sequence2, letterPolicy, NULL, &scanInformation);

	ostringstream message("");
	message << "Getting matches took " << time(NULL) - seconds << " seconds";
	//DEBUG MESSAGE
	//cout << message.str() << endl;
}


template<class LetterPolicy> void HeuristicIndex::_scanSequence(Mseq& sequence
																											 ,LetterPolicy& letterPolicy
																											 ,BuildInformation* buildInfo
																											 ,ScanInformation* scanInfo
																											 )
{
	bool build = true;
	bool useTempVectors = false;

	if (buildInfo == NULL)
	{
		build = false;
		useTempVectors = scanInfo->useTempVectors;
	}
	else
	{
		useTempVectors = buildInfo->useTempVectors;
	}

	//set up the score caches; they will only be used if build is true
	Vec<score_type> positionScoreCacheForPatterns(_numPatterns, 0);
	//this will store the corresponding to the words in the cache
	score_cache_type positionScoreCache;

	//now set up the code needed for reading in the actual letters

	//we store one key for each word
	num_keys_type numKeys = letterPolicy.getNumKeys(sequence);

	//set up the key cache
	Vec<key_list_type> keyCache(_numPatterns);

	Vec<key_list_type>::iterator kEnd = keyCache.end();
	for (Vec<key_list_type>::iterator kIt = keyCache.begin(); kIt != kEnd; ++kIt)
	{
		kIt->resize(numKeys, 0);
	}

	//these will cache our current words
	word_cache_type wordCache(numKeys);

	//these will store whether we have an 'N' in this position
	Vec<queue<bool> > anyCharHere(numKeys);
	Vec<unsigned int> anyCharCounts(numKeys, 0);

	//END INITIALIZATION

	//BEGIN RAMPUP CODE

	//now, we begin actually reading from the sequence

	//the first thing that we do is to do some rampup code

	//push an extra zero to avoid a special case for the main loop
	const word_cache_type::iterator wEnd = wordCache.end();
	word_cache_type::iterator wIt = wordCache.begin();
	Vec<queue<bool> >::iterator aIt = anyCharHere.begin();
	for (; wIt != wEnd; ++wIt, ++aIt)
	{
		for (TyphonNS::Pattern::size_type j = 0; j < _patternSpan; j++)
		{
			wIt->push_back(0);
		}
		aIt->push(false);
	}

	if (build)
	{
		for (TyphonNS::Pattern::size_type j = 0; j < _patternSpan; j++)
		{
			positionScoreCache.push_back(0);
		}
	}

	//the actual end of the sequence
	MIter actualEndIter;
	sequence.getMIEnd(actualEndIter);

	//we end this phase at the end of the pattern
	MIter rampupEnd;
	sequence.getMIBegin(rampupEnd);
	rampupEnd = rampupEnd + _patternSpan - 1;
	
	//now our rampup code reads the first patternSpan - 1 letters into the caches
	//this is stored for subsequent calls to _nextKmers()
	MIter iter;
	sequence.getMIBegin(iter);
	Seq::size_type seqPos = FIRST_SEQ_POSITION;
	for (; iter != rampupEnd && iter != actualEndIter; ++iter, ++seqPos)
	{
		//BEGIN GETTING LETTERS

		//we store the consensus character
		bool haveConsensus = false;
		seq_letter_type consensusChar = 0;

		Vec<queue<bool> >::iterator aIt = anyCharHere.begin();
		Vec<unsigned int>::iterator cIt = anyCharCounts.begin();
		MIter::iterator it = iter.getBegin();

		const word_cache_type::iterator wEnd = wordCache.end();		
		for (word_cache_type::iterator wIt = wordCache.begin(); wIt != wEnd; ++it, ++aIt, ++cIt, ++wIt)
		{
			key_type adding;
			seq_letter_type letter = letterPolicy.getLetter(iter, it);
			//if we get a gap, we use the consensus
			if (letter == GAP_CHAR)
			{
				if (!haveConsensus)
				{
					consensusChar = iter.getConsensusChar();
					haveConsensus = true;
				}
				letter = consensusChar;
				if (letter == GAP_CHAR)
				{
					letter = ANY_CHAR;
				}
			}

			if (letter == ANY_CHAR)
			{
				//make sure we don't overflow
				adding = 0;
				(*cIt)++;
				aIt->push(true);
			}
			else
			{
				aIt->push(false);
				adding = CHAR_PROFILE_IDX[static_cast<int>(static_cast<unsigned char>(letter))];
			}

			wIt->push_back(adding);
		}
		
		//END GETTING LETTERS

		if (build)
		{
			positionScoreCache.push_back(_getScore(*(buildInfo->scoreCache), seqPos));
		}

		//BEGIN UPDATING KEYS

		Vec<last_type>::const_iterator fzpEnd = _firstZeroPositions.end();

		{
			//we need to update the keys for each pattern
			Vec<last_type>::const_iterator fzpIt = _firstZeroPositions.begin();
			Vec<Vec<unsigned int> >::const_iterator fzsIt = _firstZeroShamt.begin();
			Vec<last_type>::const_iterator fopIt = _firstOnePositions.begin();
			Vec<Vec<unsigned int> >::const_iterator fosIt = _firstOneShamt.begin();			
			Vec<key_list_type>::iterator keyIt = keyCache.begin();			
			for (; fzpIt != fzpEnd; ++fzpIt, ++fzsIt, ++fopIt, ++fosIt, ++keyIt)
			{
				key_list_type::iterator kIt = keyIt->begin();
				for (word_cache_type::iterator wIt = wordCache.begin(); wIt != wEnd; ++wIt, ++kIt)
				{
					//the first thing we need to do is to subtract the letters vanishing after this shift
					const last_type::const_iterator oneEnd = fopIt->end();
					const Vec<unsigned int>::const_iterator oneShamtEnd = fosIt->end();
					last_type::const_iterator oneIt = fopIt->begin();
					Vec<unsigned int>::const_iterator oneShamIt = fosIt->begin();
					
					for (; oneIt != oneEnd; oneIt++, oneShamIt++)
					{
						*kIt -= ((*wIt)[*oneIt] << (*oneShamIt));
					}
					
					//shift everything over by one
					//originalKey = originalKey * NUM_NUCLEOTIDES
					*kIt <<= _bitsPerLetter;
					
					//now we need to add the letters appearing after this shift
					const last_type::const_iterator zeroEnd = fzpIt->end();
					const Vec<unsigned int>::const_iterator zeroShamtEnd = fzsIt->end();
					last_type::const_iterator zeroIt = fzpIt->begin();
					Vec<unsigned int>::const_iterator zeroShamIt = fzsIt->begin();
					
					for (; zeroIt != zeroEnd; zeroIt++, zeroShamIt++)
					{
						*kIt += ((*wIt)[*zeroIt] << (*zeroShamIt));
					}
				}
			}
		}
	
		//END UPDATING KEYS

		if (build)
		{
			//BEGIN UPDATING SCORES

			//we need to update the keys for each pattern
			Vec<last_type>::const_iterator fzpIt = _firstZeroPositions.begin();
			Vec<Vec<unsigned int> >::const_iterator fzsIt = _firstZeroShamt.begin();
			Vec<last_type>::const_iterator fopIt = _firstOnePositions.begin();
			Vec<Vec<unsigned int> >::const_iterator fosIt = _firstOneShamt.begin();			
			Vec<score_type>::iterator scoreForPatternIt = positionScoreCacheForPatterns.begin();			
			for (; fzpIt != fzpEnd; ++fzpIt, ++fzsIt, ++fopIt, ++fosIt, ++scoreForPatternIt)
			{
				//the first thing we need to do is to subtract the scores vanishing after this shift
				const last_type::const_iterator oneEnd = fopIt->end();
				last_type::const_iterator oneIt = fopIt->begin();
				
				for (; oneIt != oneEnd; oneIt++)
				{
					*scoreForPatternIt -= positionScoreCache[*oneIt];
				}
				
				//now we need to add the score appearing after this shift
				const last_type::const_iterator zeroEnd = fzpIt->end();
				last_type::const_iterator zeroIt = fzpIt->begin();
				
				for (; zeroIt != zeroEnd; zeroIt++)
				{
					*scoreForPatternIt += positionScoreCache[*zeroIt];
				}
			}
		
			//END UPDATING SCORES
		}

		for (word_cache_type::iterator wIt = wordCache.begin(); wIt != wEnd; ++wIt)
		{
			wIt->pop_front();
		}

		if (build)
		{
			positionScoreCache.pop_front();
		}
	}	

	//to avoid a special case in the main loopa
	--iter;
	//since we increment iter at the beginning of each iteration
	--actualEndIter;

	//END RAMPUP CODE

	if (build)
	{
		Vec<fast_direct_kmer_map>::iterator fEnd = buildInfo->fastFirstOccurrence->end();
		for(Vec<fast_direct_kmer_map>::iterator fIt = buildInfo->fastFirstOccurrence->begin(); fIt != fEnd; ++fIt)
		{
			fIt->resize(_numUniqueKmers, 0);
		}
		
		if (useTempVectors)
		{
			buildInfo->fastCount->resize(_numPatterns);
			Vec<fast_direct_kmer_count_map>::iterator fEnd = buildInfo->fastCount->end();
			for(Vec<fast_direct_kmer_count_map>::iterator fIt = buildInfo->fastCount->begin(); fIt != fEnd; ++fIt)
			{
				fIt->resize(_numUniqueKmers, 0);
			}
		}
		else
		{
			buildInfo->slowCount->resize(_numPatterns);
		}
		
		//it will need at least this much
		buildInfo->nextOccurrence->resize(_numPatterns);
		buildInfo->kmerPosToSeqPos->resize(_numPatterns);
		Vec<Seq::size_type>::size_type minSize = sequence.getLength() / _numPatterns;
		Vec<Vec<Seq::size_type> >::iterator tEnd = buildInfo->nextOccurrence->end();
		Vec<Vec<Seq::size_type> >::iterator kIt = buildInfo->kmerPosToSeqPos->begin();
		for (Vec<Vec<Seq::size_type> >::iterator tIt = buildInfo->nextOccurrence->begin(); tIt != tEnd; ++tIt, ++kIt)
		{
			tIt->reserve(minSize);
			kIt->reserve(minSize);
			while (tIt->size() < FIRST_SEQ_POSITION)
			{
				tIt->push_back(0);
				//since the first kmer position is 1, all before it map to zero
				kIt->push_back(0);
			}
		}

	}
	
	//this is temporary; it stores for each kmer the last position in the sequence we've found
	Vec<fast_direct_kmer_map> fastLastKmerOccurrence;
	Vec<slow_direct_kmer_map> slowLastKmerOccurrence;
	fastLastKmerOccurrence.resize(_numPatterns);
	slowLastKmerOccurrence.resize(_numPatterns);

	Vec<fast_direct_kmer_map> fastLastSeqOccurrence;
	Vec<slow_direct_kmer_map> slowLastSeqOccurrence;
	fastLastSeqOccurrence.resize(_numPatterns);
	slowLastSeqOccurrence.resize(_numPatterns);

	//this exists solely for checking duplicates in the kmers we're indexing
	if (useTempVectors)
	{
		Vec<fast_direct_kmer_map>::iterator fsEnd = fastLastSeqOccurrence.end();
		Vec<fast_direct_kmer_map>::iterator fkIt = fastLastKmerOccurrence.begin();
		for (Vec<fast_direct_kmer_map>::iterator fsIt = fastLastSeqOccurrence.begin(); fsIt != fsEnd; ++fsIt)
		{
			fsIt->resize(_numUniqueKmers, 0);
			
			if (build)
			{
				fkIt->resize(_numUniqueKmers, 0);
				++fkIt;
			}
		}
	}
	//BEGIN MAIN LOOP

	//this stores the position in the sequence
	//store base 0 since we index from base 0
	Seq::size_type indexPosition = 0;
	Seq::size_type sequencePosition = FIRST_SEQ_POSITION;
	//this stores the position in the larger space of kmers; at each sequence position
	//we may index more than one kmer
	Vec<Seq::size_type> kmerPositions(_numPatterns, FIRST_SEQ_POSITION - 1);
	
	for (; iter != actualEndIter; ++indexPosition, ++sequencePosition, ++seqPos)
	{
		if (sequencePosition % 100000 == 0)
		{
			
			ostringstream message;
			message << "Fetching kmers at position " << sequencePosition;
			//DEBUG COMMENT
			//cerr << "Debug: " << message.str() << endl;
		}

		++iter;

		//BEGIN GETTING LETTERS

		//we store the consensus character
		bool haveConsensus = false;
		seq_letter_type consensusChar = 0;

		Vec<queue<bool> >::iterator aIt = anyCharHere.begin();
		Vec<unsigned int>::iterator cIt = anyCharCounts.begin();
		MIter::iterator it = iter.getBegin();

		const word_cache_type::iterator wEnd = wordCache.end();		
		for (word_cache_type::iterator wIt = wordCache.begin(); wIt != wEnd; ++it, ++aIt, ++cIt, ++wIt)
		{
			key_type adding;
			seq_letter_type letter = letterPolicy.getLetter(iter, it);

			//if we get a gap, we use the consensus
			if (letter == GAP_CHAR)
			{
				if (!haveConsensus)
				{
					consensusChar = iter.getConsensusChar();
					haveConsensus = true;
				}
				letter = consensusChar;
				if (letter == GAP_CHAR)
				{
					letter = ANY_CHAR;
				}
			}

			if (letter == ANY_CHAR)
			{
				//make sure we don't overflow
				adding = 0;
				(*cIt)++;
				aIt->push(true);
			}
			else
			{
				aIt->push(false);
				adding = CHAR_PROFILE_IDX[static_cast<int>(static_cast<unsigned char>(letter))];
			}

			//if we're removing an 'N', adjust the count
			if (aIt->front())
			{
				(*cIt)--;
			}
			aIt->pop();

			wIt->push_back(adding);
		}

		//END GETTING LETTERS

		if (build)
		{
			//BEGIN GETTING SCORES

			positionScoreCache.push_back(_getScore(*(buildInfo->scoreCache), seqPos));

			//END GETTING SCORES
		}

		//BEGIN UPDATING KEYS

		Vec<last_type>::const_iterator fzpEnd = _firstZeroPositions.end();

		//we need to update the keys for each pattern
		Vec<last_type>::const_iterator fzpIt = _firstZeroPositions.begin();
		Vec<Vec<unsigned int> >::const_iterator fzsIt = _firstZeroShamt.begin();
		Vec<last_type>::const_iterator fopIt = _firstOnePositions.begin();
		Vec<Vec<unsigned int> >::const_iterator fosIt = _firstOneShamt.begin();			
		Vec<key_list_type>::iterator keyIt = keyCache.begin();			

		for (; fzpIt != fzpEnd; ++fzpIt, ++fzsIt, ++fopIt, ++fosIt, ++keyIt)
		{
			key_list_type::iterator kIt = keyIt->begin();	
			for (word_cache_type::iterator wIt = wordCache.begin(); wIt != wEnd; ++wIt, ++kIt)
			{
				//the first thing we need to do is to subtract the letters vanishing after this shift
				const last_type::const_iterator oneEnd = fopIt->end();
				const Vec<unsigned int>::const_iterator oneShamtEnd = fosIt->end();
				last_type::const_iterator oneIt = fopIt->begin();
				Vec<unsigned int>::const_iterator oneShamIt = fosIt->begin();

				for (; oneIt != oneEnd; oneIt++, oneShamIt++)
				{

					*kIt -= ((*wIt)[*oneIt] << (*oneShamIt));
				}

				//shift everything over by one
				//originalKey = originalKey * NUM_NUCLEOTIDES
				*kIt <<= _bitsPerLetter;

				//now we need to add the letters appearing after this shift
				const last_type::const_iterator zeroEnd = fzpIt->end();
				const Vec<unsigned int>::const_iterator zeroShamtEnd = fzsIt->end();
				last_type::const_iterator zeroIt = fzpIt->begin();
				Vec<unsigned int>::const_iterator zeroShamIt = fzsIt->begin();

				for (; zeroIt != zeroEnd; zeroIt++, zeroShamIt++)
				{
					*kIt += ((*wIt)[*zeroIt] << (*zeroShamIt));
				}
			}
		}

		//END UPDATING KEYS

		Vec<bool> bestPosition(_numPatterns, false);
		if (build)
		{
			//BEGIN UPDATING SCORES
			list<score_pattern_num_type> bestScoringPatterns;
			{
				//go in reverse so ties go to the earliest ones

				//we need to update the keys for each pattern
				Vec<last_type>::reverse_iterator fzpEnd = _firstZeroPositions.rend();

				Vec<last_type>::reverse_iterator fzpIt = _firstZeroPositions.rbegin();
				Vec<Vec<unsigned int> >::reverse_iterator fzsIt = _firstZeroShamt.rbegin();
				Vec<last_type>::reverse_iterator fopIt = _firstOnePositions.rbegin();
				Vec<Vec<unsigned int> >::reverse_iterator fosIt = _firstOneShamt.rbegin();			
				Vec<score_type>::reverse_iterator scoreForPatternIt = positionScoreCacheForPatterns.rbegin();			
				Vec<TyphonNS::Pattern>::size_type patternCachePosition = _numPatterns - 1;
				for (; fzpIt != fzpEnd; ++fzpIt, ++fzsIt, ++fopIt, ++fosIt, ++scoreForPatternIt, --patternCachePosition)
				{
					//the first thing we need to do is to subtract the scores vanishing after this shift
					const last_type::const_iterator oneEnd = fopIt->end();
					last_type::const_iterator oneIt = fopIt->begin();

					for (; oneIt != oneEnd; oneIt++)
					{
						*scoreForPatternIt -= positionScoreCache[*oneIt];
					}

					//now we need to add the score appearing after this shift
					const last_type::const_iterator zeroEnd = fzpIt->end();
					last_type::const_iterator zeroIt = fzpIt->begin();

					for (; zeroIt != zeroEnd; zeroIt++)
					{
						*scoreForPatternIt += positionScoreCache[*zeroIt];
					}

					bestScoringPatterns.insert(upper_bound(bestScoringPatterns.begin(), bestScoringPatterns.end(), *scoreForPatternIt, &HeuristicIndexNS::compare_for_score_pattern), score_pattern_num_type(*scoreForPatternIt, patternCachePosition));				
				}
			}

			//now let's determine how many patterns we'll use
			unsigned int maxNum = 0;
			if (buildInfo->lookupNumPatterns)
			{
				maxNum = (*(buildInfo->maxNumPatternsPerPosition))[sequencePosition];
			}
			else
			{
				maxNum = buildInfo->patternsPerPosition;
			}

			//we start at the end, since we want the highest scoring value
			unsigned int numBest = 0;

			list<score_pattern_num_type>::reverse_iterator bEnd = bestScoringPatterns.rend();
			for (list<score_pattern_num_type>::reverse_iterator bIt = bestScoringPatterns.rbegin(); bIt != bEnd; ++bIt)
			{
				bestPosition[bIt->second] = true;

				numBest++;

				if (numBest >= maxNum)
				{
					break;
				}
			}

			//END UPDATING SCORES
		}

		for (word_cache_type::iterator wIt = wordCache.begin(); wIt != wEnd; ++wIt)
		{
			wIt->pop_front();
		}
		if (build)
		{
			positionScoreCache.pop_front();
			if (!(*(buildInfo->shouldUse))[sequencePosition])
			{
				continue;
			}
		}

		//BEGIN PROCESSING KEYS

		Vec<key_list_type>::iterator keyEnd = keyCache.end();
		Vec<fast_direct_kmer_map>::iterator flsIt = fastLastSeqOccurrence.begin();
		Vec<slow_direct_kmer_map>::iterator slsIt = slowLastSeqOccurrence.begin();

		//for use while scanning
		Vec<fast_seq_pos_iterator_map>::iterator ffiIt = _fastFirstIterator.begin();
		Vec<seq_pos_map>::iterator spIt = _seqPosList.begin();
		//for use while building
		Vec<fast_direct_kmer_map>::iterator flkIt = fastLastKmerOccurrence.begin();
		Vec<slow_direct_kmer_map>::iterator slkIt = slowLastKmerOccurrence.begin();
		Vec<Vec<Seq::size_type> >::iterator tnoIt;
		Vec<Vec<Seq::size_type> >::iterator tkIt;
		Vec<fast_direct_kmer_map>::iterator ffoIt;
		Vec<fast_direct_kmer_count_map>::iterator fcIt;
		Vec<slow_direct_kmer_count_map>::iterator scIt;
		if (build)
		{
			tnoIt = buildInfo->nextOccurrence->begin();
			tkIt = buildInfo->kmerPosToSeqPos->begin();
			ffoIt = buildInfo->fastFirstOccurrence->begin();
			fcIt = buildInfo->fastCount->begin();
			scIt = buildInfo->slowCount->begin();
		}

		Vec<Seq::size_type>::iterator kpIt = kmerPositions.begin();
		Vec<bool>::iterator bpIt = bestPosition.begin();

		bool needToIncrement = false;

		for (Vec<key_list_type>::iterator keyIt = keyCache.begin(); keyIt != keyEnd; ++keyIt)
		{
			if (needToIncrement)
			{
				if (!build)
				{
					++ffiIt;
					++spIt;
				}
				else
				{
					++flkIt;
					++slkIt;
					++tnoIt;
					++tkIt;
					++ffoIt;
					++fcIt;
					++scIt;
					++kpIt;
					++bpIt;
				}
				++flsIt;
				++slsIt;

			}
			else
			{
				needToIncrement = true;
			}

			if (build)
			{
				if (!(*bpIt))
				{
					continue;
				}
			}

			const key_list_type::iterator kEnd = keyIt->end();
			cIt = anyCharCounts.begin();
			for (key_list_type::iterator kIt = keyIt->begin(); kIt != kEnd; ++kIt, ++cIt)
			{
				if (*cIt > 0)
				{
					//ignore this key; we have some N's in it
					continue;
				}
				//now go ahead and add this
				key_type currentKey = *kIt;
				Seq::size_type currentLastSeqOccurrence = 0;
				if (useTempVectors)
				{
					currentLastSeqOccurrence = (*flsIt)[currentKey];
				}
				else
				{
					currentLastSeqOccurrence = (*slsIt)[currentKey];
				}

				if (currentLastSeqOccurrence != sequencePosition)
				{

					if (build)
					{
						//that means this is the first time we've seen this kmer at this position
						++(*kpIt);

						//kmerPosToSeqPos[kmerPosition] = indexPosition
						tkIt->push_back(indexPosition);

						//nextOccurrence[kmerPosition] = 0 since this is now the end
						tnoIt->push_back(0);

						Seq::size_type currentLastKmerOccurrence = 0;

						if (useTempVectors)
						{
							currentLastKmerOccurrence = (*flkIt)[currentKey];
						}
						else
						{
							currentLastKmerOccurrence = (*slkIt)[currentKey];
						}
						if (currentLastKmerOccurrence == 0)
						{
							//first time we found this kmer
							(*ffoIt)[currentKey] = *kpIt;
						}
						else
						{
							//we need to chain this
							//chain the previous last to the current
							(*tnoIt)[currentLastKmerOccurrence] = *kpIt;
						}
					}
					else
					{
						//first time we found this kmer
						seq_pos_map::iterator cEnd = (*ffiIt)[currentKey + 1];
						for (seq_pos_map::iterator cIt = (*ffiIt)[currentKey]; cIt != cEnd; ++cIt)
						{
							//subtract the first seq position because we want the seeds to be base 0
							Seq::size_type position1 = *cIt;
							Seq::size_type position2 = indexPosition;
							LocalHit::size_type seedLength = _kLength;

							bool added = scanInfo->m->_addHit(*(scanInfo->sequenceIndexed), position1, sequence, position2, seedLength);

							scanInfo->m->_totalMatchesPreExtension++;
							if (added)
							{
								scanInfo->m->_totalMatches++;
							}
						}
					}
					if (useTempVectors)
					{
						if (build)
						{
							(*fcIt)[currentKey]++;
							(*flkIt)[currentKey] = *kpIt;
						}
						(*flsIt)[currentKey] = sequencePosition;
					}
					else
					{
						if (build)
						{
							(*scIt)[currentKey]++;
							(*slkIt)[currentKey] = *kpIt;
						}
						(*slsIt)[currentKey] = sequencePosition;
					}

				}
			}
		}
		//END ADDING KEYS

	}
	//END MAIN LOOP

	if (build)
	{
		//now we need to copy from nextOccurrence to next occurrence
		Vec<unsigned int>::iterator nkEnd = buildInfo->numKmers->end();
		
		Vec<Seq::size_type>::iterator kpIt = kmerPositions.begin();
		for (Vec<unsigned int>::iterator nkIt = buildInfo->numKmers->begin(); nkIt != nkEnd; ++nkIt, ++kpIt)
		{
			*nkIt = *kpIt - FIRST_SEQ_POSITION + 1;
		}
	}
}

#endif
