#ifndef _index_h_included_
#define _index_h_included_

#include "System.h"
#include "MatchFinder.h"
#include "Mseq.h"
#include "ProbVector.h"
#include "ScoringParameters.h"
#include "LocalHit.h"
#include "Pattern.h"
#include "NextPatternSetGenerator.h"
#include <map>
#include <deque>
#include <ext/hash_map>
#include "Constants.h"
#include "Profile.h"
#include <cassert>
#include <list>
#include <queue>

/**
 * This is the main index class
 */

class Index
{

 public:

	/* Constants */
	/* ========= */
	static const Seq::size_type FIRST_SEQ_POSITION;
	static const Seq::size_type PRE_FIRST_KMER_POSITION;

	/* Types */
	/* ===== */

	typedef LocalHit::score_type score_type;

	typedef float num_devs_type;
	typedef float avg_pattern_num_type;
	typedef Vec<TyphonNS::Pattern>::size_type pattern_num_type;
	typedef float avg_occurence_num_type;
	typedef Seq::size_type occurrence_num_type;
	typedef pair<score_type, pattern_num_type> score_pattern_num_type;

	typedef pair<score_type, Seq::size_type> score_position_type;
	typedef float pattern_score_type;
	typedef pair<pattern_score_type, unsigned int> pattern_score_key_type;

	typedef score_type conservation_score;
	typedef pair<Seq::size_type, conservation_score> size_to_conservation_type; 
	typedef unsigned char seq_letter_type;

	
	/* Constructor */
	/* =========== */
	
	/**
	 * Build an Index
	 * @param sequence: The sequence from which to build the index
	 * @param numDevs: kmers occuring more frequently that numDevs
	 *                 deviations above the mean will be discarded
	 *                 Specify 0 to discard no kmers
	 * @param patternSet: The patterns that we will use for indexing/scanning
	 *                    If it contains an empty pattern, we will fill it with dynamically
	                      chosen patterns
	 * @avgNumPatternsPerPosition: the avg num of patterns to index per position
	 */

  template<class LetterPolicy> Index(const ProbVector& sequence, const RegionList& r, LetterPolicy letterPolicy, NextPatternSetGenerator& generator, const float numDevs, const Vec<TyphonNS::Pattern>& patternSet, avg_pattern_num_type avgNumPatternsPerPosition, const unsigned int order = 0);

	/**
	 * Scan an Index
	 * @param sequence1 The sequence to scan
	 * @param sequence2 The sequence to scan
	 * @param scoringParameters The parameters to use for scoring
	 * @param start2 The offset that the 
	 * @param offset: the offset in a larger sequence of this one; this is just for bookeeping	 
	 */

  template<class LetterPolicy> void scanIndex(LetterPolicy letterPolicy, Mseq& sequence1, Mseq& sequence2, MatchFinder& m, ScoringParameters<score_type>& scoringParameters, const Seq::size_type indexOffset, const Seq::size_type queryOffset);

 private:

	/* Types */
	/* ===== */

	typedef Mseq::size_type num_keys_type;
	typedef Vec<Seq::size_type>::size_type key_type;
	typedef Vec<Vec<key_type>::size_type> last_type;
	typedef Vec<key_type> key_list_type;
	typedef deque<key_type> current_key_letters_type;
	typedef Vec<current_key_letters_type> word_cache_type;
	typedef deque<score_type> score_cache_type;

	typedef Vec<Seq::size_type> seq_pos_map;
	typedef Vec<Seq::size_type> fast_direct_kmer_map;
	typedef Vec<seq_pos_map::iterator> fast_seq_pos_iterator_map;
	typedef Vec<unsigned int> fast_direct_kmer_count_map;
	typedef hash_map_type<key_type, Seq::size_type> slow_direct_kmer_map;
	typedef hash_map_type<key_type, seq_pos_map::iterator> slow_seq_pos_iterator_map;
	typedef hash_map_type<key_type, unsigned int> slow_direct_kmer_count_map;

	/**
	 * Helper method for scanning an entire sequence
	 */

	typedef struct BI
	{
		Vec<fast_direct_kmer_count_map>* fastCount;
		Vec<slow_direct_kmer_count_map>* slowCount;
		Vec<fast_direct_kmer_map>* fastFirstOccurrence;
		Vec<Vec<Seq::size_type> >* nextOccurrence;
		Vec<Vec<Seq::size_type> >* kmerPosToSeqPos;
		Vec<fast_direct_kmer_map>* fastLastKmerOccurrence;
		Vec<slow_direct_kmer_map>* slowLastKmerOccurrence;
		Vec<Seq::size_type>* kmerPositions;
		bool useTempVectors;
		Seq::size_type start;
		Seq::size_type end;
	} BuildInformation;

	typedef struct SI
	{
		Mseq* sequenceIndexed;
		Mseq* sequenceScanned;
		MatchFinder* m;
		bool useTempVectors;
	} ScanInformation;

	/**
	 * This method scans the sequence and either indexes or looks up every
	 * pattern. It uses the letter specified by the LetterPolicy
	 */

	template<class LetterPolicy> void _scanSequence(const ProbVector& sequence
																									,LetterPolicy& letterPolicy
																									,const Vec<Vec<TyphonNS::Pattern>::size_type> patternSubset
																									,BuildInformation* buildInfo
																									,ScanInformation* scanInfo
																									);
	//MEMBERS

	/**
	 * This stores how many bits we'll use for each letter in the key
	 */
	unsigned int _bitsPerLetter;

	/**
	 * Bookeeping information on memory
	 */
	unsigned int _totalMemory;
	unsigned int _maxMemory;
	unsigned int _permanentVectorsSize;
	unsigned int _numUniqueKmers;

	/**
	 * This stores information on the class of patterns that we'll be building
	 */
	pattern_num_type _numPatterns;
	Vec<TyphonNS::Pattern> _patterns;
	TyphonNS::Pattern::size_type _kLength;	
	TyphonNS::Pattern::size_type _numZeros;	
	TyphonNS::Pattern::size_type _patternSpan;

	/**
	 * We use this to implement non-consecutive patterns
	 * These pointers tell us what positions to subtract and
	 * what positions to add after each shift of a new character
	 */

	Vec<last_type> _firstZeroPositions;
	Vec<Vec<unsigned int> > _firstZeroShamt;
	Vec<last_type> _firstOnePositions;
	Vec<Vec<unsigned int> > _firstOneShamt;

	/**
	 * This is the storage structure for the index
	 */
	//this is a vector of sequence positions
	Vec<seq_pos_map> _seqPosList;
	//for each kmer values, this points to the starting point in the seq pos list
	//the kmervalue + 1 points to one beyond the ending position
	Vec<fast_seq_pos_iterator_map> _fastFirstIterator;

};

template<class LetterPolicy> Index::Index(const ProbVector& sequence, const RegionList& regionList, LetterPolicy letterPolicy, NextPatternSetGenerator& generator, const float numDevs, const Vec<TyphonNS::Pattern>& patternSet, avg_pattern_num_type avgNumPatternsPerPosition, const unsigned int order)
{
	time_t totalSeconds = time(NULL);

	//determine whether we need to pick a new pattern set or not
	_numPatterns = patternSet.size();
	if (_numPatterns == 0)
	{
		return;
	}
	else
	{
		_numPatterns = patternSet.size();
		_patterns.clear();
		_patterns.resize(_numPatterns);
		copy(patternSet.begin(), patternSet.end(), _patterns.begin());

		TyphonNS::Pattern::size_type maxKLength = 0;
		for (Vec<TyphonNS::Pattern>::iterator it = _patterns.begin(); it != _patterns.end(); ++it)
		{
			if (it->getKLength() > maxKLength)
			{
				maxKLength = it->getKLength();
			}
		}

		if (maxKLength == 0)
		{
			_patterns.clear();
		}
		else
		{
			for (Vec<TyphonNS::Pattern>::iterator it = _patterns.begin(); it != _patterns.end(); ++it)
			{
				if (it->getKLength() < maxKLength)
				{
					it->pad(maxKLength - it->getKLength(), true);
				}
			}
			//now all have the same weight.  We need to ensure all have the same span
			TyphonNS::Pattern::size_type maxPatternSpan = 0;
			for (Vec<TyphonNS::Pattern>::iterator it = _patterns.begin(); it != _patterns.end(); ++it)
			{
				if (it->getPatternLength() > maxPatternSpan)
				{
					maxPatternSpan = it->getPatternLength();
				}
			}

			for (Vec<TyphonNS::Pattern>::iterator it = _patterns.begin(); it != _patterns.end(); ++it)
			{
				if (it->getPatternLength() < maxPatternSpan)
				{
					it->pad(maxKLength - it->getKLength(), false);
				}
			}
			_kLength = maxKLength;
			_patternSpan = maxPatternSpan;
			_numZeros = _patternSpan - _kLength;
		}
	}

	//set up our data structures
	_numUniqueKmers = static_cast<unsigned int>(pow(static_cast<float>(BIO::NUM_NUCLEOTIDES), static_cast<float>(_kLength)));

	//determine whether we'll use vectors or hash maps for our internal and temporary data structures
	bool useTempVectors = true;
	{
		//determine the memory we'll assume
		_totalMemory = 500000000;
		_maxMemory = _totalMemory / 2;

		_permanentVectorsSize = _numPatterns * sizeof(Seq::size_type) * _numUniqueKmers;
		//temp vectors are a count for each kmer in each index, a total count for each kmer for each pattern, a last sequence position, and a last kmer position
		unsigned int numCounts = 1;

		unsigned int totalTempVectorsSize = numCounts * _numPatterns * sizeof(unsigned int) * _numUniqueKmers + sizeof(Seq::size_type) * _numUniqueKmers * 2 * _numPatterns;

		//determine whether to use vectors or hash maps
		if (_permanentVectorsSize + totalTempVectorsSize < _maxMemory)
		{
			useTempVectors = true;
		}
		else
		{
			useTempVectors = false;
		}		

		//DEBUG MESSAGE
		ostringstream message;
		message << "Assuming total memory of " << _totalMemory / 1000000<< "MB";
		message << endl << "Debug: Using vectors for internal storage";

		if (useTempVectors)
		{
			message << endl << "Debug: Using vectors for temporary storage";
		}
		else
		{
			message << endl << "Debug: Using hash tables for temporary storage";
		}
		//cerr << "Debug: " << message.str() << endl;
	}

	_bitsPerLetter = static_cast<unsigned int>(ceil(log(static_cast<double>(BIO::NUM_NUCLEOTIDES)) / M_LN2));
	//make sure that we won't overflow our keys
	if (_bitsPerLetter * _kLength >= sizeof(key_type) * CHAR_BIT)
	{
		throw BIO::E_BioEx("W is too big");
	}
	
	//go about actually building the index

	//BEGIN INITIALIZATION

	//required for converting letters to numbers
	initialize_prof_idx();

	if (avgNumPatternsPerPosition > _numPatterns)
	{
		avgNumPatternsPerPosition = _numPatterns;
	}

	//store the patterns in a better format
	//now set up the positions that we'll need to track
	_firstZeroPositions.resize(_numPatterns);
	_firstZeroShamt.resize(_numPatterns);
	_firstOnePositions.resize(_numPatterns);
	_firstOneShamt.resize(_numPatterns);
	
	Vec<TyphonNS::Pattern>::const_iterator pEnd = _patterns.end();
	Vec<last_type>::iterator fzpIt = _firstZeroPositions.begin();
	Vec<Vec<unsigned int> >::iterator fzsIt = _firstZeroShamt.begin();
	Vec<last_type>::iterator fopIt = _firstOnePositions.begin();
	Vec<Vec<unsigned int> >::iterator fosIt = _firstOneShamt.begin();
	for (Vec<TyphonNS::Pattern>::const_iterator pIt = _patterns.begin(); pIt != pEnd; ++pIt, ++fzpIt, ++fzsIt, ++fopIt, ++fosIt)
	{
		bool seenOne = false;
		//the position off the end is a zero
		bool seenZero = true;
		unsigned int shamt = _kLength;
		for (TyphonNS::Pattern::size_type i = 0; i < _patternSpan; i++)
		{
			if ((*pIt)[i])
			{
				--shamt;
				//if we were just in a region of zeros, the last position is the last zero
				if (seenZero)
				{
					fopIt->push_back(i);
					fosIt->push_back(shamt * _bitsPerLetter);
				}
				seenOne = true;
				seenZero = false;
			}
			else
			{
				//if we were just in a region of ones, the last position is a last one
				if (seenOne)
				{
					fzpIt->push_back(i);
					fzsIt->push_back((shamt) * _bitsPerLetter);
				}
				seenZero = true;
				seenOne = false;
			}
		}
		if (seenOne)
		{
			fzpIt->push_back(_patternSpan);
			fzsIt->push_back(0);
		}
	}

	Vec<unsigned int> numKmers(_numPatterns, 0);
	//this is temporary; store the counts of each kmer
	Vec<fast_direct_kmer_count_map> fastCount(_numPatterns);
	Vec<slow_direct_kmer_count_map> slowCount(_numPatterns);

	//we make a temp because we don't know the size a priori
	//as it grows, it will have too much memory
	//so we want to initialize the actual one to trunc the un-needed memory

	//this stores the first occurrence of each kmer for each pattern
	Vec<fast_direct_kmer_map> fastFirstOccurrence(_numPatterns);
	//this is a linked list; each kmer position points to the next kmer position
	//at which the same kmer occurred
	Vec<Vec<Seq::size_type> > nextOccurrence(_numPatterns);
	//this maps the kmer position to the sequence position at which it occurred
	Vec<Vec<Seq::size_type> > kmerPosToSeqPos(_numPatterns);
	//this stores the last kmer occurrence for each kmer value
	Vec<fast_direct_kmer_map> fastLastKmerOccurrence;
	Vec<slow_direct_kmer_map> slowLastKmerOccurrence;

	//this stores the position in the larger space of kmers
	//at each sequence position we may index more than one kmer
	Vec<Seq::size_type> kmerPositions(_numPatterns, PRE_FIRST_KMER_POSITION);

	//resize all of the temps
	Vec<fast_direct_kmer_map>::iterator fEnd = fastFirstOccurrence.end();
	for(Vec<fast_direct_kmer_map>::iterator fIt = fastFirstOccurrence.begin(); fIt != fEnd; ++fIt)
	{
		fIt->resize(_numUniqueKmers, 0);
	}
	
	if (useTempVectors)
	{
		fastCount.resize(_numPatterns);
		Vec<fast_direct_kmer_count_map>::iterator fEnd = fastCount.end();
		for(Vec<fast_direct_kmer_count_map>::iterator fIt = fastCount.begin(); fIt != fEnd; ++fIt)
		{
			fIt->resize(_numUniqueKmers, 0);
		}

		fastLastKmerOccurrence.resize(_numPatterns);
		Vec<fast_direct_kmer_map>::iterator fkEnd = fastLastKmerOccurrence.end();
		for (Vec<fast_direct_kmer_map>::iterator fkIt = fastLastKmerOccurrence.begin(); fkIt != fkEnd; ++fkIt)
		{
			fkIt->resize(_numUniqueKmers, 0);
		}
	}
	else
	{
		slowCount.resize(_numPatterns);
		slowLastKmerOccurrence.resize(_numPatterns);
	}
	
	//it will need at least this much
	Vec<Seq::size_type>::size_type minSize = sequence.size() / _numPatterns;
	Vec<Vec<Seq::size_type> >::iterator tEnd = nextOccurrence.end();
	Vec<Vec<Seq::size_type> >::iterator kIt = kmerPosToSeqPos.begin();
	for (Vec<Vec<Seq::size_type> >::iterator tIt = nextOccurrence.begin(); tIt != tEnd; ++tIt, ++kIt)
	{
		tIt->reserve(minSize);
		kIt->reserve(minSize);
		while (tIt->size() <= PRE_FIRST_KMER_POSITION)
		{
			tIt->push_back(0);
			//since the first kmer position is 1, all before it map to zero
			kIt->push_back(0);
		}
	}
	
	//this is a utility to pass the parameters to the helper function
	BuildInformation buildInformation;
	buildInformation.fastCount = &fastCount;
	buildInformation.slowCount = &slowCount;
  buildInformation.fastFirstOccurrence = &fastFirstOccurrence;
  buildInformation.nextOccurrence = &nextOccurrence;
  buildInformation.kmerPosToSeqPos = &kmerPosToSeqPos;
  buildInformation.useTempVectors = useTempVectors;
	buildInformation.fastLastKmerOccurrence = &fastLastKmerOccurrence;
	buildInformation.slowLastKmerOccurrence = &slowLastKmerOccurrence;
	buildInformation.kmerPositions = &kmerPositions;

	//use our generator
	generator.initialize(sequence, regionList, patternSet, static_cast<unsigned int>(avgNumPatternsPerPosition * sequence.size()), order);

	while (generator.hasMoreBudget())
	{
		Region nextRegion;
		Vec<Vec<TyphonNS::Pattern>::size_type> patternSubset;
		generator.nextPatternSet(nextRegion, patternSubset);
		
		buildInformation.start = nextRegion.getStart();
		buildInformation.end = nextRegion.getEnd();

		_scanSequence(sequence, letterPolicy, patternSubset, &buildInformation, NULL);
	}

	//now we need to copy from nextOccurrence to next occurrence
	for (Vec<unsigned int>::size_type i = 0; i != _numPatterns; i++)
	{
		numKmers[i] = kmerPositions[i] - PRE_FIRST_KMER_POSITION;
	}

	//BEGIN POST PROCESSING

	if (numDevs)
	{
		time_t seconds = time(NULL);

		Vec<unsigned int>::iterator nkEnd = numKmers.end();
		Vec<fast_direct_kmer_count_map>::iterator fcIt = fastCount.begin();
		Vec<slow_direct_kmer_count_map>::iterator scIt = slowCount.begin();
		Vec<fast_direct_kmer_map>::iterator ffoIt = fastFirstOccurrence.begin();

		for (Vec<unsigned int>::iterator nkIt = numKmers.begin(); nkIt != nkEnd; ++nkIt, ++fcIt, ++scIt, ++ffoIt)
		{

			const num_devs_type mean = static_cast<float>(*nkIt) / _numUniqueKmers;

			num_devs_type stddev = 0;
			num_devs_type totalNum = 0;

			if (useTempVectors)
			{
				//iterate through the first positions and clear out the frequent ones
				const fast_direct_kmer_count_map::iterator cEnd = fcIt->end();
				for (fast_direct_kmer_count_map::iterator cIt = fcIt->begin(); cIt != cEnd; ++cIt)
				{
					if (*cIt > 0)
					{
						num_devs_type diff = *cIt - mean;
						stddev += diff * diff;
						totalNum++;
					}
				}
			}
			else
			{
				const slow_direct_kmer_count_map::iterator cEnd = scIt->end();
				for (slow_direct_kmer_count_map::iterator cIt = scIt->begin(); cIt != cEnd; ++cIt)
				{
					if (cIt->second > 0)
					{
						num_devs_type diff = cIt->second - mean;
						stddev += diff * diff;
						totalNum++;
					}
				}
			}

			stddev += mean * mean * (_numUniqueKmers - totalNum);
			stddev /= _numUniqueKmers;

			//sqrt doesn't give correct performance when optimized for some reason
			//stddev = sqrt(stddev);
			stddev = ceil(pow(stddev, static_cast<num_devs_type>(.5)));

			//DEBUG COMMENT
			ostringstream message("");
			message << "Average of " << mean << " positions per kmer; " << _numUniqueKmers << " unique but " << totalNum << " appear";

			//DEBUG COMMENT
			message << endl << "Debug: Standard deviation is " << stddev;

			unsigned int numAbove = 0;
			const float threshold = numDevs * stddev;

			if (useTempVectors)
			{
				//iterate through the first positions and clear out the frequent ones
				fast_direct_kmer_map::iterator fIt = ffoIt->begin();

				const fast_direct_kmer_count_map::iterator cEnd = fcIt->end();
				unsigned int temp = 0;
				for (fast_direct_kmer_count_map::iterator cIt = fcIt->begin(); cIt != cEnd; cIt++, fIt++, temp++)
				{
					if (*fIt)
					{
						const float diff = *cIt - mean;
						if (diff > threshold)
						{
							numAbove++;
							*fIt = 0;
						}
					}
				}
			}
			else
			{
				const slow_direct_kmer_count_map::iterator cEnd = scIt->end();
				for (slow_direct_kmer_count_map::iterator cIt = scIt->begin(); cIt != cEnd; ++cIt)
				{
					if (cIt->second > 0)
					{
						const float diff = cIt->second - mean;
						if (diff > threshold)
						{
							numAbove++;
							(*ffoIt)[cIt->first] = 0;
						}
					}
				}
			}

			//DEBUG MESSAGE
			message << endl << "Debug: " << numAbove << " are above " << numDevs << " standard deviations";
			
			//cerr << "Debug: " << message.str() << endl;
		}

		{
			ostringstream message("");
			message << "Post processing took " << time(NULL) - seconds << " seconds";
			//cout << message.str() << endl;
		}
	}
	
	//END POST PROCESSING

	//BEGIN TRANSFER TO FINAL STORAGE
	_seqPosList.resize(_numPatterns);
	_fastFirstIterator.resize(_numPatterns);

	Vec<fast_direct_kmer_map>::iterator ffoEnd = fastFirstOccurrence.end();
	Vec<fast_direct_kmer_map>::iterator ffoIt = fastFirstOccurrence.begin();
	Vec<Vec<Seq::size_type> >::iterator tkIt = kmerPosToSeqPos.begin();
	Vec<Vec<Seq::size_type> >::iterator tnoIt = nextOccurrence.begin();
	Vec<seq_pos_map>::iterator spIt = _seqPosList.begin();
	Vec<fast_seq_pos_iterator_map>::iterator ffiIt = _fastFirstIterator.begin();
	for (; ffoIt != ffoEnd; ++ffoIt, ++tnoIt, ++spIt, ++ffiIt, ++tkIt)
	{
		Seq::size_type seqPosListSize = 0;

		fast_direct_kmer_map::reverse_iterator dIt = ffoIt->rbegin();
		fast_direct_kmer_map::reverse_iterator dEnd = ffoIt->rend();
		for(; dIt != dEnd; ++dIt)
		{
			for (Seq::size_type currentKmerPosition = *dIt; currentKmerPosition != 0; currentKmerPosition = (*tnoIt)[currentKmerPosition])		
			{
				seqPosListSize++;
			}
		}
		spIt->resize(seqPosListSize);
		seq_pos_map::iterator currentIterator = spIt->end();

		ffiIt->resize(ffoIt->size() + 1);
		//go in reverse order
		fast_seq_pos_iterator_map::reverse_iterator fIt = ffiIt->rbegin();
		dIt = ffoIt->rbegin();

		for(; dIt != dEnd; ++dIt, ++fIt)
		{
			*fIt = currentIterator;
			for (Seq::size_type currentKmerPosition = *dIt; currentKmerPosition != 0; currentKmerPosition = (*tnoIt)[currentKmerPosition])		
			{
				currentIterator--;
				*currentIterator = (*tkIt)[currentKmerPosition];
			}
		}
		*fIt = currentIterator;

		assert(currentIterator == spIt->begin());
	}

	//END TRANSFER TO FINAL STORAGE

	{
		ostringstream message("");
		unsigned int totalSize = 0;
		for (Vec<Vec<Seq::size_type> >::iterator it = nextOccurrence.begin(); it != nextOccurrence.end(); ++it)
		{
			totalSize += it->size();
		}
		message << "Debug: Total size is " << totalSize;
		//cerr << message.str() << endl;
	}	

	{
		ostringstream message("");
		message << "Built index on " << sequence.size() - _patternSpan + 1 << " kmers; took " << time(NULL) - totalSeconds << " seconds";
		//cout << message.str() << endl;
	}
}

template<class LetterPolicy> void Index::scanIndex(LetterPolicy letterPolicy, Mseq& sequence1, Mseq& sequence2, MatchFinder& m, ScoringParameters<score_type>& scoringParameters, const Seq::size_type indexOffset, const Seq::size_type queryOffset)
{
	m.initialize(scoringParameters);
	m._clearStats();
	//determine whether we'll use vectors or hash maps for our internal and temporary data structures
	bool useTempVectors = true;
	{
		unsigned int maxTempMemory = _maxMemory > _permanentVectorsSize ? _maxMemory - _permanentVectorsSize : 0;

		unsigned int totalTempVectorsSize = _numPatterns * sizeof(Seq::size_type) * _numUniqueKmers;

		//determine whether to use vectors or hash maps
		if (totalTempVectorsSize < maxTempMemory)
		{
			useTempVectors = true;
		}
		else
		{
			useTempVectors = false;
		}

		//DEBUG MESSAGE
		ostringstream message;
		message << "Assuming total memory of " << _totalMemory << "B";
		if (useTempVectors)
		{
			message << endl << "Debug: Using vectors for temporary storage";
		}
		else
		{
			message << endl << "Debug: Using hash tables for temporary storage";
		}
		//cerr << "Debug: " << message.str() << endl;

	}

	//set up the start offsets
	m._start1Offset = indexOffset;
	m._start2Offset = queryOffset;

	time_t seconds = time(NULL);
	
	//get all of the matches

	ScanInformation scanInformation;
	scanInformation.sequenceIndexed = &sequence1;
	scanInformation.sequenceScanned = &sequence2;
	scanInformation.m = &m;
	scanInformation.useTempVectors = useTempVectors;

	//for now, use them all
	Vec<Vec<TyphonNS::Pattern>::size_type> patternSubset;
	for (Vec<TyphonNS::Pattern>::size_type i = 0; i < _patterns.size(); i++)
	{
		patternSubset.push_back(i);
	}

	//we need to create a prob vector
	ProbVector pv(sequence2);
	_scanSequence(sequence2, letterPolicy, patternSubset, NULL, &scanInformation);

	ostringstream message("");
	message << "Getting matches took " << time(NULL) - seconds << " seconds";
	//DEBUG MESSAGE
	//cout << message.str() << endl;
}

template<class LetterPolicy> void Index::_scanSequence(const ProbVector& sequence
																											 ,LetterPolicy& letterPolicy
																											 ,const Vec<Vec<TyphonNS::Pattern>::size_type> patternSubset
																											 ,BuildInformation* buildInfo
																											 ,ScanInformation* scanInfo
																											 )
{
	bool build = true;
	bool useTempVectors = false;

	if (buildInfo == NULL)
	{
		build = false;
		useTempVectors = scanInfo->useTempVectors;
	}
	else
	{
		useTempVectors = buildInfo->useTempVectors;
	}

	if (build)
	{
		if (buildInfo->end - buildInfo->start + 1 < _patternSpan)
		{
			return;
		}
	}
	else
	{
		if (sequence.size() < _patternSpan)
		{
			return;
		}
	}

	//initialize the pattern information
	Vec<TyphonNS::Pattern>::size_type numPatterns = patternSubset.size();

	//now set up the code needed for reading in the actual letters

	//we store however many keys we'll need for each word
	num_keys_type numKeys = letterPolicy.getNumKeys(sequence);

	//set up the key cache; one key set for each pattern
	Vec<key_list_type> keyCache(numPatterns);

	//each pattern needs a set of keys (how many are determined by the letter policy)
	Vec<key_list_type>::iterator kEnd = keyCache.end();
	for (Vec<key_list_type>::iterator kIt = keyCache.begin(); kIt != kEnd; ++kIt)
	{
		kIt->resize(numKeys, 0);
	}

	//these will cache our current words
	word_cache_type wordCache(numKeys);

	//these will store whether we have an 'N' in this position

	//END INITIALIZATION

	//BEGIN RAMPUP CODE

	//now, we begin actually reading from the sequence

	//the first thing that we do is to do some rampup code

	//push an extra zero to avoid a special case for the main loop
	const word_cache_type::iterator wEnd = wordCache.end();
	word_cache_type::iterator wIt = wordCache.begin();
	for (; wIt != wEnd; ++wIt)
	{
		for (TyphonNS::Pattern::size_type j = 0; j < _patternSpan; j++)
		{
			wIt->push_back(0);
		}
	}

	//the actual end of the sequence
	ProbVector::size_type actualEndIter = sequence.size();
	if (build)
	{
		actualEndIter = buildInfo->end + 1;
	}

	//we end this phase at the end of the pattern
	ProbVector::size_type rampupEnd = 0;
	if (build)
	{
		rampupEnd += buildInfo->start;
	}
	rampupEnd += _patternSpan - 1;
	
	//now our rampup code reads the first patternSpan - 1 letters into the caches
	//this is stored for subsequent calls to _nextKmers()
	ProbVector::size_type iter = 0;
	if (build)
	{
		iter += buildInfo->start;
	}

	for (; iter != rampupEnd && iter != actualEndIter; ++iter)
	{
		//BEGIN GETTING LETTERS

		//we store the consensus character
		const word_cache_type::iterator wEnd = wordCache.end();		
		for (word_cache_type::iterator wIt = wordCache.begin(); wIt != wEnd; ++wIt)
		{
			key_type adding = letterPolicy.getLetter(sequence, iter);
			wIt->push_back(adding);
		}

		//END GETTING LETTERS

		//BEGIN UPDATING KEYS

		//we need to update the keys for each pattern
		for (Vec<last_type>::size_type i = 0; i != numPatterns; ++i)
		{
			Vec<last_type>::size_type curPattern = patternSubset[i];
			key_list_type::iterator kIt = keyCache[i].begin();
			for (word_cache_type::iterator wIt = wordCache.begin(); wIt != wEnd; ++wIt, ++kIt)
			{
				//the first thing we need to do is to subtract the letters vanishing after this shift
				const last_type::const_iterator oneEnd = _firstOnePositions[curPattern].end();
				const Vec<unsigned int>::const_iterator oneShamtEnd = _firstOneShamt[curPattern].end();
				last_type::const_iterator oneIt = _firstOnePositions[curPattern].begin();
				Vec<unsigned int>::const_iterator oneShamIt = _firstOneShamt[curPattern].begin();

				for (; oneIt != oneEnd; oneIt++, oneShamIt++)
				{
					*kIt -= ((*wIt)[*oneIt] << (*oneShamIt));
				}

				//shift everything over by one
				//originalKey = originalKey * NUM_NUCLEOTIDES
				*kIt <<= _bitsPerLetter;

				//now we need to add the letters appearing after this shift
				const last_type::const_iterator zeroEnd = _firstZeroPositions[curPattern].end();
				const Vec<unsigned int>::const_iterator zeroShamtEnd = _firstZeroShamt[curPattern].end();
				last_type::const_iterator zeroIt = _firstZeroPositions[curPattern].begin();
				Vec<unsigned int>::const_iterator zeroShamIt = _firstZeroShamt[curPattern].begin();

				for (; zeroIt != zeroEnd; zeroIt++, zeroShamIt++)
				{
					*kIt += ((*wIt)[*zeroIt] << (*zeroShamIt));
				}
			}
		}

		//END UPDATING KEYS

		for (word_cache_type::iterator wIt = wordCache.begin(); wIt != wEnd; ++wIt)
		{
			wIt->pop_front();
		}
	}

	//to avoid a special case in the main loopa
	--iter;
	//since we increment iter at the beginning of each iteration
	--actualEndIter;

	//END RAMPUP CODE
	
	Vec<fast_direct_kmer_map> fastLastSeqOccurrence;
	Vec<slow_direct_kmer_map> slowLastSeqOccurrence;
	fastLastSeqOccurrence.resize(numPatterns);
	slowLastSeqOccurrence.resize(numPatterns);

	//this exists solely for checking duplicates in the kmers we're indexing
	if (useTempVectors)
	{
		Vec<fast_direct_kmer_map>::iterator fsEnd = fastLastSeqOccurrence.end();
		for (Vec<fast_direct_kmer_map>::iterator fsIt = fastLastSeqOccurrence.begin(); fsIt != fsEnd; ++fsIt)
		{
			fsIt->resize(_numUniqueKmers, 0);
		}
	}

	//BEGIN MAIN LOOP

	//this stores the position in the sequence
	//store base 0 since we index from base 0
	Seq::size_type indexPosition = 0;
	Seq::size_type sequencePosition = FIRST_SEQ_POSITION;

	if (build)
	{
		indexPosition += buildInfo->start;
		sequencePosition += buildInfo->start;
	}
	
	for (; iter != actualEndIter; ++indexPosition, ++sequencePosition)
	{
		if (sequencePosition % 100000 == 0)
		{
			
			ostringstream message;
			message << "Fetching kmers at position " << sequencePosition;
			//DEBUG COMMENT
			//cerr << "Debug: " << message.str() << endl;
		}

		++iter;

		//BEGIN GETTING LETTERS

		//we store the consensus character
		const word_cache_type::iterator wEnd = wordCache.end();		
		for (word_cache_type::iterator wIt = wordCache.begin(); wIt != wEnd; ++wIt)
		{
			key_type adding = letterPolicy.getLetter(sequence, iter);
			wIt->push_back(adding);
		}

		//END GETTING LETTERS

		//BEGIN UPDATING KEYS

		//we need to update the keys for each pattern
		for (Vec<last_type>::size_type i = 0; i != numPatterns; ++i)
		{
			Vec<last_type>::size_type curPattern = patternSubset[i];
			key_list_type::iterator kIt = keyCache[i].begin();	
			for (word_cache_type::iterator wIt = wordCache.begin(); wIt != wEnd; ++wIt, ++kIt)
			{
				//the first thing we need to do is to subtract the letters vanishing after this shift
				const last_type::const_iterator oneEnd = _firstOnePositions[curPattern].end();
				const Vec<unsigned int>::const_iterator oneShamtEnd = _firstOneShamt[curPattern].end();
				last_type::const_iterator oneIt = _firstOnePositions[curPattern].begin();
				Vec<unsigned int>::const_iterator oneShamIt = _firstOneShamt[curPattern].begin();

				for (; oneIt != oneEnd; oneIt++, oneShamIt++)
				{
					*kIt -= ((*wIt)[*oneIt] << (*oneShamIt));
				}

				//shift everything over by one
				//originalKey = originalKey * NUM_NUCLEOTIDES
				*kIt <<= _bitsPerLetter;

				//now we need to add the letters appearing after this shift
				const last_type::const_iterator zeroEnd = _firstZeroPositions[curPattern].end();
				const Vec<unsigned int>::const_iterator zeroShamtEnd = _firstZeroShamt[curPattern].end();
				last_type::const_iterator zeroIt = _firstZeroPositions[curPattern].begin();
				Vec<unsigned int>::const_iterator zeroShamIt = _firstZeroShamt[curPattern].begin();

				for (; zeroIt != zeroEnd; zeroIt++, zeroShamIt++)
				{
					*kIt += ((*wIt)[*zeroIt] << (*zeroShamIt));
				}
			}
		}

		//END UPDATING KEYS

		for (word_cache_type::iterator wIt = wordCache.begin(); wIt != wEnd; ++wIt)
		{
			wIt->pop_front();
		}

		//BEGIN PROCESSING KEYS

		for (Vec<key_list_type>::size_type j = 0; j != numPatterns; ++j)
		{
			Vec<key_list_type>::size_type curPattern = patternSubset[j];
			const key_list_type::iterator kEnd = keyCache[j].end();
			for (key_list_type::iterator kIt = keyCache[j].begin(); kIt != kEnd; ++kIt)
			{
				//now go ahead and add this
				key_type currentKey = *kIt;
				Seq::size_type currentLastSeqOccurrence = 0;
				if (useTempVectors)
				{
					currentLastSeqOccurrence = fastLastSeqOccurrence[j][currentKey];
				}
				else
				{
					currentLastSeqOccurrence = slowLastSeqOccurrence[j][currentKey];
				}

				if (currentLastSeqOccurrence != sequencePosition)
				{
					if (build)
					{
						//that means this is the first time we've seen this kmer at this position
						++(*buildInfo->kmerPositions)[curPattern];

						//kmerPosToSeqPos[kmerPosition] = indexPosition
						(*buildInfo->kmerPosToSeqPos)[curPattern].push_back(indexPosition);

						//nextOccurrence[kmerPosition] = 0 since this is now the end
						(*buildInfo->nextOccurrence)[curPattern].push_back(0);

						Seq::size_type currentLastKmerOccurrence = 0;

						if (useTempVectors)
						{
							currentLastKmerOccurrence = (*buildInfo->fastLastKmerOccurrence)[curPattern][currentKey];
						}
						else
						{
							currentLastKmerOccurrence = (*buildInfo->slowLastKmerOccurrence)[curPattern][currentKey];
						}
						if (currentLastKmerOccurrence == PRE_FIRST_KMER_POSITION)
						{
							//first time we found this kmer
							(*buildInfo->fastFirstOccurrence)[curPattern][currentKey] = (*buildInfo->kmerPositions)[curPattern];
						}
						else
						{
							//we need to chain this
							//chain the previous last to the current
							(*buildInfo->nextOccurrence)[curPattern][currentLastKmerOccurrence] = (*buildInfo->kmerPositions)[curPattern];
						}
						if (useTempVectors)
						{
							(*buildInfo->fastCount)[curPattern][currentKey]++;
							(*buildInfo->fastLastKmerOccurrence)[curPattern][currentKey] = (*buildInfo->kmerPositions)[curPattern];
						}
						else
						{
							(*buildInfo->slowCount)[curPattern][currentKey]++;
							(*buildInfo->slowLastKmerOccurrence)[curPattern][currentKey] = (*buildInfo->kmerPositions)[curPattern];
						}
					}
					else
					{

						//first time we found this kmer
						seq_pos_map::iterator cEnd = _fastFirstIterator[curPattern][currentKey + 1];
						for (seq_pos_map::iterator cIt = _fastFirstIterator[curPattern][currentKey]; cIt != cEnd; ++cIt)
						{
							//subtract the first seq position because we want the seeds to be base 0
							Seq::size_type position1 = *cIt;
							Seq::size_type position2 = indexPosition;

							LocalHit::size_type seedLength = _kLength;

							bool added = scanInfo->m->_addHit(*(scanInfo->sequenceIndexed), position1, *(scanInfo->sequenceScanned), position2, seedLength);

							scanInfo->m->_totalMatchesPreExtension++;
							if (added)
							{
								scanInfo->m->_totalMatches++;
							}
						}
					}
					if (useTempVectors)
					{
						fastLastSeqOccurrence[j][currentKey] = sequencePosition;
					}
					else
					{
						slowLastSeqOccurrence[j][currentKey] = sequencePosition;
					}
				}
			}
			//END ADDING KEYS
		}
	}
	//END MAIN LOOP
}

#endif
