#ifndef _localhit_h_included_
#define _localhit_h_included_

#include <iostream>
#include "System.h"
#include "Seq.h"
#include "Mseq.h"
#include "MIter.h"
#include "Profile.h"
#include "ScoringParameters.h"

/**
 * The LocalHit class represents small ungapped alignments
 * It stores the positions that the two alignments match and provides holders
 * for information about the match, as well as utilites for extending the 
 * hit
 */

class LocalHit
{
 public:

	/* Types */
	/* ===== */

	typedef Seq::size_type size_type;
	typedef int diagonal_type;
	typedef int score_type;
	typedef ScoringParameters<score_type>::ScoringMatrix matrix_type;

	static const size_type OFFSET = 1;

	enum OutputType { LEVEL1, LEVEL2, LEVEL3 };

	/* Constructors */
	/* ============ */

	/*
	 * Construct a local hit of length l.  p1 is the position in 
	 * the first sequence, p2 is the position in the second sequence
	 */
	LocalHit(size_type p1, size_type p2, size_type l, score_type initialScore) : _pos1(p1), _pos2(p2), _diagonal(p2 - p1), _length(l), _score(initialScore), _rightLength(0), _leftLength(0) {}

	/* Methods */
	/* ======= */

	/**
	 * Get the _diagonal this hit lies on
	 */
  diagonal_type getDiagonal() const { return _diagonal; }
	/**
	 * Get the starting position in sequence1
	 */
  size_type getPos1Begin() const { return _pos1; }
	/**
	 * Get the last position in sequence1
	 */
  size_type getPos1End() const { return _pos1 + getLength() - 1; }
	/**
	 * Get the starting position in sequence2
	 */
  size_type getPos2Begin() const { return _pos2; }
	/**
	 * Get the last position in sequence2
	 */
  size_type getPos2End() const { return _pos2 + getLength() - 1; }
	/**
	 * Get the length of the hit
	 */
  size_type getLength() const { return _length + _rightLength + _leftLength; }
	/**
	 * Get the score of the hit
	 */
	score_type getScore() const { return _score; }
	/**
	 * Set the score of the hit
	 */
	void setScore(score_type score) { _score = score; }
	/**
	 * Add to the score of the hit
	 */
	void addToScore(score_type score) { _score += score; }
	/**
	 * Extend this hit to the right
	 */
  void extendRight() { _rightLength++;}

	/**
	 * Inverse right extension
	 */
  void undoLastExtendRight() { _rightLength--;}

	/**
	 * Inverse right extension
	 */
  void undoExtendRight(Seq::size_type amt) { _rightLength -= amt;}

	/**
	 * Extend this hit to the left
	 */
  void extendLeft() { _pos1--; _pos2--; _leftLength++; }

	/**
	 * Inverse left extension
	 */
  void undoLastExtendLeft() { _pos1++; _pos2++; _leftLength--; }

	/**
	 * Inverse right extension
	 */
  void undoExtendLeft(Seq::size_type amt) { _pos1 += amt; _pos2 += amt; _leftLength -= amt; }

	/**
	 * Extend this hit to the right
	 */
  void extendRight(size_type n) { _rightLength += n; }
	/**
	 * Extend this hit to the left
	 */
  void extendLeft(size_type n) { _pos1 -= n; _pos2 -= n; _leftLength += n; }
	/**
	 * Reset the score of the hit
	 */
	void clearScore() { _score = 0; }

	/**
	 * Translate this hit by offset1 in the sequence1 direction
	 * and by offset2 in the sequence2 direction
	 */
	//void translate(size_type offset1, size_type offset2) { _pos1 += offset1; _pos2 += offset2; }

	/**
	 * Tests whether this completely contains the query
	 */
	bool contains(const LocalHit& h2) const
		{
			if (h2.getDiagonal() != _diagonal)
			{
				return false;
			}
			return (getPos1Begin() <= h2.getPos1Begin() && getPos1End() >= h2.getPos1End());
		}

	void writeOut(const OutputType level, ostream& outStream, Mseq& sequence1, Mseq& sequence2, const unsigned int off1, const unsigned int off2, const matrix_type& matrix) const
		{
			if (level == LEVEL1)
			{
				outStream << getPos1Begin() + OFFSET + off1 << " " << getPos1End() + OFFSET + off1 << " " << getPos2Begin() + OFFSET + off2 << " " << getPos2End() + OFFSET + off2 << " + " << getScore();
			}
			else if (level == LEVEL2)
			{

				unsigned int percentIdentity = static_cast<unsigned int>(100 * getNumberOfIdentities(sequence1, sequence2, matrix) / getLength());
				outStream << "\tl " << getPos1Begin() + OFFSET + off1 << " " << getPos2Begin() + OFFSET + off2 << " " << getPos1End() + OFFSET + off1 << " " << getPos2End() + off2 << " " << percentIdentity << endl;
			}
		}

	void writeBoundary(const bool begin, const unsigned int off1, const unsigned int off2, ostream& outStream) const
		{
			if (begin)
			{
				outStream << "\tb " << getPos1Begin() + OFFSET + off1 << " " << getPos2Begin() + OFFSET + off2 << endl;
			}
			else
			{
				outStream << "\te " << getPos1End() + OFFSET + off1 << " " << getPos2End() + OFFSET + off2 << endl;
			}
		}

	size_type getNumberOfIdentities(Mseq& sequence1, Mseq& sequence2, const matrix_type& matrix, const bool identities = true) const
		{
			//iterators in the sequence
			MIter it1;
			sequence1.getMIBegin(it1);
			it1 = it1 + getPos1Begin();
			MIter end1;
			sequence1.getMIBegin(end1);
			end1 = end1 + getPos1End() + 1;			
			MIter it2;
			sequence2.getMIBegin(it2);
			it2 = it2 + getPos2Begin();

			Seq::size_type numIdentities = 0;

			for(; it1 != end1; ++it1, ++it2)
			{
				if (identities)
				{
					if (it1.getConsensusChar() == it2.getConsensusChar())
					{
						numIdentities++;
					}
				}
				else
				{
					matrix_type::size_type pos1 = CHAR_PROFILE_IDX[static_cast<int>(static_cast<unsigned char>(it1.getConsensusChar()))];
					matrix_type::size_type pos2 = CHAR_PROFILE_IDX[static_cast<int>(static_cast<unsigned char>(it2.getConsensusChar()))];
					if (matrix[pos1][pos2] > 0)
					{
						numIdentities++;
					}
				}
			}
			return numIdentities;
		}


	/* Operators */
	/* ========= */

	bool operator<(const LocalHit& h2) const { return getPos1Begin() < h2.getPos1Begin(); }

 private:

  size_type _pos1;
  size_type _pos2;
  diagonal_type _diagonal;
  size_type _length;
	score_type _score;
  size_type _rightLength;
  size_type _leftLength;
};

namespace LocalHitNS
{
	bool compare_scores(const LocalHit& h1, const LocalHit& h2);
}


#endif
