#include "MandalaWrapper.h"

#include "optimizer.h"
#include "eval-dp-cond.h"
#include "eval-dp.h"
#include "vector.h"
#include "prng.h"

#include <deque>
#include <sstream>

unsigned int nEvals = 0;

void MandalaWrapper::getMandalaModel(TyphonMarkovModel& m, MandalaModel& mm)
{
	Vec<TyphonMarkovModel> tmms;
	for (unsigned int i = 0; i < m.getOrder(); ++i)
	{
		tmms.push_back(TyphonMarkovModel(i));
	}
	tmms.push_back(m);

	for (Vec<TyphonMarkovModel>::reverse_iterator it = tmms.rbegin() + 1; it != tmms.rend(); ++it)
	{
		(it - 1)->projectTo(*it);
	}

	ostringstream ss;
	for (unsigned int i = 0; i <= m.getOrder(); i++)
	{
		ss << i << endl;
		for (unsigned int j = 0; j < tmms[i].getNumHistories(); j++)
		{
			//first output the binary
			unsigned int curNum = j;
			for (unsigned int k = 0; k < i; k++)
			{
				ss << ((curNum >> (i - k - 1)) & 1);
			}
			if (i)
			{
				ss << " ";
			}
			//now we need to actually output the numbers
			ss << tmms[i].getProb(j) << " 0 0" << endl;
		}
	}
	istringstream is(ss.str());
	mm.readFromStream(is);
}

double MandalaWrapper::findNextBestPattern(unsigned int length, TyphonMarkovModel& tmm, Vec<TyphonNS::Pattern>& patternsAlready, unsigned int patternWeight, unsigned int patternSpan, TyphonNS::Pattern& resultPattern)
{
	//build the model
	MandalaModel model(tmm.getOrder());
	getMandalaModel(tmm, model);
	//build the evaluator
	Evaluator* evaluator = NULL;

	Vector<MandalaPattern> fixedPatterns;
	for (unsigned int i = 0; i < patternsAlready.size(); i++)
	{
		fixedPatterns.add(MandalaWrapper::typhonToMandalaPattern(patternsAlready[i]));
	}

	if (fixedPatterns.length())
	{
		evaluator = new DynamicProgrammingCondEvaluator(static_cast<int>(length), &model, fixedPatterns);
	}
	else
	{
		evaluator = new DynamicProgrammingEvaluator(static_cast<int>(length), &model);
	}

	Vector<MandalaPattern> bestPatterns;
	double bestPMatch = -1.0;
	double pMatch;

	for (unsigned int r = 0; r < NUM_RESTARTS; r++)
	{
		// Construct the initial pattern at random
		MandalaPattern p = _createRandomPattern(patternWeight, patternSpan);
		Vector<MandalaPattern> resultPatterns;
		resultPatterns.add(p);
		
		pMatch = findLocallyBestPattern(resultPatterns, patternSpan, evaluator);
		if (pMatch > bestPMatch)
		{
			bestPMatch = pMatch;
			bestPatterns = resultPatterns.copy();
		}
	}
	delete evaluator;
	resultPattern = mandalaToTyphonPattern(bestPatterns[0]);
	if (resultPattern.getPatternLength() < patternSpan)
	{
		resultPattern.pad(patternSpan - resultPattern.getPatternLength(), false);
	}
	return bestPMatch;
}

pair<double, Vec<TyphonNS::Pattern>::size_type> MandalaWrapper::bestPattern(unsigned int length, TyphonMarkovModel& tmm, const Vec<TyphonNS::Pattern>& candidatePatterns, const Vec<Vec<TyphonNS::Pattern>::size_type>& patternsAlready)
{
	//build the model
	MandalaModel model(tmm.getOrder());
	getMandalaModel(tmm, model);
	//build the evaluator
	Evaluator* evaluator = NULL;

	Vector<MandalaPattern> fixedPatterns;
	for (unsigned int i = 0; i < patternsAlready.size(); i++)
	{
		fixedPatterns.add(MandalaWrapper::typhonToMandalaPattern(candidatePatterns[patternsAlready[i]]));
	}

	if (fixedPatterns.length())
	{
		evaluator = new DynamicProgrammingCondEvaluator(static_cast<int>(length), &model, fixedPatterns);
	}
	else
	{
		evaluator = new DynamicProgrammingEvaluator(static_cast<int>(length), &model);
	}

	double bestPMatch = -1.0;
	Vec<TyphonNS::Pattern>::size_type bestPatternPos = 0;

	unsigned int numPatternsAlready = patternsAlready.size();
	for (Vec<TyphonNS::Pattern>::size_type i = 0; i < candidatePatterns.size(); ++i)
	{
		bool ignore = false;
		for (unsigned int j = 0; j != numPatternsAlready; ++j)
		{
			if (patternsAlready[j] == i)
			{
				ignore = true;
				break;
			}
		}
		if (ignore)
		{
			continue;
		}
		Vector<Pattern> curPattern;
		curPattern.add(typhonToMandalaPattern(candidatePatterns[i]));
		double curMatch = evaluator->evaluatePatterns(curPattern, bestPMatch);
		if (curMatch > bestPMatch)
		{
			bestPMatch = curMatch;
			bestPatternPos = i;
		}
	}
	delete evaluator;

	pair<double, Vec<TyphonNS::Pattern>::size_type> result;
	result.first = bestPMatch;
	result.second = bestPatternPos;

	return result;
}

double MandalaWrapper::evaluatePattern(unsigned int length, TyphonMarkovModel& tmm, const Vec<TyphonNS::Pattern>& candidatePatterns, const TyphonNS::Pattern::size_type patternToEvaluate, const Vec<Vec<TyphonNS::Pattern>::size_type>& patternsAlready)
{
	//build the model
	MandalaModel model(tmm.getOrder());

	getMandalaModel(tmm, model);
	//build the evaluator
	Evaluator* evaluator = NULL;

	Vector<MandalaPattern> fixedPatterns;
	for (unsigned int i = 0; i < patternsAlready.size(); i++)
	{
		fixedPatterns.add(MandalaWrapper::typhonToMandalaPattern(candidatePatterns[patternsAlready[i]]));
	}

	if (fixedPatterns.length())
	{
		evaluator = new DynamicProgrammingCondEvaluator(static_cast<int>(length), &model, fixedPatterns);
	}
	else
	{
		evaluator = new DynamicProgrammingEvaluator(static_cast<int>(length), &model);
	}

	Vector<Pattern> curPattern;
	curPattern.add(typhonToMandalaPattern(candidatePatterns[patternToEvaluate]));
	return evaluator->evaluatePatterns(curPattern);
}

double MandalaWrapper::evaluatePatterns(unsigned int length, TyphonMarkovModel& tmm, const Vec<TyphonNS::Pattern>& candidatePatterns)
{
	//build the model
	MandalaModel model(tmm.getOrder());

	getMandalaModel(tmm, model);
	//build the evaluator
	Evaluator* evaluator = new DynamicProgrammingEvaluator(static_cast<int>(length), &model);

	Vector<Pattern> curPattern;
	for (unsigned int i = 0; i < candidatePatterns.size(); i++)
	{
		curPattern.add(typhonToMandalaPattern(candidatePatterns[i]));
	}
	return evaluator->evaluatePatterns(curPattern);
}


MandalaWrapper::MandalaPattern MandalaWrapper::typhonToMandalaPattern(const TyphonNS::Pattern& from)
{
  int* posns = new int[from.getKLength()];
	unsigned int j = 0;
	for (unsigned int i = 0; i < from.getPatternLength(); i++)
	{
		if (from[i])
		{
			posns[j++] = i;
		}
	}
	MandalaPattern to(posns, from.getKLength());
	delete posns;

	return to;
}

TyphonNS::Pattern MandalaWrapper::mandalaToTyphonPattern(const MandalaPattern& from)
{
	Vec<bool> boolPattern(from.span());
	for (unsigned int i = 0; i < static_cast<unsigned int>(from.weight()); i++)
	{
		boolPattern[from[i]] = true;
	}
	return TyphonNS::Pattern(boolPattern);
}

MandalaWrapper::MandalaPattern MandalaWrapper::_createRandomPattern(unsigned int weight, unsigned int span)
{
  int* posns = new int[weight];
  unsigned int nChosen = 1;
  
  posns[0] = 0; // for uniqueness
  do
	{
		int posn = genrand_int32() % (span - 1) + 1;
		
		bool dup = false;
		for (unsigned int j = 0; j < nChosen; j++)
		{
			if (posns[j] == posn)
			{
				dup = true;
				break;
			}
		}
		
		if (!dup)
		{
			posns[nChosen++] = posn;
		}
	}
  while (nChosen < weight);
  
  MandalaPattern p(posns, weight);
  delete posns;

  return p;
}


