#ifndef _matchfinder_h_included_
#define _matchfinder_h_included_

#include "LocalHit.h"
#include "LocalAlignment.h"
#include "Pattern.h"
#include "Mseq.h"
#include <string>
#include <set>
#include "System.h"
#include "ScoringParameters.h"
#include "Scorer.h"
#include "Seq.h"

class MatchFinder
{
	friend class HeuristicIndex;
	friend class Index;

 public:

	/**
	 * Constants
	 */

	static const double DEFAULT_E_VALUE = 10;
	static const double DEFAULT_CUTOFF_THRESHOLD = 23;
	static const double DEFAULT_GAPPED_CUTOFF_THRESHOLD = 40;


	/* Types */
	/* ===== */
	typedef LocalHit::score_type score_type;
	typedef unsigned long long statistics_type;
	static const LocalHit::diagonal_type NO_HIT = -1;
	static const unsigned int MAX_PER_LINE = 60;

	enum ScoringType { ConsensusScoring, PSIScoring, SPScoring};

	/* Constructors */
	/* ============ */

	/**
	 * Build MatchFinder
	 * finalThreshold is the final score that a hit needs to be output
	 * cutoffThreshold is the point at which we stop extending the hit (when it falls this far below the max
	 *   score so far)
	 * maxDistance is the maximum distance for two hits to be apart before we chain them.  If this is zero, then
	 *   we extend every hit
	 *   in the hit
	 * maxHitLength is the maximum length of returned hits; pass 0 for infinity
	 */

  MatchFinder(const double bitSemiFinalThreshold
							,const double bitFinalThreshold
							,const double bitThresholdToDoGapped
							,const double semiEValue
							,const double eValue
							,const double eValueToDoGapped
							,const double bitSemiCutoffThreshold
							,const double bitCutoffThreshold
							,const double bitGappedCutoff
							,const bool doGapped
							,const Seq::size_type hspSizeForGapped
							,const Seq::size_type maxDistance
							,const ScoringType scoringType
							,const Seq::size_type maxHitLength = 0
							) : _bitSemiFinalThreshold(bitSemiFinalThreshold), _bitFinalThreshold(bitFinalThreshold), _bitThresholdToDoGapped(bitThresholdToDoGapped),  _semiEValue(semiEValue), _eValue(eValue), _eValueToDoGapped(eValueToDoGapped), _bitSemiCutoffThreshold(bitSemiCutoffThreshold), _bitCutoffThreshold(bitCutoffThreshold), _bitGappedCutoff(bitGappedCutoff), _doGapped(doGapped), _hspSizeForGapped(hspSizeForGapped), _maxDistance(maxDistance), _combine(false), _scoringType(scoringType), _maxHitLength(maxHitLength), _start1Offset(0), _start2Offset(0), MINF(numeric_limits<score_type>::min())
		{
			_clearStats();
		}

	/* Methods */
	/* ======= */

	/*
	 * These are initialization functions for the sequences and scoring stuff
	 */
	void setSequence1(Mseq& sequence1
									,ScoringParameters<score_type>& scoringParameters);
	void setSequence2(Mseq& sequence2);
	void initialize(ScoringParameters<score_type>& scoringParameters);

	statistics_type getNumHitsExtended() { return _numHitsExtended; }
	statistics_type getNumHitsExtendedMore() { return _numHitsExtendedMore; }
	statistics_type getNumGappedExtensions() { return _numGappedExtensions; }
	statistics_type getNumCellsExplored() { return _numCellsExplored; }
  statistics_type getTotalMatchesPreExtension() { return _totalMatchesPreExtension; }
  statistics_type getTotalMatches() { return _totalMatches; }


	/**
	 * Print the stats we have from the most recent getMatches call
	 */
	void writeOutResults(Mseq& sequence1, Mseq& sequence2, ostream& outStream, LocalHit::OutputType outputType, double indexTime, double scanTime);

 private:

	/**
	 * Start fresh with statistics
	 */
	void _clearStats();

	/**
	 * This takes a hit and adds it to the list of hits if it is high scoring after extension
	 * It returns whether or not the hit was added
	 */
	bool _addHit(Mseq& sequence1, Seq::size_type position1, Mseq& sequence2, Seq::size_type position2, TyphonNS::Pattern::size_type seedLength);

	/**
	 * This takes a hit and tries to extend it
	 * It returns whether or not the hit is high scoring
	 */
	bool _extendHitC(LocalHit& seed);
	bool _extendHitC(LocalHit& seed, score_type cutoffThreshold);
	//	bool _extendHitPSI(LocalHit& seed);
	//bool _extendHitPSI(LocalHit& seed, score_type cutoffThreshold);

	/**
	 * This checks if the hit is COMPLETELY contained in some alignment
	 */
	bool _containedInAlignment(const LocalHit& hit);

	LocalHit _getCenterPoint(const LocalHit& hit);

	/**
	 * This takes an alignment and does a gapped extension of it
	 */
	bool _doGappedExtension(LocalHit& hit);

	/**
	 * This does a gapped extension of an alignment in one direction
	 */
	enum GappedDirection {BACKWARD, FORWARD};
	enum AlignmentChar {DIAG_TO_MATCH, DIAG_TO_GAP_X, DIAG_TO_GAP_Y, SEQUENCE1_TO_MATCH, SEQUENCE1_TO_GAP_X, SEQUENCE1_TO_GAP_Y, SEQUENCE2_TO_MATCH, SEQUENCE2_TO_GAP_X, SEQUENCE2_TO_GAP_Y, ALIGNMENT_BEGIN, NO_VALUE};
	enum MatrixType {MATCH, GAP_X, GAP_Y};
	void _doGappedExtension(LocalAlignment& alignment, GappedDirection direction);

	/**
	 * Utility method to write out a LocalHit to a file
	 */
	MatchFinder::statistics_type _writeOutHit(Mseq& sequence1, Mseq& sequence2, LocalHit hit, ostream& tempOut, LocalHit::OutputType outputType, bool recalculateScore = false) const;

	/**
	 * Utility method for writing out lines of blast
	 */
	void _printBlastLine(Seq::size_type queryStart, Seq::size_type queryEnd, const Seq::size_type queryOff, const string& queryString, Seq::size_type sbjctStart, Seq::size_type sbjctEnd, const Seq::size_type sbjctOff, const string& sbjctString, const string& matchString, ostream& outStream) const;

	/**
	 * Utility method to write out a LocalAlignment to a file
	 */
	 MatchFinder::statistics_type _writeOutAlignment(Mseq& sequence1, Mseq& sequence2, LocalAlignment alignment, ostream& tempOut, LocalHit::OutputType outputType) const;



	/* Members */
	/* ======= */

	/**
	 * Info on the sequences
	 */
	Seq::size_type _length1;
	Seq::size_type _length2;
	Mseq::size_type _size1;
	Mseq::size_type _size2;
	
	/**
	 * The scoring parameters
	 */
	score_type _semiFinalThreshold;
	score_type _finalThreshold;
	score_type _thresholdToDoGapped;
	double _bitSemiFinalThreshold;
	double _bitFinalThreshold;
	double _bitThresholdToDoGapped;

	double _semiEValue;
	double _eValue;
	double _eValueToDoGapped;
	score_type _semiCutoffThreshold;
	score_type _cutoffThreshold;
	score_type _gappedCutoff;
	double _bitSemiCutoffThreshold;
	double _bitCutoffThreshold;
	double _bitGappedCutoff;

	bool _doGapped;
	Seq::size_type _hspSizeForGapped;

	Seq::size_type _maxDistance;

	bool _combine;

	ScoringType _scoringType;

	/**
	 * This is the maximum hit length we'll allow; we'll split hits at the END however, so
	 * some long hits may be split into smaller hits, some of which aren't high scoring
	 */
	Seq::size_type _maxHitLength;

	/**
	 * Internal
	 */
	typedef Vec<unsigned int> consensus_type;
	consensus_type _consensus1;
	consensus_type _consensus2;

	string _sequence1Name;
	string _sequence2Name;

	typedef	ScoringParameters<score_type>::ScoringMatrix scoring_matrix;
	scoring_matrix _psiMatrix;
	Seq::size_type _psiMatrixLength;
	scoring_matrix _matrix;
	score_type _gapOpen;

	scoring_matrix _spMatrix;

	double _lambda;
	double _K;
	double _effectiveLength1;
	double _effectiveLength2;
	
	/**
	 * This stores the offset to be added to the sequence; we always index them base 0
	 * but they could be subsequences of a true sequence, in which case hits at position
	 * correspond to some other position; those positions are stored in these offsets
	 */
	Seq::size_type _start1Offset;
	Seq::size_type _start2Offset;

	/**
	 * These store the data associated with the hits we've found
	 */

	typedef Vec<LocalHit::diagonal_type> fast_hit_storage_type;
	typedef hash_multimap_type<LocalHit::diagonal_type, LocalHit> hits_to_write_type;
	typedef Vec<LocalAlignment> alignments_to_write_type;

	LocalHit::diagonal_type _diagonalOffset;

	statistics_type _numHitsExtended;
	statistics_type _numHitsExtendedMore;
	statistics_type _numGappedExtensions;
	statistics_type _numCellsExplored;
  statistics_type _totalMatchesPreExtension;
  statistics_type _totalMatches;

	/**
	 * This stores the most recent high scoring hit for a given diagonal
	 */
  //hit_storage_type _hits;
  fast_hit_storage_type _fastHits;

	/**
	 * This stores the most recent seed for a given diagonal
	 */
  fast_hit_storage_type _lastSeeds;
	/**
	 * These are potentially high scoring hits that have been bumped out of the _hits structure
	 * and need to potentially be written to disk
	 */
	set<LocalHit::diagonal_type> _diagonalsToWrite;
	hits_to_write_type _hitsToWrite;
	alignments_to_write_type _alignmentsToWrite;
	hits_to_write_type _alignmentComponentsFound;

	/* Minus infinity (depends on the score_type) */
	score_type MINF;
	unsigned int GAP_INDEX;

};

#endif
