#ifndef _pattern_h_included_
#define _pattern_h_included_

#include "System.h"
#include "Seq.h"
#include "Vec.h"
#include <string>
#include <iostream>

/**
 * The Pattern class represents a seed mask that we use
 * when scanning for local hits
 *
 * It is essentially a bit vector, and we consider the 
 * 1s as neccessary matches for a hit and the 0s as
 * don't cares
 */

namespace TyphonNS
{

	class Pattern
		{
		public:
			
			typedef Vec<bool>::size_type size_type;
			
			/* Constructors */
			/* ============ */
			
			/**
			 * Build a pattern from a bit vector
			 */
			Pattern(const Vec<bool> p);
			
			/**
			 * Build a contiguous pattern
			 */
			Pattern(const Vec<bool>::size_type w);
			
			/**
			 * Build an empty pattern (length 0)
			 */
			Pattern();
			/**
			 * Build a pattern from a string representation
			 * The string rep is a string of 0s and 1s
			 * that maps to a bit vector in the obvious way
			 */
			Pattern(const char * const stringRep);
			
			/**
			 * Build a pattern from an unsigned integer
			 * The least significant bit is the first position in the pattern
			 */
			Pattern(const unsigned int intRep, const Pattern::size_type size);
			
			/**
			 * Build a pattern from a vector of positions where the pattern has zeros
			 */
			Pattern(const Vec<Pattern::size_type>& zeroPositions, Pattern::size_type size);
			
			/* Methods */
			/* ======= */
			
			/**
			 * The length of the pattern; this is the total number
			 * of 1s and 0s
			 */
			size_type getPatternLength() const { return _patternLength; }
			
			/**
			 * The number of true entries in this pattern; that is,
			 * the number of 1s in the bit vector
			 */
			size_type getKLength() const { return _kLength; }
			
			/**
			 * Tells whether this pattern is all 1s
			 */
			bool isContiguous() const { return _patternLength == _kLength; }
			
			/**
			 * Convert this pattern to a string
			 */
			std::string toString() const;
			
			/**
			 * Convert this pattern to a Vec<bool>
			 */
			Vec<bool> toVecBool() const { return _pattern; }
			
			/**
			 * Return a pattern that is only newSize weight
			 */
			void trunc(Pattern::size_type newSize);

			/**
			 * Return a pattern that is only newSize span
			 */
			void trim(Pattern::size_type newSize);
			
			/**
			 * Increase size by numToAdd with value
			 */
			void pad(Pattern::size_type numToAdd, bool value);
			
			
			/* Operators */
			/* ========= */
			
			/**
			 * Get the value of the pattern at a given position
			 */
			bool operator[](size_type pos) const;

			/**
			 * Compare two patterns
			 */
			bool operator==(const Pattern& pattern) const;
			
			/**
			 * Print out a pattern
			 */
			
			friend std::ostream& operator<<(std::ostream& out, const Pattern&);

			
			
		private:
			
			/**
			 * Internal storage
			 */
			Vec<bool> _pattern;
			size_type _kLength;
			size_type _patternLength;
		};
}	
#endif
	
