#ifndef _phytree_h_included_
#define _phytree_h_included_

#include "System.h"
#include "PhyTreeNode.h"
#include <istream>

/**
 * This class represents a phylogenetic tree
 */

class PhyTree
{
 public:

	typedef PhyTreeNode::branch_length_type branch_length_type;
	typedef PhyTreeNode::prob_type prob_type;

	PhyTree(istream& is, Mseq& sequence, string queryName = "")
		{
			_root = new PhyTreeNode(is, sequence, queryName);
			//now, we want to reorient the tree to put the query at the root
			_rotate(queryName);
			double defaultLength = 0.05;
			PhyTreeNode::_setNoBranchLengthToLength(_root, defaultLength);
		}

	~PhyTree()
		{
			delete _root;
		}

	void getProbVector(ProbVector& pv
										 ,const bool memOpt
										 ,const prob_type alpha = 1
										 ,const prob_type gamma = 2
										 ,const Vec<prob_type>& f = Vec<prob_type>(0))
		{
			_root->getProbVector(pv, memOpt, alpha, gamma, f);
			ProbVector::size_type length = pv.size();
			for (ProbVector::size_type i = 0; i < length; i++)
			{
				prob_type maxProb = pv.getMaxLetterProb(i);

				const double vShift = .25;
				const double hShift = .25;
				const double ePower = 4;
				const double scale = (1 - vShift) / (exp(ePower * (1 - hShift)));
				
				pv.setMaxLetterProb(i, scale * exp(ePower * (maxProb - hShift)) + vShift);
			}
		}

 private:

	void _rotate(string queryName)
		{
			if (queryName.size())
			{
				PhyTreeNode* node = PhyTreeNode::_presentInTree(_root, queryName);
				branch_length_type newBranchLength = node->_branchLengthToParent;
				assert(node != NULL);
				if (node->_parent)
				{
					PhyTreeNode* parent = node->_parent;
					//first, remove the node from the tree
					Vec<PhyTreeNode*>::iterator it = parent->_children.begin();
					for (; it != parent->_children.end(); ++it)
					{
						if ((*it)->_name == queryName)
						{
							break;
						}
					}
					assert(it != parent->_children.end());
					parent->_children.erase(it);
					delete node;

					//the root is now the parent of this node
					if (parent == _root)
					{
						//just in case the root now has only one child
						Vec<string> dummyNames;
						PhyTreeNode::_eliminateNodes(_root, dummyNames, false);
					}
					else
					{
						//makeroot can guarantee that parent is not NULL
						_root = parent;
						PhyTreeNode::_makeRoot(_root);
						_root->_parent = NULL;

					}
				}
				_root->_branchLengthToParent = newBranchLength;
			}
		}

	PhyTreeNode* _root;

};

#endif
