#include "PhyTreeNode.h"
#include <iostream>
#include <cmath>

PhyTreeNode::PhyTreeNode(istream& is, Mseq& sequence, const string& queryName) : _parent(NULL), _branchLengthToParent(0)
{
	initialize_prof_idx();	
	_getNode(this, is);

	//first get the sequence names
	Vec<string> sequenceNames;
	if (queryName.size())
	{
		sequenceNames.push_back(queryName);
	}
	//eliminate all nodes not present in the multiple alignment
	for (Mseq::iterator it = sequence.getBegin(); it != sequence.getEnd(); ++it)
	{
		sequenceNames.push_back(it->getTitle());
	}
	_eliminateNodes(this, sequenceNames);

	//now insert all nodes in the alignment into the tree
	for (Vec<string>::const_iterator it = sequenceNames.begin(); it != sequenceNames.end(); ++it)
	{
		if (!_presentInTree(this, *it))
		{
			//we may need to move this one down
			if (_children.size() == 0 && _name.size())
			{
				PhyTreeNode* newChild = new PhyTreeNode(this);
				newChild->_branchLengthToParent = _branchLengthToParent;
				newChild->_name = _name;
				_children.push_back(newChild);
				_branchLengthToParent = 0;
				_name.clear();
			}
			PhyTreeNode* newChild = new PhyTreeNode(this);
			newChild->_branchLengthToParent = 0;
			newChild->_name = *it;
			_children.push_back(newChild);
		}
	}
	_makeBinary(this);
	_setDepth(this);
	//now, go ahead and store the sequences
	_storeSequences(this, sequence);
}

PhyTreeNode::~PhyTreeNode()
{
	for (Vec<PhyTreeNode*>::iterator it = _children.begin(); it != _children.end(); ++it)
	{
		delete (*it);
	}
}

void PhyTreeNode::getProbVector(ProbVector& pv, const bool memOpt, const prob_type baseAlpha, const prob_type baseBeta, const Vec<prob_type>& fOrig)
{
	//cerr << "Generating probabilities..." << endl;
	Vec<prob_type> f(fOrig);
	if (f.size() < BIO::NUM_NUCLEOTIDES)
	{
		f.clear();
		f.resize(BIO::NUM_NUCLEOTIDES, 1.0 / BIO::NUM_NUCLEOTIDES);
	}

	Vec<ProbVector> childPv(1);
	prob_type betaToAlpha = baseBeta / baseAlpha;

	//branch_length_type prevBranchLength = _branchLengthToParent;
	//set all of the branch lengths just to get some variety
	//branch_length_type allBranches = numSubstitutionsForAll;
	//_setNoBranchLengthToLength(this, allBranches);
	//except for this one
	//_branchLengthToParent = prevBranchLength;

	if (memOpt)
	{
		//do it in pieces
		const unsigned int increment = 1000000;
		const unsigned int length = _getLength();
		childPv.front().resize(length);
		for (unsigned int curBegin = 0, curEnd = curBegin + increment; curBegin < length; curBegin = curEnd, curEnd = (length - curEnd) > increment ? curEnd + increment : length)
		{
			Vec<ProbVector> childPvSub(1);
			_getProbVector(childPvSub.front(), curBegin, curEnd, betaToAlpha, f);


			for (unsigned int i = curBegin; i < curEnd; i++)
			{
				childPv.front().copyFrom(i, childPvSub.front(), i - curBegin);
			}
		}
	}
	else
	{
		_getProbVector(childPv.front(), 0, _getLength(), betaToAlpha, f);
	}

	//double numSubstitutionsForQueryToUse = numSubstitutionsForQuery;
	//if (!_branchLengthToParent)
	//{
	//	numSubstitutionsForQueryToUse = numSubstitutionsForAll;
	//}
	Vec<branch_length_type> branchLengths(1, _getBranchLength(_branchLengthToParent, betaToAlpha, f));
	_evolveFromChildren(childPv, branchLengths, pv, betaToAlpha, f);

	//now renormalize
	//const unsigned int gapIndex = CHAR_PROFILE_IDX[static_cast<unsigned char>(GAP_CHAR)];
	//for (ProbVector::size_type i = 0; i < length; i++)
	{
		//prob_type noGapProb = 1 - pv.getProbability(i, gapIndex);
		//for (ProbVector::size_type j = 0; j < BIO::NUM_NUCLEOTIDES; j++)
		{
			//pv.setProbability(i, j, pv.getProbability(i,j) * noGapProb);
		}
	}

}

void PhyTreeNode::_getProbVector(ProbVector& pv, unsigned int beginPos, unsigned int endPos, const prob_type betaToAlpha, const Vec<prob_type>& f)
{
	pv.clear();
	//base case
	if (_children.size() == 0)
	{
		//copy into the vector
		pv.resize(endPos - beginPos);
		Seq::iterator endIt = _sequence.getBegin() + endPos;
		Seq::iterator beginIt = _sequence.getBegin() + beginPos;
		Seq::size_type i = 0;
		for(Seq::iterator it = beginIt; it != endIt; ++it, ++i)
		{
			unsigned char letter = static_cast<unsigned char>(*it);
			if (letter == ANY_CHAR)
			{
				letter = GAP_CHAR;
			}
			pv.setProbability(i, CHAR_PROFILE_IDX[letter], 1);
		}
	}
	else
	{
		Vec<ProbVector> childPv(_children.size());
		Vec<unsigned int> indices(_children.size());
		Vec<pair<unsigned int, unsigned int> > depths(indices.size());
		for (unsigned int i = 0; i < _children.size(); i++)
		{
			depths[i] = make_pair(_children[i]->_depth, i);
		}
		sort(depths.begin(), depths.end());
		for (unsigned int i = 0; i < depths.size(); i++)
		{
			indices[i] = depths[i].second;
		}

		for (Vec<PhyTreeNode*>::size_type i = 0; i != indices.size(); ++i)
		{
			_children[indices[i]]->_getProbVector(childPv[indices[i]], beginPos, endPos, betaToAlpha, f);
		}
		Vec<ProbVector>::iterator pIt = childPv.begin();
		for (Vec<PhyTreeNode*>::iterator it = _children.begin(); it != _children.end(); ++it, ++pIt)
		{
			//check if we need to compute a branch length
// 			if (!(*it)->_branchLengthToParent)
// 			{
// 				//this always should happen
// 				assert(_children.size() > 1);
// 				branch_length_type totalBranchLengths = 0;
// 				Vec<ProbVector>::iterator spIt = childPv.begin();
// 				for (Vec<PhyTreeNode*>::iterator sIt = _children.begin(); sIt != _children.end(); ++sIt, ++spIt)
// 				{
// 					if (sIt == it)
// 					{
// 						continue;
// 					}

// 					ProbVector& first = *pIt;
// 					ProbVector& second = *spIt;
// 					ProbVector::size_type length = first.size();
// 					//count the total number of differences; we will use this to compute the branch lengths
// 					prob_type totalDifferences = 0;
// 					//get the start
// 					ProbVector::size_type start = 0;
// 					unsigned int gapIndex = CHAR_PROFILE_IDX[static_cast<unsigned char>(GAP_CHAR)];
// 					while (start < length && (first.getProbability(start, gapIndex) > .5 || second.getProbability(start, gapIndex) > .5))
// 					{
// 						start++;
// 					}
// 					ProbVector::size_type end = first.size() - 1;
// 					while (end > start && (first.getProbability(end, gapIndex) > .5 || second.getProbability(end, gapIndex) > .5))
// 					{
// 						end--;
// 					}
					
// 					end++;
// 					ProbVector::size_type effectiveLength = 0;

// 					for (ProbVector::size_type i = start; i < end; ++i)
// 					{
// 						if (first.getProbability(i, gapIndex) > .5 || second.getProbability(i, gapIndex) > .5)
// 						{
// 							continue;
// 						}
// 						effectiveLength++;
// 						for (ProbVector::size_type j = 0; j < BIO::NUM_NUCLEOTIDES; j++)
// 						{
// 							for (ProbVector::size_type k = 0; k < BIO::NUM_NUCLEOTIDES; k++)
// 							{
// 								if (j == k)
// 								{
// 									continue;
// 								}
// 								totalDifferences += first.getProbability(i, j) * second.getProbability(i, k);
// 							}
// 						}
// 					}
// 					totalDifferences /= effectiveLength;
// 					branch_length_type newBranchLength = _getBranchLength(totalDifferences, betaToAlpha, f);
// 					//to each is half what's between
// 					newBranchLength /= 2;
// 					totalBranchLengths += newBranchLength;
// 				}
// 				branch_length_type newBranchLength = totalBranchLengths / (_children.size() - 1);
// 				(*it)->_branchLengthToParent = newBranchLength;
// 			}
		}
		Vec<branch_length_type> branchLengths(_children.size());
		for (unsigned int i = 0; i < _children.size(); ++i)
		{
			branchLengths[i] = _getBranchLength(_children[i]->_branchLengthToParent, betaToAlpha, f);
		}
		_evolveFromChildren(childPv, branchLengths, pv, betaToAlpha, f);
	}
}

unsigned int PhyTreeNode::_getLength() const
{
	if (!_children.size())
	{
		return _sequence.getLength();
	}
	else
	{
		return _children.front()->_getLength();
	}
}

PhyTreeNode::branch_length_type PhyTreeNode::_getBranchLength(double subsPerSite, const prob_type betaToAlpha, const Vec<prob_type>& f)
{
	//just do a simple calculation here; uniform model
	branch_length_type newBranchLength = 1;
	
	if ((1 - subsPerSite) <= 1.0 / BIO::NUM_NUCLEOTIDES)
	{
		newBranchLength = INF_BRANCH_LENGTH;
	}
	else
	{
		newBranchLength = subsPerSite / (2 * betaToAlpha + 1);
	}
	return newBranchLength;
}

void PhyTreeNode::_evolveFromChildren(Vec<ProbVector>& childPv, Vec<branch_length_type>& branchLengths, ProbVector& pv, const prob_type betaToAlpha, const Vec<prob_type>& f)
{
	//resize the vector
	pv.resize(childPv[0].size());
	ProbVector::size_type length = pv.size();
	for (ProbVector::size_type i = 0; i < length; i++)
	{
		for (ProbVector::size_type j = 0; j < BIO::NUM_ALPHABET; j++)
		{
			pv.setProbability(i, j, 0);
		}
	}

	const unsigned int gapIndex = CHAR_PROFILE_IDX[static_cast<unsigned char>(GAP_CHAR)];

	prob_type totalWeights = 0;
	for (Vec<branch_length_type>::iterator it = branchLengths.begin(); it != branchLengths.end(); ++it)
	{
		totalWeights += (1 / *it);
	}

	Vec<ProbVector>::iterator pIt = childPv.begin();

	for (Vec<branch_length_type>::iterator it = branchLengths.begin(); it != branchLengths.end(); ++it, ++pIt)
	{			
		//build up a matrix to cache the values
		Vec<Vec<prob_type> > matrix;
		matrix.resize(BIO::NUM_NUCLEOTIDES);
		for (Vec<prob_type>::size_type i = 0; i < BIO::NUM_NUCLEOTIDES; i++)
		{
			matrix[i].resize(NUM_NUCLEOTIDES);
			for (Vec<prob_type>::size_type j = 0; j < BIO::NUM_NUCLEOTIDES; j++)
			{
				matrix[i][j] = PhyTreeNodeNS::kimura(i, j, *it, betaToAlpha, f);
			}
		}

		//now we can go ahead and actually compute the likelihood
		ProbVector::size_type length = pv.size();

		for (ProbVector::size_type i = 0; i < length; i++)
		{
			prob_type weight = (1 / (*it));

			prob_type curGapContribution = pIt->getProbability(i, gapIndex) * weight;


			//first do the letters; they only consider probability of evolving to other letters
			for (ProbVector::size_type j = 0; j < BIO::NUM_NUCLEOTIDES; j++)
			{
				//P(L) = sum_B P(L|B)P(B)
				prob_type curContribution = 0;
				for (ProbVector::size_type k = 0; k < BIO::NUM_NUCLEOTIDES; k++)					
				{
					prob_type curChildValue = pIt->getProbability(i, k);
					curContribution += curChildValue * matrix[j][k];
				}

				//it's possible that curContribution could have p=0, if all gapped. In this case, we ignore
				if (curContribution)
				{
					prob_type mult = pv.getProbability(i,j);

					if (!mult)
					{
						mult = 1;
					}
					pv.setProbability(i, j, mult * curContribution);
				}
			}

			//P(L) = sum_B P(L|B)P(B)

			prob_type curProb = pv.getProbability(i, gapIndex);
			prob_type newProb = curProb + curGapContribution;// * remainGap + (1 - curContribution) * changeToGap;

			pv.setProbability(i, gapIndex, newProb);
		}

	}
	//renormalize
	for (ProbVector::size_type i = 0; i < length; i++)
	{
		prob_type totalProb = 0;
		for (ProbVector::size_type j = 0; j < BIO::NUM_NUCLEOTIDES; j++)
		{
			totalProb += pv.getProbability(i, j);
		}

		pv.setProbability(i, gapIndex, pv.getProbability(i, gapIndex) / totalWeights);

		for (ProbVector::size_type j = 0; j < BIO::NUM_NUCLEOTIDES; j++)
		{
			pv.setProbability(i, j, totalProb ? (pv.getProbability(i,j) / totalProb) : 0);
		}
	}
}

bool PhyTreeNode::_getNode(PhyTreeNode* node, istream& is)
{
	//eat white space
	while (is.good() && isspace(is.peek()))
	{
		is.get();
	}
	if (!is.good())
	{
		return false;
	}
	if (is.peek() == '(')
	{
		//internal node
		//eat the opening
		is.get();
		bool doneChildren = false;
		while (!doneChildren)
		{
			//we need to read the children
			node->_children.push_back(new PhyTreeNode(node));

			bool gotChild = _getNode(node->_children.back(), is);
			if (!gotChild)
			{
				//shrink it back down
				delete node->_children.back();
				node->_children.resize(node->_children.size() - 1);				
			}
			//now see if more children
			if (!is.good() || is.get() == ')')
			{
				doneChildren = true;
			}
		}
		if (is.good() && is.peek() == ':')
		{
			is.get();
			//get the weight
			if (is.good())
			{
				//branch lengths are substitutions per site
				is >> node->_branchLengthToParent;
			}
		}
		if (node->_children.size() == 0)
		{
			return false;
		}
		else if (node->_children.size() == 1)
		{
			PhyTreeNode* firstChild = node->_children.front();
			node->_children.clear();
			node->_children.resize(firstChild->_children.size());
			copy(firstChild->_children.begin(), firstChild->_children.end(), node->_children.begin());
			double numSubstitutions = node->_branchLengthToParent + firstChild->_branchLengthToParent;
			node->_branchLengthToParent = numSubstitutions;
			node->_name = firstChild->_name;
			for (Vec<PhyTreeNode*>::iterator it = node->_children.begin(); it != node->_children.end(); ++it)
			{
				(*it)->_parent = node;
			}
			//ensure children aren't deleted
			firstChild->_children.clear();
			delete firstChild;
			return true;
		}
		else
		{
			return true;
		}
	}
	else
	{
		//leaf node
		//go until we see a comma or a closed parenthesis
		while (is.good() && is.peek() != ')' && is.peek() != ',' && is.peek() != ' ' && is.peek() != ':')
		{
			node->_name += is.get();
		}
		if (is.good() && is.peek() == ':')
		{
			is.get();
			//get the weight
			if (is.good())
			{
				is >> node->_branchLengthToParent;
			}
		}
		return node->_name.size() > 0;
	}
}

PhyTreeNode::branch_length_type PhyTreeNode::_getMinBranchLength(PhyTreeNode* node)
{
	if (node->_children.size() == 0)
	{
		return node->_branchLengthToParent;
	}
	else
	{
		branch_length_type minBranch = node->_branchLengthToParent;
		for (Vec<PhyTreeNode*>::iterator it = node->_children.begin(); it != node->_children.end(); ++it)
		{
			branch_length_type curChildBranch = _getMinBranchLength(*it);
			if (!minBranch || (curChildBranch && curChildBranch < minBranch))
			{
				minBranch = curChildBranch;
			}
		}
		return minBranch;
	}
}

void PhyTreeNode::_setNoBranchLengthToLength(PhyTreeNode* node, const branch_length_type length)
{
	if (!node->_branchLengthToParent)
	{
		node->_branchLengthToParent = length;
	}
	for (Vec<PhyTreeNode*>::iterator it = node->_children.begin(); it != node->_children.end(); ++it)
	{
		_setNoBranchLengthToLength(*it, length);
	}
}

bool PhyTreeNode::_eliminateNodes(PhyTreeNode* node, const Vec<string>& names, const bool checkNames)
{
	if (node->_children.size() == 0)
	{
		if (checkNames)
		{
			//if we got a name we have a leaf
			for (Vec<string>::const_iterator it = names.begin(); it != names.end(); ++it)
			{
				if (*it == node->_name)
				{
					return true;
				}
			}
			return false;
		}
		else
		{
			return true;
		}
	}
	else
	{
		Vec<PhyTreeNode*> newChildren;
		for (Vec<PhyTreeNode*>::iterator it = node->_children.begin(); it != node->_children.end(); ++it)
		{
			if (_eliminateNodes(*it, names, checkNames))
			{
				newChildren.push_back(*it);
			}
			else
			{
				delete *it;
			}
		}
		if (newChildren.size() != node->_children.size())
		{
			node->_children.clear();
			node->_children.resize(newChildren.size());
			copy(newChildren.begin(), newChildren.end(), node->_children.begin());

			if (node->_children.size() == 1)
			{
				PhyTreeNode* firstChild = node->_children.front();
				node->_children.clear();
				node->_children.resize(firstChild->_children.size());
				copy(firstChild->_children.begin(), firstChild->_children.end(), node->_children.begin());
				double numSubstitutions = node->_branchLengthToParent + firstChild->_branchLengthToParent;
				node->_branchLengthToParent = numSubstitutions;
				node->_name = firstChild->_name;
				for (Vec<PhyTreeNode*>::iterator it = node->_children.begin(); it != node->_children.end(); ++it)
				{
					(*it)->_parent = node;
				}
				//ensure children aren't deleted
				firstChild->_children.clear();
				delete firstChild;
				return true;
			}
			else 
			{
				return node->_children.size() > 0;
			}
		}
		else
		{
			return true;
		}
	}
}

PhyTreeNode* PhyTreeNode::_presentInTree(PhyTreeNode* node, const string& name)
{
	if (node->_name == name)
	{
		return node;
	}
	else
	{
		for (Vec<PhyTreeNode*>::iterator it = node->_children.begin(); it != node->_children.end(); ++it)
		{
			PhyTreeNode* result = _presentInTree(*it, name);
			if (result)
			{
				return result;
			}
		}
		return NULL;
	}
}

void PhyTreeNode::_makeBinary(PhyTreeNode* node)
{
	for (Vec<PhyTreeNode*>::iterator it = node->_children.begin(); it != node->_children.end(); ++it)
	{
		_makeBinary(*it);
	}

	//now we need to do some work
	while (node->_children.size() > 2)
	{
		branch_length_type minChildBranchLength = node->_children[0]->_branchLengthToParent > node->_children[1]->_branchLengthToParent ? node->_children[1]->_branchLengthToParent : node->_children[0]->_branchLengthToParent;
		// we need to create the new nodes
		PhyTreeNode* newChild = new PhyTreeNode(node);
		newChild->_branchLengthToParent = minChildBranchLength / 2;
		newChild->_children.push_back(node->_children[0]);
		newChild->_children.back()->_branchLengthToParent -= (minChildBranchLength / 2);
		newChild->_children.push_back(node->_children[1]);
		newChild->_children.back()->_branchLengthToParent -= (minChildBranchLength / 2);
		for (Vec<PhyTreeNode*>::iterator it = newChild->_children.begin(); it != newChild->_children.end(); ++it)
		{
			(*it)->_parent = newChild;
		}

		node->_children.erase(node->_children.begin());
		node->_children.front() = newChild;
	}
}

void PhyTreeNode::_storeSequences(PhyTreeNode* node, Mseq& sequence)
{
	if (node->_children.size() > 0)
	{
		for (Vec<PhyTreeNode*>::iterator it = node->_children.begin(); it != node->_children.end(); ++it)
		{
			_storeSequences(*it, sequence);
			node->_name += "/" + (*it)->_name;
		}
	}
	else
	{
		for (Mseq::iterator it = sequence.getBegin(); it != sequence.getEnd(); ++it)
		{
			if (it->getTitle() == node->_name)
			{
				//copy the sequence
				node->_sequence = *it;
			}
		}
	}
}

void PhyTreeNode::_makeRoot(PhyTreeNode* node)
{
	assert(node->_parent != NULL);
	Vec<PhyTreeNode*>& parentsChildren = node->_parent->_children;
	if (node->_parent->_parent == NULL)
	{
		for (Vec<PhyTreeNode*>::iterator it = parentsChildren.begin(); it != parentsChildren.end(); ++it)
		{
			if (*it != node)
			{
				node->_children.push_back(*it);
				double numSubstitutions = node->_branchLengthToParent + (*it)->_branchLengthToParent;
				(*it)->_branchLengthToParent = numSubstitutions;
				(*it)->_parent = node;
			}
		}
		node->_parent->_children.clear();
		delete node->_parent;
		return;
	}
	else
	{
		node->_children.push_back(node->_parent);
		Vec<PhyTreeNode*>::iterator it = parentsChildren.begin();
		for (; it != parentsChildren.end(); ++it)
		{
			if (*it == node)
			{
				break;
			}
		}
		parentsChildren.erase(it);
		_makeRoot(node->_parent);
		node->_parent->_parent = node;
	}
}

PhyTreeNode::branch_length_type PhyTreeNode::_closestNode(PhyTreeNode* node, string& name)
{
	if (!node->_children.size())
	{
		name = node->_name;
		return node->_branchLengthToParent;
	}
	else
	{
		string minName;
		branch_length_type minBranchLength = 0;
		string myName = "";

		for (Vec<PhyTreeNode*>::iterator it = node->_children.begin(); it != node->_children.end(); ++it)
		{
			string curName;
			branch_length_type curBranchLength = _closestNode(*it, curName);

			if (!minName.size() || curBranchLength < minBranchLength)
			{
				minName = curName;
				minBranchLength = curBranchLength;
			}
		}
		name = minName;
		return minBranchLength + node->_branchLengthToParent;
	}
}

unsigned int PhyTreeNode::_setDepth(PhyTreeNode* node)
{
	unsigned int depth = 0;
	for (Vec<PhyTreeNode*>::iterator it = node->_children.begin(); it != node->_children.end(); ++it)
	{
		unsigned int curDepth = _setDepth(*it);
		if (curDepth > depth)
		{
			depth = curDepth;
		}
	}
	node->_depth = depth + 1;
	return node->_depth;
}

ostream& operator<<(ostream& out, PhyTreeNode& t)
{
	if (t._children.size() == 0)
	{
		out << t._name;
		out << ":" << t._branchLengthToParent;
	}
	else
	{
		//out << t._branchLengthToParent << "-" << "(" << endl;
		out << "(";
		for (unsigned int i = 0; i < t._children.size(); i++)
		{
			//out << i << ":" << endl;
			if (i != 0)
			{
				out << ",";
			} 
			out << *(t._children[i]);
		}
		out << ")";// << endl;
		out << ":" << t._branchLengthToParent;
	}

	return out;
}	

PhyTreeNode::prob_type PhyTreeNodeNS::kimura(const unsigned int char1, const unsigned int char2, const PhyTreeNode::prob_type alphaT, const PhyTreeNode::prob_type betaToAlpha, const Vec<PhyTreeNode::prob_type>& f)
{
	PhyTreeNode::prob_type betaT = alphaT * betaToAlpha;

	if (char1 != char2)
	{
		if ((char1 == A_INDEX && char2 == G_INDEX) || (char1 == G_INDEX && char2 == A_INDEX) || (char1 == C_INDEX && char2 == T_INDEX) || (char1 == T_INDEX && char2 == C_INDEX))
		{
			return (1 - exp(-4 * betaT)) / 4;
		}		
		else
		{
			return (1 + exp(-4 * betaT) - 2 * exp(-2 * (alphaT + betaT))) / 4;
		}
	}
	else
	{
		return 1 - 2 * ((1 - exp(-4 * betaT)) / 4) - ((1 + exp(-4 * betaT) - 2 * exp(-2 * (alphaT + betaT))) / 4);

	}
}


PhyTreeNode::prob_type PhyTreeNodeNS::toGap(const bool alreadyGapped, const PhyTreeNode::prob_type alphaT, const PhyTreeNode::prob_type betaToAlpha, const Vec<PhyTreeNode::prob_type>& f)
{
	PhyTreeNode::prob_type betaT = alphaT * betaToAlpha;

	if (!alreadyGapped)
	{
		return 2 * ((1 - exp(-4 * betaT)) / 4) + ((1 + exp(-4 * betaT) - 2 * exp(-2 * (alphaT + betaT))) / 4);
	}
	else
	{
		return 1 - 2 * ((1 - exp(-4 * betaT)) / 4) - ((1 + exp(-4 * betaT) - 2 * exp(-2 * (alphaT + betaT))) / 4);

	}
}


