#ifndef _phytreenode_h_included_
#define _phytreenode_h_included_

#include "System.h"
#include "Mseq.h"
#include "ProbVector.h"
#include "Profile.h"
#include "Vec.h"
#include <istream>

/**
 * This class represents a node in a phylogenetic tree
 */

class PhyTreeNode
{
	friend class PhyTree;

 public:

	typedef double branch_length_type;
	typedef ProbVector::prob_type prob_type;

	PhyTreeNode(istream& is, Mseq& sequence, const string& queryName = "");

	PhyTreeNode(PhyTreeNode* parent) : _parent(parent), _branchLengthToParent(0) {  }

	PhyTreeNode(const PhyTreeNode& t) { }

	~PhyTreeNode();

	void getProbVector(ProbVector& pv, const bool memOpt = false, const prob_type alpha = 2, const prob_type beta = 1, const Vec<prob_type>& f = Vec<prob_type>(0));

	friend std::ostream& operator<<(std::ostream& out, PhyTreeNode&);

 private:

	static const double INF_BRANCH_LENGTH = 50;

	void _getProbVector(ProbVector& pv, unsigned int begin, unsigned int end, const prob_type betaToAlpha, const Vec<prob_type>& f);

	unsigned int _getLength() const;
	
	static branch_length_type _getBranchLength(double numSubstitutionsPerSite, const prob_type betaToAlpha, const Vec<prob_type>& f);

	static void _evolveFromChildren(Vec<ProbVector>& childPv, Vec<branch_length_type>& branchLengths, ProbVector& pv, const prob_type betaToAlpha, const Vec<prob_type>& f);

	static bool _getNode(PhyTreeNode* node, istream& is);

	static branch_length_type _getMinBranchLength(PhyTreeNode* node);

	static void _setNoBranchLengthToLength(PhyTreeNode* node, const branch_length_type length);

	static bool _eliminateNodes(PhyTreeNode* node, const Vec<string>& names, const bool checkNames = true);

	static PhyTreeNode* _presentInTree(PhyTreeNode* node, const string& name);

	static void _makeBinary(PhyTreeNode* node);

	static void _storeSequences(PhyTreeNode* node, Mseq& sequence);

	static void _makeRoot(PhyTreeNode* node);

	static branch_length_type _closestNode(PhyTreeNode* node, string& name);

	static unsigned int _setDepth(PhyTreeNode* node);

	//this is the data storage
	string _name;
	unsigned int _depth;
	Vec<PhyTreeNode*> _children;
	PhyTreeNode* _parent;
	branch_length_type _branchLengthToParent;
	Seq _sequence;

};

namespace PhyTreeNodeNS
{

	const static unsigned int A_INDEX = NUCL_A_PROF_IDX;
	const static unsigned int C_INDEX = NUCL_C_PROF_IDX;
	const static unsigned int T_INDEX = NUCL_T_PROF_IDX;
	const static unsigned int G_INDEX = NUCL_G_PROF_IDX;

	PhyTreeNode::prob_type kimura(const unsigned int char1, const unsigned int char2, const PhyTreeNode::prob_type alphaT, const PhyTreeNode::prob_type betaToAlpha, const Vec<PhyTreeNode::prob_type>& f);

	PhyTreeNode::prob_type toGap(const bool alreadyGapped, const PhyTreeNode::prob_type alphaT, const PhyTreeNode::prob_type betaToAlpha, const Vec<PhyTreeNode::prob_type>& f);
}

#endif
