#ifndef _regionlist_h_included_
#define _regionlist_h_included_

#include "System.h"
#include "Region.h"
#include "Vec.h"
#include <map>
#include <set>

class RegionList
{
 public:

	typedef double prob_type;
	typedef unsigned int region_num_type;
	typedef pair<prob_type, prob_type> region_attr_type;

	static prob_type getWeight(const region_attr_type t)
		{
			return t.first;
		}

	static prob_type getConsLevel(const region_attr_type t)
		{
			return t.second;
		}

	void clear()
		{
			_regions.clear();
			_regionToNumMap.clear();
			_regionTypes.clear();
			_regionTypeList.clear();
		}

	void addRegion(const Region& r, prob_type weight, prob_type consLevel)
		{
			//first check if the region already exists
			region_attr_type aT(weight, consLevel);
			region_num_type regionNum = 0;
			if (_regionTypeList.find(aT) != _regionTypeList.end())
			{
				regionNum = _regionTypeList[aT];
			}
			else
			{
				regionNum = _regionTypes.size();
				pair<region_attr_type, region_num_type> p(aT, regionNum);
				_regionTypeList.insert(p);
				_regionTypes.push_back(region_attr_type(weight, consLevel));
			}
			_regions.push_back(r);
			_regionToNumMap.push_back(regionNum);
		}

	void getRegionTypes(Vec<region_attr_type>& regionTypeList) const
		{
			std::set<region_attr_type> types;
			for (map<region_attr_type, region_num_type>::const_iterator it = _regionTypeList.begin(); it != _regionTypeList.end(); ++it)
			{
				types.insert(it->first);
			}				
			regionTypeList.resize(types.size());
			copy(types.begin(), types.end(), regionTypeList.begin());
			sort(regionTypeList.begin(), regionTypeList.end());
		}
	
	region_num_type getNumRegions() const
		{
			return _regions.size();
		}
	Region getRegion(unsigned int position) const
		{
			return _regions[position];
		}

	void setStart(unsigned int position, unsigned int start)
		{
			_regions[position].setStart(start);
		}

	void setEnd(unsigned int position, unsigned int end)
		{
			_regions[position].setEnd(end);
		}

	region_attr_type getRegionType(unsigned int position) const
		{
			return _regionTypes[_regionToNumMap[position]];
		}

	void breakRegions(unsigned int maxRegionLength)
		{
			Vec<Region> regionCopy(_regions.size());
			copy(_regions.begin(), _regions.end(), regionCopy.begin());

			Vec<region_num_type> regionToNumMapCopy(_regionToNumMap.size());
			copy(_regionToNumMap.begin(), _regionToNumMap.end(), regionToNumMapCopy.begin());

			_regions.clear();
			_regionToNumMap.clear();

			unsigned int numRegions = regionCopy.size();
			const unsigned int rubberMax = static_cast<unsigned int>(maxRegionLength * 1.5);
			for (unsigned int i = 0; i < numRegions;)
			{
				Region& r = regionCopy[i];
				if (r.getLength() < rubberMax)
				{
					_regions.push_back(r);
					_regionToNumMap.push_back(regionToNumMapCopy[i]);
					++i;
				}
				else
				{
					unsigned int newEnd = r.getStart() + maxRegionLength - 1;
					Region newRegion(r.getStart(), newEnd);
					r.setStart(newEnd + 1);
					_regions.push_back(newRegion);
					_regionToNumMap.push_back(regionToNumMapCopy[i]);
				}
			}
		}
	
	void printStats() const
		{

			for (map<region_attr_type, region_num_type>::const_iterator it = _regionTypeList.begin(); it != _regionTypeList.end(); ++it)
			{
				cout << "For weight of " << it->first.first << " and cons level of " << it->first.second << ":" << endl;
				region_num_type num = it->second;
				unsigned int minSize = 10000000;
				unsigned int maxSize = 0;
				unsigned int avgSize = 0;
				unsigned int numTotal = 0;
				Region minRegion(0,0);
				for (unsigned int i = 0; i < _regions.size(); i++)
				{
					if (_regionToNumMap[i] == num)
					{
						const Region& r = _regions[i];
						unsigned int curSize = r.getEnd() - r.getStart() + 1;
						if (curSize > maxSize)
						{
							maxSize = curSize;
						}
						if (curSize < minSize)
						{
							minSize = curSize;
							minRegion = r;
						}
						avgSize += curSize;
						numTotal++;
					}
				}
				cout << "Min size: " << minSize << " (" << minRegion.getStart() << "-" << minRegion.getEnd() << ")" << endl;
				cout << "Max size: " << maxSize << endl;
				cout << "Tot size: " << avgSize << endl;
				if (numTotal > 0)
				{
					avgSize /= numTotal;
				}
				cout << "Avg size: " << avgSize << endl;
				cout << "Num tot : " << numTotal << endl;
			}
		}
	
 private:
	
	Vec<Region> _regions;
	Vec<region_num_type> _regionToNumMap;
	Vec<region_attr_type> _regionTypes;
	map<region_attr_type, region_num_type> _regionTypeList;

};

#endif
