#include "HeuristicIndex.h"
#include "Util.h"
#include "Index.h"
#include "MatchFinder.h"
#include "AllLetterPolicy.h"
#include "ConsensusLetterPolicy.h"
#include "MandalaWrapper.h"
#include "OnePatternSetSelector.h"
#include "FilePatternSetSelector.h"
#include "CachedPatternSetSelector.h"
#include "LocalHit.h"
#include "Pattern.h"
#include "PhyTree.h"
#include "RegionList.h"
#include "ProbVector.h"
#include "SimpleHMM.h"
#include "TyphonMarkovModel.h"
#include "NextPatternSetGenerator.h"
#include "NextPatternSetGeneratorOneRegion.h"
#include "NextPatternSetGeneratorByRegionClass.h"
#include "gnu_getopt.h"
#include <ctime>
#include <string>
#include <sstream>
#include <ostream>

void printUsage();
void printHelp();
void printLongUsage();


static struct option longOptions[] = 
	{
		{"help", 0, 0, 0}, //0
		//these are parameters for running the aligner
		{"startdb", 1, 0, 0}, //1
		{"enddb", 1, 0, 0}, //2
		{"startq", 1, 0, 0}, //3
		{"endq", 1, 0, 0}, //4
		{"mhl", 1, 0, 0}, //5
		{"o", 1, 0, 0}, //6
		{"outputlevel", 1, 0, 0}, //7
		{"w", 1, 0, 0}, //8
		{"s", 1, 0, 0}, //9
		{"pattern", 1, 0, 0}, //10
		{"devs", 1, 0, 0}, //11
		{"S1", 1, 0, 0}, //12
		{"S", 1, 0, 0}, //13
		{"gapS", 1, 0, 0}, //14
		{"gap", 0, 0, 0}, //15
		{"E1", 1, 0, 0}, //16
		{"E", 1, 0, 0}, //17
		{"gapE", 1, 0, 0}, //18
		{"X1", 1, 0, 0}, //19
		{"X2", 1, 0, 0}, //20
		{"gapX", 1, 0, 0}, //21
		{"hitdist", 1, 0, 0}, //22
		{"matrix", 1, 0, 0}, //23
		{"nodust", 0, 0, 0}, //24
		{"dustl", 1, 0, 0}, //25
		{"P", 1, 0, 0}, //26
		{"B", 1, 0, 0}, //27
		{"sepq", 0, 0, 0}, //28
		{"runblast", 0, 0, 0}, //29
		{"runallpairs", 0, 0, 0}, //30
		{"runpattern", 0, 0, 0}, //31
		{"queryname", 1, 0, 0}, //32
		{"numregionclasses", 1, 0, 0}, //33
		{"tree", 1, 0, 0}, //34
		{"patternfile", 1, 0, 0}, //35
		{"memopt", 0, 0, 0}, //36
		{0, 0, 0, 0}
	};



int main(int argc, char** argv)
{
	//time_t seconds = time(NULL);

	//sub ranges to index/query
	Seq::size_type start1 = 0;
	Seq::size_type end1 = 0;
	bool haveEnd1 = false;
	Seq::size_type start2 = 0;
	Seq::size_type end2 = 0;
	bool haveEnd2 = false;
	Seq::size_type maxHitLength = 0;

	//where we write out the hits
	bool writeOutToFile = false;
	string outFileName;
	LocalHit::OutputType outputType = LocalHit::LEVEL1;

	//this specifies the properties of the pattern we'll index
	TyphonNS::Pattern::size_type minPatternWeight = 7;
	TyphonNS::Pattern::size_type maxPatternWeight = 12;
	TyphonNS::Pattern::size_type minPatternSpanDiff = 6;
	TyphonNS::Pattern::size_type maxPatternSpanDiff = 10;

	TyphonNS::Pattern::size_type patternWeight = 0;
	TyphonNS::Pattern::size_type patternSpan = 0;
	TyphonNS::Pattern pattern;
	Vec<TyphonNS::Pattern> patterns;

	//filtering for the index
	float numDevs = 0;
	bool haveNumDevs = false;

	//scoring parameters; these defaults are in MatchFinder
	double eValue = 0;
	double semiEValue = 0;
	double eValueToDoGapped = 0;
	double bitFinalThreshold = 0;
	double bitSemiFinalThreshold = 0;
	double bitThresholdToDoGapped = 0;
	double bitCutoffThreshold = 0;
	double bitSemiCutoffThreshold = 0;
	double bitCutoffForGapped = 0;
	bool doGapped = false;
	Seq::size_type hspSizeForGapped = 0;

	//extension/scoring information
	Seq::size_type maxDistance = 0;
	MatchFinder::ScoringType scoringType = MatchFinder::ConsensusScoring;

	//scoring parameter information
	bool hasConservationScoringParamFileName = false;
	string conservationScoringParamFileName;
	bool hasScoringParamFileName = false;
	string scoringParamFileName;

	//Typhon specific parameters
	Vec<TyphonNS::Pattern>::size_type maxNumPatterns = 20;
	Vec<TyphonNS::Pattern>::size_type numPatterns = 0;
	Vec<TyphonNS::Pattern>::size_type maxBudget = 20;
	double avgNumPatternsPerPositionForIndex = 0;

	//by default, run Typhon consensus
	enum ProgramType {BLAST, PATTERN, TYPHON};
	ProgramType program = TYPHON;
	bool runAllPairs = false;

	//dust options
	bool dustIndex = false;
	bool dustQuery = true;
	unsigned int dustLevel = 0;

	//assorted options
	bool separateQueries = false;
	float ratioToIndex = 1;

	unsigned int minRegionLength = 32;
	unsigned int numRegionClasses = 40;
	double simLevel = .6;
	string queryName = "";
	string treeFile = "";

	string patternFile = "";
	bool havePatternFile = false;

	bool memOpt = false;

	while (true)
	{
		int longIndex;

		int c = gnu_getopt_long_only(argc, argv, "", longOptions, &longIndex);

		if (c == -1)
		{
			break;
		}
		switch(c)
		{
			case 0:
			{
				switch(longIndex)
				{
					case 0:
					{
						printLongUsage();
						exit(0);
					}
					case 1:
					{
						start1 = atoi(optarg) - 1;
						break;
					}
					case 2:
					{
						end1 = atoi(optarg) - 1;
						haveEnd1 = true;
						break;
					}
					case 3:
					{
						start2 = atoi(optarg) - 1;
						break;
					}
					case 4:
					{

						end2 = atoi(optarg) - 1;
						haveEnd2 = true;
						break;
					}
					case 5:
					{
						maxHitLength = atoi(optarg);
						break;
					}
					case 6:
					{
						outFileName = optarg;
						writeOutToFile = true;
						break;
					}
					case 7:
					{
						outputType = static_cast<LocalHit::OutputType>(atoi(optarg));
						break;
					}
					case 8:
					{
						patternWeight = atoi(optarg);
						break;
					}
					case 9:
					{
						patternSpan = atoi(optarg);
						break;
					}
					case 10:
					{
						unsigned int stringLength = strlen(optarg);
						char patternString[stringLength + 1];
						strcpy(patternString, optarg);
						pattern = TyphonNS::Pattern(patternString);
						break;
					}
					case 11:
					{
						numDevs = atof(optarg);
						haveNumDevs = true;
						break;
					}
					case 12:
					{
						bitSemiFinalThreshold = atof(optarg);
						break;
					}
					case 13:
					{
						bitFinalThreshold = atof(optarg);
						break;
					}
					case 14:
					{
						bitThresholdToDoGapped = atof(optarg);
						break;
					}
					case 15:
					{
						doGapped = true;
						break;
					}
					case 16:
					{
						semiEValue = atof(optarg);
						break;
					}
					case 17:
					{
						eValue = atof(optarg);
						break;
					}
					case 18:
					{
						eValueToDoGapped = atof(optarg);
						break;
					}
					case 19:
					{
						bitSemiCutoffThreshold = atof(optarg);
						break;
					}
					case 20:
					{
						bitCutoffThreshold = atof(optarg);
						break;
					}
					case 21:
					{
						bitCutoffForGapped = atof(optarg);
						break;
					}
					case 22:
					{
						maxDistance = atoi(optarg);
						break;
					}
					case 23:
					{
						hasScoringParamFileName = true;
						scoringParamFileName = optarg;
						break;
					}
					case 24:
					{
						dustQuery = false;
						break;
					}
					case 25:
					{
						dustLevel = atoi(optarg);
						break;
					}
					case 26:
					{
						numPatterns = atoi(optarg);
						break;
					}
					case 27:
					{
						avgNumPatternsPerPositionForIndex = atof(optarg);
						break;
					}
					case 28:
					{
						separateQueries = true;
						break;
					}
					case 29:
					{
						program = BLAST;
						break;
					}
					case 30:
					{
						runAllPairs = true;
						break;
					}
					case 31:
					{
						program = PATTERN;
						break;
					}
					case 32:
					{
						queryName = optarg;
						break;
					}
					case 33:
					{
						numRegionClasses = atoi(optarg);
						break;
					}
					case 34:
					{
						treeFile = optarg;
						break;
					}
					case 35:
					{
						havePatternFile = true;
						patternFile = optarg;
						break;
					}
					case 36:
					{
						memOpt = true;
					}
				}
			}
		}
	}

	//now we need to fill in defaults for every thing that the user didn't enter
	const TyphonNS::Pattern::size_type DEFAULT_WEIGHT = 10;

	if (program != BLAST)
	{
		if (!patternWeight)
		{
			patternWeight = DEFAULT_WEIGHT;
		}
		if (patternWeight < minPatternWeight)
		{
			cerr << "Pattern weight must be >= " << minPatternWeight << "; setting to " << minPatternWeight << endl;
			patternWeight = minPatternWeight;
		}
		else if (patternWeight > maxPatternWeight)
		{
			cerr << "Pattern weight must be <= " << maxPatternWeight << "; setting to " << maxPatternWeight << endl;
			patternWeight = maxPatternWeight;
		}

		if (!patternSpan || patternSpan < patternWeight)
		{
			patternSpan = patternWeight + 8;
		}
		else if (patternSpan < patternWeight)
		{
			patternSpan = patternWeight;
		}

		if (patternSpan - patternWeight < minPatternSpanDiff)
		{
			cerr << "Pattern span must be >= " << patternWeight + minPatternSpanDiff << "; setting to " << patternWeight + minPatternSpanDiff << endl;
			patternSpan = patternWeight + minPatternSpanDiff;
		}
		else if (patternSpan - patternWeight > maxPatternSpanDiff)
		{
			cerr << "Pattern span must be <= " << patternWeight + maxPatternSpanDiff << "; setting to " << patternWeight + maxPatternSpanDiff << endl;
			patternSpan = patternWeight + maxPatternSpanDiff;
		}
	}
	else
	{
		if (pattern.getKLength() == 0)
		{
			if (!patternWeight)
			{
				pattern = TyphonNS::Pattern(DEFAULT_WEIGHT);
			}
			else
			{
				pattern = TyphonNS::Pattern(patternWeight);
			}
		}
	}

	if (!haveNumDevs)
	{
		numDevs = 5;
	}

	if (doGapped && !hspSizeForGapped)
	{
		hspSizeForGapped = 11;
	}

	if (!numPatterns)
	{
		numPatterns = 10;
	}

	if (numPatterns > maxNumPatterns)
	{
		cerr << "Number of patterns must be <= " << maxNumPatterns << "; setting to " << maxNumPatterns << endl;
		numPatterns = maxNumPatterns;
	}

	if (avgNumPatternsPerPositionForIndex == 0)
	{
		avgNumPatternsPerPositionForIndex = 2;
	}

	if (avgNumPatternsPerPositionForIndex > maxBudget)
	{
		cerr << "Budget must be <= " << maxBudget << "; setting to " << maxBudget << endl;
		numPatterns = maxNumPatterns;
	}

	if (!dustLevel)
	{
		dustLevel = 15;
	}

	if (program != BLAST)
	{
		if (havePatternFile)
		{
			FilePatternSetSelector fps;
			fps.getBestPatterns(patternFile, patternWeight, patternSpan, numPatterns, patterns);
			
		}
		
		{
			CachedPatternSetSelector cps;
			cps.getBestPatterns(patternWeight, patternSpan, numPatterns, patterns);
		}
	}

	//need to start from setting up and reading in
	if (optind + 1 >= argc)
	{
		printHelp();
		exit(0);
	}
	try
	{
		//get the sequence file names
		string firstFileName = argv[optind++];
		string secondFileName = argv[optind++];

		//get the scoring parameters filenames
		ScoringParameters<MatchFinder::score_type> sMain;
		if (hasScoringParamFileName)
		{
			try
			{
				sMain.read_from_file(scoringParamFileName);
			}
			catch (const BIO::E_BioEx& e)
			{
				throw BIO::E_BioEx("Problem with scoring parameter file `" + scoringParamFileName + "': " + e.getMessage());
			}
		}
		else
		{
			string tempS = "SubstitutionMatrix\n91 -114  -31 -123 -25\n-114  100 -125  -31 -25\n-31 -125  100 -114 -25\n-123  -31 -114   91 -25\n-25  -25  -25  -25   0\nGapOpen 400";
			sMain.read_from_string(tempS);
		}

		ScoringParameters<MatchFinder::score_type> sConservation;
		if (hasConservationScoringParamFileName)
		{
			try
			{
				sConservation.read_from_file(conservationScoringParamFileName);
			}
			catch (const BIO::E_BioEx& e)
			{
				throw BIO::E_BioEx("Problem with scoring parameter file `" + conservationScoringParamFileName + "': " + e.getMessage());
			}
		}
		else
		{
			string tempS = "SubstitutionMatrix\n1 0  0 0 -1\n0  1 0  0 -1\n0 0 1 0 -1\n0  0 0   1 -1\n-1  -1  -1  -1   -1";
			sConservation.read_from_string(tempS);
		}

		Mseq sequence1;

		if (start1 || end1)
		{
			//we have to read some subset of the string
			Mseq fullSequence1;
			try
			{
				fullSequence1.readFromFile(firstFileName);
			}
			catch (BIO::E_BioEx& e)
			{
				throw BIO::E_BioEx("Problem with alignment in `" + firstFileName + "': " + e.getMessage());
			}

			//since end is base 0 and the end is inclusive
			Seq::size_type lastSequence1Pos = fullSequence1.getLength() - 1;
			if (!haveEnd1)
			{
				end1 = lastSequence1Pos;
			}

			//let's do some checks on the start/end to make sure they are correct
			{
				const Seq::size_type firstPosition = 0;

				if (start1 < firstPosition)
				{
					cerr << "First start value less than 1; setting to 1" << endl;
					start1 = firstPosition;
				}
				else if (start1 > lastSequence1Pos)
				{
					cerr << "First start value greater than first sequence length; setting to sequence length" << endl;
					start1 = lastSequence1Pos;
				}

				if (end1 < firstPosition)
				{
					cerr << "First end value less than 1; setting to 1" << endl;
					end1 = firstPosition;
				}
				else if (end1 > lastSequence1Pos)
				{
					cerr << "First end value greater than first sequence length; setting to sequence length" << endl;
					end1 = lastSequence1Pos;
				}

				if (end1 < start1)
				{
					throw BIO::E_BioEx("Start value 1 must be less than or equal to end value 1");
				}
			}
			//since we store start and end base 0, and this method takes base 1, need to add 1 to start
			sequence1.copyFrom(fullSequence1, start1 + 1, end1 - start1 + 1);
		}
		else
		{
			try
			{
				sequence1.readFromFile(firstFileName);
			}
			catch (const BIO::E_BioEx& e)
			{
				throw BIO::E_BioEx("Problem with alignment in `" + firstFileName + "': " + e.getMessage());
			}
		}
		//do dusting if required
		if (dustIndex)
		{
			sequence1.dustMseq(dustLevel);
		}

		HeuristicIndex* i = NULL;
		Index* in = NULL;

		//build the right kind of index
		double indexTime = 0;
		double scanTime = 0;

		if (program == TYPHON || program == PATTERN)
		{
			ifstream treeF(treeFile.c_str());
			PhyTree tree(treeF, sequence1, queryName);
			ProbVector pv;
			clock_t startTime = clock();
			cerr << "Construting profile...";
			tree.getProbVector(pv, memOpt);
			cerr << "in " << (clock() - startTime) / ((double)CLOCKS_PER_SEC) << " seconds" << endl;
// 			if (TEMP_PVS || TEMP_PVE)
// 			{
// 				if (!TEMP_PVS)
// 				{
// 					TEMP_PVS = 1;
// 				}
// 				if (!TEMP_PVE)
// 				{
// 					TEMP_PVE = pv.size() + 1;
// 				}
// 				//print probvector
// 				for (unsigned int i = TEMP_PVS - 1 ; i < TEMP_PVE - 1; i++)
// 				{
// 					cout << i << "\t";
// 					for (int j = 0; j < 4; j++)
// 					{
// 						cout << j << ":" << pv.getProbability(i,j) << " ";
// 					}
// 					cout << 4 << ":" << pv.getPresentProbability(i) << " ";
// 					cout << 5 << ":" << pv.getMaxLetterProb(i) << " ";
// 					cout << endl;
// 				}
// 				exit(0);
// 			}

			SimpleHMM hmm(minRegionLength);
			RegionList r;
			startTime = clock();
			cerr << "Decoding profile...";
			hmm.adaptiveDecode(pv, numRegionClasses, r);
			cerr << "in " << (clock() - startTime) / ((double)CLOCKS_PER_SEC) << " seconds" << endl;
			if (program == PATTERN)
			{
				numPatterns = static_cast<unsigned int>(avgNumPatternsPerPositionForIndex);
			}

			OnePatternSetSelector ops;
			ops.getBestPatterns(simLevel, numPatterns, patternWeight, patternSpan, patterns);

			ConsensusLetterPolicy policy;

			NextPatternSetGenerator* generator = NULL;
			if (program == PATTERN)
			{
				generator = new NextPatternSetGeneratorOneRegion();
			}
			else
			{
				generator = new NextPatternSetGeneratorByRegionClass();
			}

			cerr << "Building index...";
			startTime = clock();
			in = new Index(pv, r, policy, *generator, numDevs, patterns, avgNumPatternsPerPositionForIndex);
			indexTime = (clock() - startTime) / ((double)CLOCKS_PER_SEC);
			cerr << "in " << indexTime << " seconds" << endl;
			delete generator;
		}
		else
		{
			if (program == BLAST)
			{
				cerr << "Building index...";
				clock_t startTime = clock();
				Vec<TyphonNS::Pattern> patternSet(1, pattern);
				if (runAllPairs)
				{
					AllLetterPolicy policy;
					i = new HeuristicIndex(sequence1, policy, sConservation, numDevs, patternSet, 0, 0, 1, ratioToIndex);
				}
				else
				{
					ConsensusLetterPolicy policy;
					i = new HeuristicIndex(sequence1, policy, sConservation, numDevs, patternSet, 0, 0, 1, ratioToIndex);
				}
				indexTime = (clock() - startTime) / ((double)CLOCKS_PER_SEC);
				cerr << "in " << indexTime << " seconds" << endl;

			}
			// else
// 			{
// 				Vec<TyphonNS::Pattern> patternSet(numPatterns);
// 				if (runAllPairs)
// 				{
// 					AllLetterPolicy policy;
// 					i = new HeuristicIndex(sequence1, policy, sConservation, numDevs, patternSet, patternWeight, patternSpan - patternWeight, avgNumPatternsPerPositionForIndex, ratioToIndex);
// 				}
// 				else
// 				{
// 					ConsensusLetterPolicy policy;
// 					i = new HeuristicIndex(sequence1, policy, sConservation, numDevs, patternSet, patternWeight, patternSpan - patternWeight, avgNumPatternsPerPositionForIndex, ratioToIndex);
// 				}
// 			}
		}

		MatchFinder m(bitSemiFinalThreshold
									,bitFinalThreshold
									,bitThresholdToDoGapped
									,semiEValue
									,eValue
									,eValueToDoGapped
									,bitSemiCutoffThreshold
									,bitCutoffThreshold
									,bitCutoffForGapped
									,doGapped
									,hspSizeForGapped
									,maxDistance
									,scoringType
									,maxHitLength
									);
		m.setSequence1(sequence1, sMain);

		//now do the second sequence
		bool doneFile = false;
		bool newFile = true;
		char buffer[BIO::IO_BUF_SIZE];
		ifstream is;
	
		/* Associate buffer with stream */
		is.rdbuf()->pubsetbuf(buffer, BIO::IO_BUF_SIZE);

		/* Open the file */
		if (!(is.rdbuf()->open(secondFileName.c_str(), ios_base::in)))
		{
			throw BIO::E_OpenFailedEx();
		}

		//cerr << "Pre scan took " << time(NULL) - seconds << " seconds" << endl;

		ostream* outStreamPtr = &cout;

		ofstream outFile;
		if (writeOutToFile)
		{
			outFile.open(outFileName.c_str());
			if (outFile)
			{
				outStreamPtr = &outFile;
			}
		}

		ostream& outStream = *outStreamPtr;

		while (!doneFile)
		{
			Mseq sequence2;
			
			{
				Mseq fullSequence2;

				if (!separateQueries)
				{
					doneFile = true;
					try
					{
						fullSequence2.readFromFile(is);
					}
					catch (const BIO::E_BioEx& e)
					{
						throw BIO::E_BioEx("Problem with alignment in `" + secondFileName + "': " + e.getMessage());
					}
				}
				else
				{
					try
					{
						doneFile = !fullSequence2.readSeqFromFile(is);
					}
					catch (const BIO::E_BioEx& e)
					{
						throw BIO::E_BioEx("Problem with alignment in `" + secondFileName + "': " + e.getMessage());
					}
				}

				Seq::size_type lastSequence2Pos = fullSequence2.getLength() - 1;
				
				if (!haveEnd2)
				{
					end2 = lastSequence2Pos;
				}

				{
					const Seq::size_type firstPosition = 0;
					
					if (start2 < firstPosition)
					{
						start2 = firstPosition;
					}
					else if (start2 > lastSequence2Pos)
					{
						start2 = lastSequence2Pos;
					}
					
					if (end2 < firstPosition)
					{
						end2 = firstPosition;
					}
					else if (end2 > lastSequence2Pos)
					{
						end2 = lastSequence2Pos;
					}
					
					if (end2 < start2)
					{
						throw BIO::E_BioEx("Start value 2 must be less than or equal to end value 2");
					}
				}
				//since we store start and end base 0, and this method takes base 1, need to add 1 to start
				sequence2.copyFrom(fullSequence2, start2 + 1, end2 - start2 + 1);
			}
			
			if (dustQuery)
			{
				sequence2.dustMseq(dustLevel);
			}

			m.setSequence2(sequence2);
			if (in != NULL)
			{
					ConsensusLetterPolicy policy;
					clock_t startTime = clock();
					cerr << "Scanning index...";
					in->scanIndex(policy, sequence1, sequence2, m, sMain, start1, start2);
					scanTime = (clock() - startTime) / ((double)CLOCKS_PER_SEC);
					cerr << "in " << scanTime << " seconds" << endl;
			}
			else
			{
				clock_t startTime = clock();
				cerr << "Scanning index...";

				if (runAllPairs)
				{
					AllLetterPolicy policy;
					i->scanIndex(policy, sequence1, sequence2, m, sMain, start1, start2);
				}
				else
				{
					ConsensusLetterPolicy policy;
					i->scanIndex(policy, sequence1, sequence2, m, sMain, start1, start2);
				}
				scanTime = (clock() - startTime) / ((double)CLOCKS_PER_SEC);
				cerr << "in " << scanTime << " seconds" << endl;
			}

			m.writeOutResults(sequence1, sequence2, outStream, outputType, indexTime, scanTime);
			
			newFile = false;
		}
		
		if (!(is.rdbuf()->close()))
		{
			throw BIO::E_CloseFailedEx();
		}
		delete i;
	}		
	catch(const BIO::E_BioEx& e)
	{
		cerr << "Error: " << e.getMessage() << endl;
		exit(1);
	}
}

void printUsage()
{
	string usage = "";
	usage += "Usage: typhon [-options] file1 file2";
	cerr << usage << endl;
}

void printHelp()
{
	printUsage();
	cerr << "Try `typhon -help' for more information" << endl;
}

unsigned int numDescriptions = 37;

const string descriptions[] = 
	{
		"",
		"The position at which to start indexing the database",
		"The position at which to stop indexing the database",
		"The position at which to start scanning the query",
		"The position at which to stop scanning the query",
		"If this is non-zero, hits longer than this will be broken into pieces no longer than this value",
		"Write output to this file as oposed to stdout",
		"Control the output level\n0=Only positions of the hit\n1=BlastZ style output\n2=Blast long output",
		"The weight of the pattern. Ignored if a pattern is specified.",
		"The span of patterns.",
		"A pattern to run. This is used only if standard blast is run. It overrides word length.",
		"Words occuring more frequently than this number of deviations above the mean are not indexed.",
		"Threshold (in bits) at which words are kept after the first round of ungapped extensions. Overrides E1.",
		"Threshold (in bits) for HSPs (gapped or ungapped). Overrides E.",
		"Threshold (in bits) at which to do full gapped extensions on ungapped HSPs. Gapped HSPs must score higher than S to be kept as HSPs. Overrides gapE.",
		"Perform gapped extensions.",
		"E value at which words are kept after the first round of ungapped extensions.",
		"E valuefor HSPs (gapped or ungapped).",
		"E value at which to do full gapped extensions on ungapped HSPs. Gapped HSPs must score higher than S to be kept as HSPs.",
		"When scores drop this far (in bits) below the best value seen so far, first round of ungapped extensions is stopped.",
		"When scores drop this far (in bits) below the best value seen so far, final round of ungapped extensions is stopped.",
		"When scores drop this far (in bits) below the best value seen so far, gapped extensions are stopped.",
		"If this is non-zero, then only pairs of hits within a window of this size on the same diagonal are extended.",
		"The scoring matrix to use.",
		"Do not perform filtering using DUST on the query sequence.",
		"Level at which to dust query. ",
		"Number of patterns to index. Used only is standard blast is not run.",
		"Number of patterns to index on average at each position. Used only is standard blast is not run.",
		"Treat query file as a concatenation of multiple queries. Otherwise, it will be assumed to be a multiple alignment.",
		"Run the basic blast algorithm.",
		"Index not the consensus character but every letter in a sequence; valid only for runblast.",
		"Run the basic blast using a set of patterns; every position will be indexed equally",
		"The name of the query. It must match that in the tree exactly",
		"The number of region classes to use",
		"The phylogenetic tree. Entries must match the sequence names in the FASTA file exactly",
		"A file consisting of the patterns to use. Only P will be used. Format is a list of patterns, one per line, each a bit string. Each will be formatted to have the input weight (w) and span (s)",
		"Minimize memory usage; may degrade performance"
	};

const string defaults[] = 
	{
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"0",
		"10",
		"18",
		"",
		"5",
		"",
		"",
		"",
		"",
		"10",
		"E1",
		"E1",
		"23",
		"X1",
		"40",
		"0",
		"HOXD,gapOpen=400",
		"",
		"15",
		"15",
		"2",
		"",
		"",
		"",
		"",
		"",
		"40",
		"",
		"",
		""
	};

const string types[] = 
	{
		"",
		"unsigned int",
		"unsigned int",
		"unsigned int",
		"unsigned int",
		"unsigned int",
		"string",
		"[0,1,2]",
		"unsigned int",
		"unsigned int",
		"{0,1}*",
		"float",
		"unsigned int",
		"unsigned int",
		"unsigned int",
		"",
		"double",
		"double",
		"double",
		"unsigned int",
		"unsigned int",
		"unsigned int",
		"unsigned int",
		"filename",
		"",
		"unsigned int",
		"unsigned int",
		"unsigned int",
		"",
		"",
		"",
		"",
		"string",
		"unsigned int",
		"filename",
		"filename",
		""
	};

void printLongUsage()
{
	cerr << "typhon version beta" << endl << endl;
	cerr << "Copyright (C) 2004-2005 by Jason Flannick and Serafim Batzoglou, all rights reserved" << endl << endl;

	printUsage();
	string usage = "";

	//get the max length
	unsigned int maxSize = 0;
	for (unsigned int i = 1; longOptions[i].name; i++)
	{
		string curName(longOptions[i].name);
		if (curName.size() > maxSize)
		{
			maxSize = curName.size();
		}
	}
	for (unsigned int i = 1; longOptions[i].name; i++)
	{
		if (i >= numDescriptions)
		{
			continue;
		}
		string curName(longOptions[i].name);
		usage +=  "  -";
		unsigned int curEnd = curName.size() + 3;
		unsigned int curStart = curEnd + 2;
		usage += curName;
		for (unsigned int j = 0; j < curStart - curEnd; j++)
		{
			usage += " ";
		}
		//need to transform descriptions
		string description;
		unsigned int numWritten = 0;
		for (unsigned int j = 0; j < descriptions[i].size(); j++)
		{
			if (description.size() > 0 && description[description.size() - 1] == '\n')
			{
				for (unsigned int j = 0; j < curStart + 2; j++)
				{
					description += " ";
				}
				numWritten = 0;

			}
			if (numWritten > 60 && descriptions[i][j] == ' ')
			{
				description += "\n";
			}
			else
			{
				description += descriptions[i][j];
				numWritten++;
			}
		}
		usage += description;
		usage += "\n";
		if (types[i].size() > 0)
		{
			for (unsigned int j = 0; j < curStart + 2; j++)
			{
				usage += " ";
			}
			usage += "type=";
			usage += types[i];
			usage += "\n";
		}

		if (defaults[i].size() > 0)
		{
			for (unsigned int j = 0; j < curStart + 2; j++)
			{
				usage += " ";
			}

			usage += "default=";
			usage += defaults[i];
			usage += "\n";
		}
	}
	cerr << usage;
}
